# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qrl.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='qrl.proto',
  package='qrl',
  syntax='proto3',
  serialized_pb=_b('\n\tqrl.proto\x12\x03qrl\x1a\x1cgoogle/api/annotations.proto\"\x07\n\x05\x45mpty\"\x11\n\x0fGetNodeStateReq\"/\n\x10GetNodeStateResp\x12\x1b\n\x04info\x18\x01 \x01(\x0b\x32\r.qrl.NodeInfo\"\x12\n\x10GetKnownPeersReq\"U\n\x11GetKnownPeersResp\x12 \n\tnode_info\x18\x01 \x01(\x0b\x32\r.qrl.NodeInfo\x12\x1e\n\x0bknown_peers\x18\x02 \x03(\x0b\x32\t.qrl.Peer\"\x11\n\x0fGetPeersStatReq\"5\n\x10GetPeersStatResp\x12!\n\npeers_stat\x18\x01 \x03(\x0b\x32\r.qrl.PeerStat\"\x12\n\x10GetChainStatsReq\"U\n\x11GetChainStatsResp\x12\x12\n\nstate_size\x18\x01 \x01(\x04\x12\x15\n\rstate_size_mb\x18\x02 \x01(\t\x12\x15\n\rstate_size_gb\x18\x03 \x01(\t\")\n\x0bGetStatsReq\x12\x1a\n\x12include_timeseries\x18\x01 \x01(\x08\"\x84\x02\n\x0cGetStatsResp\x12 \n\tnode_info\x18\x01 \x01(\x0b\x32\r.qrl.NodeInfo\x12\r\n\x05\x65poch\x18\x02 \x01(\x04\x12\x16\n\x0euptime_network\x18\x03 \x01(\x04\x12\x19\n\x11\x62lock_last_reward\x18\x04 \x01(\x04\x12\x17\n\x0f\x62lock_time_mean\x18\x05 \x01(\x04\x12\x15\n\rblock_time_sd\x18\x06 \x01(\x04\x12\x1a\n\x12\x63oins_total_supply\x18\x07 \x01(\x04\x12\x15\n\rcoins_emitted\x18\x08 \x01(\x04\x12-\n\x10\x62lock_timeseries\x18\t \x03(\x0b\x32\x13.qrl.BlockDataPoint\"!\n\x13GetAddressFromPKReq\x12\n\n\x02pk\x18\x01 \x01(\x0c\"\'\n\x14GetAddressFromPKResp\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\"\xb2\x01\n\x0e\x42lockDataPoint\x12\x0e\n\x06number\x18\x01 \x01(\x04\x12\x12\n\ndifficulty\x18\x02 \x01(\t\x12\x11\n\ttimestamp\x18\x03 \x01(\x04\x12\x11\n\ttime_last\x18\x04 \x01(\x04\x12\x13\n\x0btime_movavg\x18\x05 \x01(\x04\x12\x12\n\nhash_power\x18\x06 \x01(\x02\x12\x13\n\x0bheader_hash\x18\x07 \x01(\x0c\x12\x18\n\x10header_hash_prev\x18\x08 \x01(\x0c\"g\n\x12GetAddressStateReq\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\x12\x1c\n\x14\x65xclude_ots_bitfield\x18\x02 \x01(\x08\x12\"\n\x1a\x65xclude_transaction_hashes\x18\x03 \x01(\x08\"7\n\x13GetAddressStateResp\x12 \n\x05state\x18\x01 \x01(\x0b\x32\x11.qrl.AddressState\"I\n\x1cGetOptimizedAddressStateResp\x12)\n\x05state\x18\x01 \x01(\x0b\x32\x1a.qrl.OptimizedAddressState\"-\n\x1aGetMultiSigAddressStateReq\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\"G\n\x1bGetMultiSigAddressStateResp\x12(\n\x05state\x18\x01 \x01(\x0b\x32\x19.qrl.MultiSigAddressState\"6\n\nIsSlaveReq\x12\x16\n\x0emaster_address\x18\x01 \x01(\x0c\x12\x10\n\x08slave_pk\x18\x02 \x01(\x0c\"\x1d\n\x0bIsSlaveResp\x12\x0e\n\x06result\x18\x01 \x01(\x08\"\"\n\x0fParseAddressReq\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\"J\n\x10ParseAddressResp\x12\x10\n\x08is_valid\x18\x01 \x01(\x08\x12$\n\x04\x64\x65sc\x18\x02 \x01(\x0b\x32\x16.qrl.AddressDescriptor\"\x1d\n\x0cGetObjectReq\x12\r\n\x05query\x18\x01 \x01(\x0c\"\xbc\x01\n\rGetObjectResp\x12\r\n\x05\x66ound\x18\x01 \x01(\x08\x12\x33\n\raddress_state\x18\x02 \x01(\x0b\x32\x1a.qrl.OptimizedAddressStateH\x00\x12/\n\x0btransaction\x18\x03 \x01(\x0b\x32\x18.qrl.TransactionExtendedH\x00\x12,\n\x0e\x62lock_extended\x18\x04 \x01(\x0b\x32\x12.qrl.BlockExtendedH\x00\x42\x08\n\x06result\"\xb7\x01\n\x10GetLatestDataReq\x12,\n\x06\x66ilter\x18\x01 \x01(\x0e\x32\x1c.qrl.GetLatestDataReq.Filter\x12\x0e\n\x06offset\x18\x02 \x01(\r\x12\x10\n\x08quantity\x18\x03 \x01(\r\"S\n\x06\x46ilter\x12\x07\n\x03\x41LL\x10\x00\x12\x10\n\x0c\x42LOCKHEADERS\x10\x01\x12\x10\n\x0cTRANSACTIONS\x10\x02\x12\x1c\n\x18TRANSACTIONS_UNCONFIRMED\x10\x03\"\xaf\x01\n\x11GetLatestDataResp\x12.\n\x0c\x62lockheaders\x18\x01 \x03(\x0b\x32\x18.qrl.BlockHeaderExtended\x12.\n\x0ctransactions\x18\x02 \x03(\x0b\x32\x18.qrl.TransactionExtended\x12:\n\x18transactions_unconfirmed\x18\x03 \x03(\x0b\x32\x18.qrl.TransactionExtended\"\x82\x01\n\x10TransferCoinsReq\x12\x13\n\x0bmaster_addr\x18\x01 \x01(\x0c\x12\x14\n\x0c\x61\x64\x64resses_to\x18\x02 \x03(\x0c\x12\x0f\n\x07\x61mounts\x18\x03 \x03(\x04\x12\x14\n\x0cmessage_data\x18\x04 \x01(\x0c\x12\x0b\n\x03\x66\x65\x65\x18\x05 \x01(\x04\x12\x0f\n\x07xmss_pk\x18\x06 \x01(\x0c\"T\n\x11TransferCoinsResp\x12?\n\x1d\x65xtended_transaction_unsigned\x18\x01 \x01(\x0b\x32\x18.qrl.TransactionExtended\"B\n\x12PushTransactionReq\x12,\n\x12transaction_signed\x18\x01 \x01(\x0b\x32\x10.qrl.Transaction\"\xca\x01\n\x13PushTransactionResp\x12\x39\n\nerror_code\x18\x01 \x01(\x0e\x32%.qrl.PushTransactionResp.ResponseCode\x12\x19\n\x11\x65rror_description\x18\x02 \x01(\t\x12\x0f\n\x07tx_hash\x18\x03 \x01(\x0c\"L\n\x0cResponseCode\x12\x0b\n\x07UNKNOWN\x10\x00\x12\t\n\x05\x45RROR\x10\x01\x12\x15\n\x11VALIDATION_FAILED\x10\x02\x12\r\n\tSUBMITTED\x10\x03\"\x82\x01\n\x14MultiSigCreateTxnReq\x12\x13\n\x0bmaster_addr\x18\x01 \x01(\x0c\x12\x13\n\x0bsignatories\x18\x02 \x03(\x0c\x12\x0f\n\x07weights\x18\x03 \x03(\r\x12\x11\n\tthreshold\x18\x04 \x01(\r\x12\x0b\n\x03\x66\x65\x65\x18\x05 \x01(\x04\x12\x0f\n\x07xmss_pk\x18\x06 \x01(\x0c\"\xa3\x01\n\x13MultiSigSpendTxnReq\x12\x13\n\x0bmaster_addr\x18\x01 \x01(\x0c\x12\x19\n\x11multi_sig_address\x18\x02 \x01(\x0c\x12\x10\n\x08\x61\x64\x64rs_to\x18\x03 \x03(\x0c\x12\x0f\n\x07\x61mounts\x18\x04 \x03(\x04\x12\x1b\n\x13\x65xpiry_block_number\x18\x05 \x01(\x04\x12\x0b\n\x03\x66\x65\x65\x18\x06 \x01(\x04\x12\x0f\n\x07xmss_pk\x18\x07 \x01(\x0c\"k\n\x12MultiSigVoteTxnReq\x12\x13\n\x0bmaster_addr\x18\x01 \x01(\x0c\x12\x12\n\nshared_key\x18\x02 \x01(\x0c\x12\x0e\n\x06unvote\x18\x03 \x01(\x08\x12\x0b\n\x03\x66\x65\x65\x18\x04 \x01(\x04\x12\x0f\n\x07xmss_pk\x18\x05 \x01(\x0c\"d\n\rMessageTxnReq\x12\x13\n\x0bmaster_addr\x18\x01 \x01(\x0c\x12\x0f\n\x07message\x18\x02 \x01(\x0c\x12\x0f\n\x07\x61\x64\x64r_to\x18\x03 \x01(\x0c\x12\x0b\n\x03\x66\x65\x65\x18\x04 \x01(\x04\x12\x0f\n\x07xmss_pk\x18\x05 \x01(\x0c\"\xad\x01\n\x0bTokenTxnReq\x12\x13\n\x0bmaster_addr\x18\x01 \x01(\x0c\x12\x0e\n\x06symbol\x18\x02 \x01(\x0c\x12\x0c\n\x04name\x18\x03 \x01(\x0c\x12\r\n\x05owner\x18\x04 \x01(\x0c\x12\x10\n\x08\x64\x65\x63imals\x18\x05 \x01(\x04\x12,\n\x10initial_balances\x18\x06 \x03(\x0b\x32\x12.qrl.AddressAmount\x12\x0b\n\x03\x66\x65\x65\x18\x07 \x01(\x04\x12\x0f\n\x07xmss_pk\x18\x08 \x01(\x0c\"\x85\x01\n\x13TransferTokenTxnReq\x12\x13\n\x0bmaster_addr\x18\x01 \x01(\x0c\x12\x14\n\x0c\x61\x64\x64resses_to\x18\x02 \x03(\x0c\x12\x14\n\x0ctoken_txhash\x18\x03 \x01(\x0c\x12\x0f\n\x07\x61mounts\x18\x04 \x03(\x04\x12\x0b\n\x03\x66\x65\x65\x18\x05 \x01(\x04\x12\x0f\n\x07xmss_pk\x18\x06 \x01(\x0c\"i\n\x0bSlaveTxnReq\x12\x13\n\x0bmaster_addr\x18\x01 \x01(\x0c\x12\x11\n\tslave_pks\x18\x02 \x03(\x0c\x12\x14\n\x0c\x61\x63\x63\x65ss_types\x18\x03 \x03(\r\x12\x0b\n\x03\x66\x65\x65\x18\x04 \x01(\x04\x12\x0f\n\x07xmss_pk\x18\x05 \x01(\x0c\"i\n\rLatticeTxnReq\x12\x13\n\x0bmaster_addr\x18\x01 \x01(\x0c\x12\x0b\n\x03pk1\x18\x02 \x01(\x0c\x12\x0b\n\x03pk2\x18\x03 \x01(\x0c\x12\x0b\n\x03pk3\x18\x04 \x01(\x0c\x12\x0b\n\x03\x66\x65\x65\x18\x05 \x01(\x04\x12\x0f\n\x07xmss_pk\x18\x06 \x01(\x0c\"H\n\x0fMiniTransaction\x12\x18\n\x10transaction_hash\x18\x01 \x01(\t\x12\x0b\n\x03out\x18\x02 \x01(\x08\x12\x0e\n\x06\x61mount\x18\x03 \x01(\x04\"$\n\x11GetTransactionReq\x12\x0f\n\x07tx_hash\x18\x01 \x01(\x0c\"\xa0\x01\n\x12GetTransactionResp\x12\x1c\n\x02tx\x18\x01 \x01(\x0b\x32\x10.qrl.Transaction\x12\x15\n\rconfirmations\x18\x02 \x01(\x04\x12\x14\n\x0c\x62lock_number\x18\x03 \x01(\x04\x12\x19\n\x11\x62lock_header_hash\x18\x04 \x01(\x0c\x12\x11\n\ttimestamp\x18\x05 \x01(\x04\x12\x11\n\taddr_from\x18\x06 \x01(\x0c\"^\n\x1fGetMiniTransactionsByAddressReq\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\x12\x15\n\ritem_per_page\x18\x02 \x01(\x04\x12\x13\n\x0bpage_number\x18\x03 \x01(\x04\"d\n GetMiniTransactionsByAddressResp\x12/\n\x11mini_transactions\x18\x01 \x03(\x0b\x32\x14.qrl.MiniTransaction\x12\x0f\n\x07\x62\x61lance\x18\x02 \x01(\x04\"Z\n\x1bGetTransactionsByAddressReq\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\x12\x15\n\ritem_per_page\x18\x02 \x01(\x04\x12\x13\n\x0bpage_number\x18\x03 \x01(\x04\"T\n\x1cGetTransactionsByAddressResp\x12\x34\n\x13transactions_detail\x18\x01 \x03(\x0b\x32\x17.qrl.GetTransactionResp\"\xb8\x02\n\x1fGetMultiSigSpendTxsByAddressReq\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\x12\x15\n\ritem_per_page\x18\x02 \x01(\x04\x12\x13\n\x0bpage_number\x18\x03 \x01(\x04\x12\x44\n\x0b\x66ilter_type\x18\x04 \x01(\x0e\x32/.qrl.GetMultiSigSpendTxsByAddressReq.FilterType\"\x91\x01\n\nFilterType\x12\x08\n\x04NONE\x10\x00\x12\x11\n\rEXECUTED_ONLY\x10\x01\x12\x10\n\x0cNON_EXECUTED\x10\x02\x12\x0b\n\x07\x45XPIRED\x10\x03\x12\x0f\n\x0bNON_EXPIRED\x10\x04\x12\x18\n\x14NON_EXECUTED_EXPIRED\x10\x05\x12\x1c\n\x18NON_EXECUTED_NON_EXPIRED\x10\x06\"X\n GetMultiSigSpendTxsByAddressResp\x12\x34\n\x13transactions_detail\x18\x01 \x03(\x0b\x32\x17.qrl.GetTransactionResp\"2\n\x0fGetVoteStatsReq\x12\x1f\n\x17multi_sig_spend_tx_hash\x18\x01 \x01(\x0c\"6\n\x10GetVoteStatsResp\x12\"\n\nvote_stats\x18\x01 \x01(\x0b\x32\x0e.qrl.VoteStats\"U\n\x1dGetInboxMessagesByAddressResp\x12\x34\n\x13transactions_detail\x18\x01 \x03(\x0b\x32\x17.qrl.GetTransactionResp\"l\n\x0cInboxMessage\x12\x11\n\taddr_from\x18\x01 \x01(\x0c\x12\x11\n\ttimestamp\x18\x02 \x01(\x04\x12\x0f\n\x07message\x18\x03 \x01(\x0c\x12\x0f\n\x07tx_hash\x18\x04 \x01(\x0c\x12\x14\n\x0c\x62lock_number\x18\x05 \x01(\x04\"R\n\x0bTokenDetail\x12\x14\n\x0ctoken_txhash\x18\x01 \x01(\x0c\x12\x0c\n\x04name\x18\x02 \x01(\x0c\x12\x0e\n\x06symbol\x18\x03 \x01(\x0c\x12\x0f\n\x07\x62\x61lance\x18\x04 \x01(\x04\"A\n\x16GetTokensByAddressResp\x12\'\n\rtokens_detail\x18\x01 \x03(\x0b\x32\x10.qrl.TokenDetail\"9\n\x0bSlaveDetail\x12\x15\n\rslave_address\x18\x01 \x01(\x0c\x12\x13\n\x0b\x61\x63\x63\x65ss_type\x18\x02 \x01(\x04\"A\n\x16GetSlavesByAddressResp\x12\'\n\rslaves_detail\x18\x01 \x03(\x0b\x32\x10.qrl.SlaveDetail\"J\n\x10LatticePKsDetail\x12\x0b\n\x03pk1\x18\x01 \x01(\x0c\x12\x0b\n\x03pk2\x18\x02 \x01(\x0c\x12\x0b\n\x03pk3\x18\x03 \x01(\x0c\x12\x0f\n\x07tx_hash\x18\x04 \x01(\x0c\"O\n\x1aGetLatticePKsByAddressResp\x12\x31\n\x12lattice_pks_detail\x18\x01 \x03(\x0b\x32\x15.qrl.LatticePKsDetail\"2\n\x0eMultiSigDetail\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\x12\x0f\n\x07\x62\x61lance\x18\x02 \x01(\x04\"R\n!GetMultiSigAddressesByAddressResp\x12-\n\x10multi_sig_detail\x18\x01 \x03(\x0b\x32\x13.qrl.MultiSigDetail\" \n\rGetBalanceReq\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\"!\n\x0eGetBalanceResp\x12\x0f\n\x07\x62\x61lance\x18\x01 \x01(\x04\"\'\n\x12GetTotalBalanceReq\x12\x11\n\taddresses\x18\x01 \x03(\x0c\"&\n\x13GetTotalBalanceResp\x12\x0f\n\x07\x62\x61lance\x18\x01 \x01(\x04\"b\n\tGetOTSReq\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\x12\x11\n\tpage_from\x18\x02 \x01(\x04\x12\x12\n\npage_count\x18\x03 \x01(\x04\x12\x1d\n\x15unused_ots_index_from\x18\x04 \x01(\x04\">\n\x11OTSBitfieldByPage\x12\x14\n\x0cots_bitfield\x18\x01 \x03(\x0c\x12\x13\n\x0bpage_number\x18\x02 \x01(\x04\"\x81\x01\n\nGetOTSResp\x12\x34\n\x14ots_bitfield_by_page\x18\x01 \x03(\x0b\x32\x16.qrl.OTSBitfieldByPage\x12\x1d\n\x15next_unused_ots_index\x18\x02 \x01(\x04\x12\x1e\n\x16unused_ots_index_found\x18\x03 \x01(\x08\"\x0e\n\x0cGetHeightReq\"\x1f\n\rGetHeightResp\x12\x0e\n\x06height\x18\x01 \x01(\x04\"\"\n\x0bGetBlockReq\x12\x13\n\x0bheader_hash\x18\x01 \x01(\x0c\")\n\x0cGetBlockResp\x12\x19\n\x05\x62lock\x18\x01 \x01(\x0b\x32\n.qrl.Block\"+\n\x13GetBlockByNumberReq\x12\x14\n\x0c\x62lock_number\x18\x01 \x01(\x04\"1\n\x14GetBlockByNumberResp\x12\x19\n\x05\x62lock\x18\x01 \x01(\x0b\x32\n.qrl.Block\"\x16\n\x14GetLocalAddressesReq\"*\n\x15GetLocalAddressesResp\x12\x11\n\taddresses\x18\x01 \x03(\x0c\"\x8d\x02\n\x08NodeInfo\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\"\n\x05state\x18\x02 \x01(\x0e\x32\x13.qrl.NodeInfo.State\x12\x17\n\x0fnum_connections\x18\x03 \x01(\r\x12\x17\n\x0fnum_known_peers\x18\x04 \x01(\r\x12\x0e\n\x06uptime\x18\x05 \x01(\x04\x12\x14\n\x0c\x62lock_height\x18\x06 \x01(\x04\x12\x17\n\x0f\x62lock_last_hash\x18\x07 \x01(\x0c\x12\x12\n\nnetwork_id\x18\x08 \x01(\t\"G\n\x05State\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0c\n\x08UNSYNCED\x10\x01\x12\x0b\n\x07SYNCING\x10\x02\x12\n\n\x06SYNCED\x10\x03\x12\n\n\x06\x46ORKED\x10\x04\"\x85\x01\n\x11\x41\x64\x64ressDescriptor\x12\x15\n\rhash_function\x18\x01 \x01(\t\x12\x18\n\x10signature_scheme\x18\x02 \x01(\t\x12\x13\n\x0btree_height\x18\x03 \x01(\r\x12\x12\n\nsignatures\x18\x04 \x01(\r\x12\x16\n\x0e\x61\x64\x64ress_format\x18\x05 \x01(\t\"\'\n\x0bStoredPeers\x12\x18\n\x05peers\x18\x01 \x03(\x0b\x32\t.qrl.Peer\"\x12\n\x04Peer\x12\n\n\x02ip\x18\x01 \x01(\t\"\x91\x03\n\x0c\x41\x64\x64ressState\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\x12\x0f\n\x07\x62\x61lance\x18\x02 \x01(\x04\x12\r\n\x05nonce\x18\x03 \x01(\x04\x12\x14\n\x0cots_bitfield\x18\x04 \x03(\x0c\x12\x1a\n\x12transaction_hashes\x18\x05 \x03(\x0c\x12-\n\x06tokens\x18\x06 \x03(\x0b\x32\x1d.qrl.AddressState.TokensEntry\x12&\n\x0elatticePK_list\x18\x07 \x03(\x0b\x32\x0e.qrl.LatticePK\x12H\n\x15slave_pks_access_type\x18\x08 \x03(\x0b\x32).qrl.AddressState.SlavePksAccessTypeEntry\x12\x13\n\x0bots_counter\x18\t \x01(\x04\x1a-\n\x0bTokensEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x04:\x02\x38\x01\x1a\x39\n\x17SlavePksAccessTypeEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\r:\x02\x38\x01\"\xe0\x03\n\x15OptimizedAddressState\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\x12\x0f\n\x07\x62\x61lance\x18\x02 \x01(\x04\x12\r\n\x05nonce\x18\x03 \x01(\x04\x12\x1e\n\x16ots_bitfield_used_page\x18\x04 \x01(\x04\x12\x1a\n\x12used_ots_key_count\x18\x05 \x01(\x04\x12\x1e\n\x16transaction_hash_count\x18\x06 \x01(\x04\x12\x14\n\x0ctokens_count\x18\x07 \x01(\x04\x12\x14\n\x0cslaves_count\x18\x08 \x01(\x04\x12\x18\n\x10lattice_pk_count\x18\t \x01(\x04\x12\x1f\n\x17multi_sig_address_count\x18\n \x01(\x04\x12\x1d\n\x15multi_sig_spend_count\x18\x0b \x01(\x04\x12\x1b\n\x13inbox_message_count\x18\x0c \x01(\x04\x12+\n#foundation_multi_sig_spend_txn_hash\x18\r \x03(\x0c\x12*\n\"foundation_multi_sig_vote_txn_hash\x18\x0e \x03(\x0c\x12\x0f\n\x07unvotes\x18\x0f \x03(\x0c\x12-\n\x13proposal_vote_stats\x18\x10 \x03(\x0b\x32\x10.qrl.Transaction\"\x93\x03\n\x14MultiSigAddressState\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\x12\x18\n\x10\x63reation_tx_hash\x18\x02 \x01(\x0c\x12\r\n\x05nonce\x18\x03 \x01(\x04\x12\x0f\n\x07\x62\x61lance\x18\x04 \x01(\x04\x12\x13\n\x0bsignatories\x18\x05 \x03(\x0c\x12\x0f\n\x07weights\x18\x06 \x03(\r\x12\x11\n\tthreshold\x18\x07 \x01(\r\x12\x1e\n\x16transaction_hash_count\x18\x08 \x01(\x04\x12\x1d\n\x15multi_sig_spend_count\x18\t \x01(\x04\x12\x1f\n\x17multi_sig_address_count\x18\n \x01(\x04\x12+\n#foundation_multi_sig_spend_txn_hash\x18\x0b \x03(\x0c\x12*\n\"foundation_multi_sig_vote_txn_hash\x18\x0c \x03(\x0c\x12\x0f\n\x07unvotes\x18\r \x03(\x0c\x12-\n\x13proposal_vote_stats\x18\x0e \x03(\x0b\x32\x10.qrl.Transaction\"\'\n\x15MultiSigAddressesList\x12\x0e\n\x06hashes\x18\x01 \x03(\x0c\"\x1a\n\x08\x44\x61taList\x12\x0e\n\x06values\x18\x01 \x03(\x0c\"\x1d\n\x08\x42itfield\x12\x11\n\tbitfields\x18\x01 \x03(\x0c\"%\n\x13TransactionHashList\x12\x0e\n\x06hashes\x18\x01 \x03(\x0c\"3\n\tLatticePK\x12\x10\n\x08kyber_pk\x18\x01 \x01(\x0c\x12\x14\n\x0c\x64ilithium_pk\x18\x02 \x01(\x0c\"0\n\rAddressAmount\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\x12\x0e\n\x06\x61mount\x18\x02 \x01(\x04\"\xd7\x01\n\x0b\x42lockHeader\x12\x13\n\x0bhash_header\x18\x01 \x01(\x0c\x12\x14\n\x0c\x62lock_number\x18\x02 \x01(\x04\x12\x19\n\x11timestamp_seconds\x18\x03 \x01(\x04\x12\x18\n\x10hash_header_prev\x18\x04 \x01(\x0c\x12\x14\n\x0creward_block\x18\x05 \x01(\x04\x12\x12\n\nreward_fee\x18\x06 \x01(\x04\x12\x13\n\x0bmerkle_root\x18\x07 \x01(\x0c\x12\x14\n\x0cmining_nonce\x18\x08 \x01(\r\x12\x13\n\x0b\x65xtra_nonce\x18\t \x01(\x04\"i\n\x13\x42lockHeaderExtended\x12 \n\x06header\x18\x01 \x01(\x0b\x32\x10.qrl.BlockHeader\x12\x30\n\x11transaction_count\x18\x02 \x01(\x0b\x32\x15.qrl.TransactionCount\"q\n\x10TransactionCount\x12/\n\x05\x63ount\x18\x01 \x03(\x0b\x32 .qrl.TransactionCount.CountEntry\x1a,\n\nCountEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\r:\x02\x38\x01\"\x91\x01\n\x13TransactionExtended\x12 \n\x06header\x18\x01 \x01(\x0b\x32\x10.qrl.BlockHeader\x12\x1c\n\x02tx\x18\x02 \x01(\x0b\x32\x10.qrl.Transaction\x12\x11\n\taddr_from\x18\x03 \x01(\x0c\x12\x0c\n\x04size\x18\x04 \x01(\x04\x12\x19\n\x11timestamp_seconds\x18\x05 \x01(\x04\"\xa6\x01\n\rBlockExtended\x12 \n\x06header\x18\x01 \x01(\x0b\x32\x10.qrl.BlockHeader\x12\x37\n\x15\x65xtended_transactions\x18\x02 \x03(\x0b\x32\x18.qrl.TransactionExtended\x12,\n\x0fgenesis_balance\x18\x03 \x03(\x0b\x32\x13.qrl.GenesisBalance\x12\x0c\n\x04size\x18\x04 \x01(\x04\"\x7f\n\x05\x42lock\x12 \n\x06header\x18\x01 \x01(\x0b\x32\x10.qrl.BlockHeader\x12&\n\x0ctransactions\x18\x02 \x03(\x0b\x32\x10.qrl.Transaction\x12,\n\x0fgenesis_balance\x18\x03 \x03(\x0b\x32\x13.qrl.GenesisBalance\"2\n\x0eGenesisBalance\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\x12\x0f\n\x07\x62\x61lance\x18\x02 \x01(\x04\"D\n\x11\x42lockMetaDataList\x12/\n\x13\x62lock_number_hashes\x18\x01 \x03(\x0b\x32\x12.qrl.BlockMetaData\"\xaa\x15\n\x0bTransaction\x12\x13\n\x0bmaster_addr\x18\x01 \x01(\x0c\x12\x0b\n\x03\x66\x65\x65\x18\x02 \x01(\x04\x12\x12\n\npublic_key\x18\x03 \x01(\x0c\x12\x11\n\tsignature\x18\x04 \x01(\x0c\x12\r\n\x05nonce\x18\x05 \x01(\x04\x12\x18\n\x10transaction_hash\x18\x06 \x01(\x0c\x12-\n\x08transfer\x18\x07 \x01(\x0b\x32\x19.qrl.Transaction.TransferH\x00\x12-\n\x08\x63oinbase\x18\x08 \x01(\x0b\x32\x19.qrl.Transaction.CoinBaseH\x00\x12\x36\n\tlatticePK\x18\t \x01(\x0b\x32!.qrl.Transaction.LatticePublicKeyH\x00\x12+\n\x07message\x18\n \x01(\x0b\x32\x18.qrl.Transaction.MessageH\x00\x12\'\n\x05token\x18\x0b \x01(\x0b\x32\x16.qrl.Transaction.TokenH\x00\x12\x38\n\x0etransfer_token\x18\x0c \x01(\x0b\x32\x1e.qrl.Transaction.TransferTokenH\x00\x12\'\n\x05slave\x18\r \x01(\x0b\x32\x16.qrl.Transaction.SlaveH\x00\x12;\n\x10multi_sig_create\x18\x0e \x01(\x0b\x32\x1f.qrl.Transaction.MultiSigCreateH\x00\x12\x39\n\x0fmulti_sig_spend\x18\x0f \x01(\x0b\x32\x1e.qrl.Transaction.MultiSigSpendH\x00\x12\x37\n\x0emulti_sig_vote\x18\x10 \x01(\x0b\x32\x1d.qrl.Transaction.MultiSigVoteH\x00\x12:\n\x0fproposal_create\x18\x11 \x01(\x0b\x32\x1f.qrl.Transaction.ProposalCreateH\x00\x12\x36\n\rproposal_vote\x18\x12 \x01(\x0b\x32\x1d.qrl.Transaction.ProposalVoteH\x00\x1a\x43\n\x08Transfer\x12\x10\n\x08\x61\x64\x64rs_to\x18\x01 \x03(\x0c\x12\x0f\n\x07\x61mounts\x18\x02 \x03(\x04\x12\x14\n\x0cmessage_data\x18\x03 \x01(\x0c\x1a+\n\x08\x43oinBase\x12\x0f\n\x07\x61\x64\x64r_to\x18\x01 \x01(\x0c\x12\x0e\n\x06\x61mount\x18\x02 \x01(\x04\x1a\x39\n\x10LatticePublicKey\x12\x0b\n\x03pk1\x18\x01 \x01(\x0c\x12\x0b\n\x03pk2\x18\x02 \x01(\x0c\x12\x0b\n\x03pk3\x18\x03 \x01(\x0c\x1a\x30\n\x07Message\x12\x14\n\x0cmessage_hash\x18\x01 \x01(\x0c\x12\x0f\n\x07\x61\x64\x64r_to\x18\x02 \x01(\x0c\x1at\n\x05Token\x12\x0e\n\x06symbol\x18\x01 \x01(\x0c\x12\x0c\n\x04name\x18\x02 \x01(\x0c\x12\r\n\x05owner\x18\x03 \x01(\x0c\x12\x10\n\x08\x64\x65\x63imals\x18\x04 \x01(\x04\x12,\n\x10initial_balances\x18\x05 \x03(\x0b\x32\x12.qrl.AddressAmount\x1aH\n\rTransferToken\x12\x14\n\x0ctoken_txhash\x18\x01 \x01(\x0c\x12\x10\n\x08\x61\x64\x64rs_to\x18\x02 \x03(\x0c\x12\x0f\n\x07\x61mounts\x18\x03 \x03(\x04\x1a\x30\n\x05Slave\x12\x11\n\tslave_pks\x18\x01 \x03(\x0c\x12\x14\n\x0c\x61\x63\x63\x65ss_types\x18\x02 \x03(\r\x1aI\n\x0eMultiSigCreate\x12\x13\n\x0bsignatories\x18\x01 \x03(\x0c\x12\x0f\n\x07weights\x18\x02 \x03(\r\x12\x11\n\tthreshold\x18\x03 \x01(\r\x1aj\n\rMultiSigSpend\x12\x19\n\x11multi_sig_address\x18\x01 \x01(\x0c\x12\x10\n\x08\x61\x64\x64rs_to\x18\x02 \x03(\x0c\x12\x0f\n\x07\x61mounts\x18\x03 \x03(\x04\x12\x1b\n\x13\x65xpiry_block_number\x18\x04 \x01(\x04\x1aH\n\x0cMultiSigVote\x12\x12\n\nshared_key\x18\x01 \x01(\x0c\x12\x0e\n\x06unvote\x18\x02 \x01(\x08\x12\x14\n\x0cprev_tx_hash\x18\x03 \x01(\x0c\x1a\x9b\t\n\x0eProposalCreate\x12\x1b\n\x13\x65xpiry_block_number\x18\x01 \x01(\x04\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x32\n\x03qip\x18\x03 \x01(\x0b\x32#.qrl.Transaction.ProposalCreate.QIPH\x00\x12\x38\n\x06\x63onfig\x18\x04 \x01(\x0b\x32&.qrl.Transaction.ProposalCreate.ConfigH\x00\x12\x36\n\x05other\x18\x05 \x01(\x0b\x32%.qrl.Transaction.ProposalCreate.OtherH\x00\x1a\x17\n\x03QIP\x12\x10\n\x08qip_link\x18\x01 \x01(\t\x1a\xed\x06\n\x06\x43onfig\x12\x18\n\x10\x63hanges_bitfield\x18\x01 \x03(\x0c\x12\x13\n\x0breorg_limit\x18\x02 \x01(\x04\x12\x17\n\x0fmax_coin_supply\x18\x03 \x01(\x04\x12,\n$complete_emission_time_span_in_years\x18\x04 \x01(\x04\x12\x1b\n\x13mining_nonce_offset\x18\x05 \x01(\x04\x12\x1a\n\x12\x65xtra_nonce_offset\x18\x06 \x01(\x04\x12!\n\x19mining_blob_size_in_bytes\x18\x07 \x01(\x04\x12\x1f\n\x17\x62lock_timing_in_seconds\x18\x08 \x01(\x04\x12 \n\x18number_of_blocks_analyze\x18\t \x01(\x04\x12\x1d\n\x15\x62lock_size_multiplier\x18\n \x01(\x04\x12%\n\x1d\x62lock_min_size_limit_in_bytes\x18\x0b \x01(\x04\x12&\n\x1etransaction_multi_output_limit\x18\x0c \x01(\x04\x12\x1a\n\x12message_max_length\x18\r \x01(\x04\x12\x1f\n\x17token_symbol_max_length\x18\x0e \x01(\x04\x12\x1d\n\x15token_name_max_length\x18\x0f \x01(\x04\x12\x1e\n\x16lattice_pk1_max_length\x18\x10 \x01(\x04\x12\x1e\n\x16lattice_pk2_max_length\x18\x11 \x01(\x04\x12\x1e\n\x16lattice_pk3_max_length\x18\x12 \x01(\x04\x12\x39\n1foundation_multi_sig_address_threshold_percentage\x18\x13 \x01(\x04\x12\x1e\n\x16proposal_threshold_per\x18\x14 \x01(\x04\x12 \n\x18proposal_default_options\x18\x15 \x03(\t\x12\x1e\n\x16\x64\x65scription_max_length\x18\x16 \x01(\x04\x12\x1a\n\x12options_max_number\x18\x17 \x01(\x04\x12\x1e\n\x16option_max_text_length\x18\x18 \x01(\x04\x12(\n proposal_config_activation_delay\x18\x19 \x01(\x04\x12\x15\n\rN_measurement\x18\x1a \x01(\x04\x12\n\n\x02kp\x18\x1b \x01(\x04\x1a\x18\n\x05Other\x12\x0f\n\x07options\x18\x01 \x03(\tB\x0e\n\x0cproposalType\x1a\x32\n\x0cProposalVote\x12\x12\n\nshared_key\x18\x01 \x01(\x0c\x12\x0e\n\x06option\x18\x02 \x01(\rB\x11\n\x0ftransactionType\"\xb8\x01\n\tVoteStats\x12\x19\n\x11multi_sig_address\x18\x01 \x01(\x0c\x12\x12\n\nshared_key\x18\x02 \x01(\x0c\x12\x13\n\x0bsignatories\x18\x03 \x03(\x0c\x12\x11\n\ttx_hashes\x18\x04 \x03(\x0c\x12\x0f\n\x07unvotes\x18\x05 \x03(\x08\x12\x1b\n\x13\x65xpiry_block_number\x18\x06 \x01(\x04\x12\x14\n\x0ctotal_weight\x18\x07 \x01(\x04\x12\x10\n\x08\x65xecuted\x18\x08 \x01(\x08\"\xb7\x01\n\x11ProposalVoteStats\x12\x11\n\taddr_from\x18\x01 \x01(\x0c\x12\x12\n\nshared_key\x18\x02 \x01(\x0c\x12\x15\n\rproposal_type\x18\x03 \x01(\t\x12\x18\n\x10weight_by_option\x18\x04 \x03(\x04\x12\x1b\n\x13\x65xpiry_block_number\x18\x05 \x01(\x04\x12\x10\n\x08\x65xecuted\x18\x06 \x01(\x08\x12\x1b\n\x13number_of_tx_hashes\x18\x07 \x01(\x04\"-\n\x0eProposalRecord\x12\x1b\n\x13number_of_tx_hashes\x18\x01 \x01(\x04\"!\n\tTokenList\x12\x14\n\x0ctoken_txhash\x18\x01 \x03(\x0c\"R\n\x0cTokenBalance\x12\x0f\n\x07\x62\x61lance\x18\x01 \x01(\x04\x12\x10\n\x08\x64\x65\x63imals\x18\x02 \x01(\x04\x12\x0f\n\x07tx_hash\x18\x03 \x01(\x0c\x12\x0e\n\x06\x64\x65lete\x18\x04 \x01(\x08\"E\n\rSlaveMetadata\x12\x13\n\x0b\x61\x63\x63\x65ss_type\x18\x01 \x01(\x04\x12\x0f\n\x07tx_hash\x18\x02 \x01(\x0c\x12\x0e\n\x06\x64\x65lete\x18\x03 \x01(\x08\"E\n\x11LatticePKMetadata\x12\x0f\n\x07\x65nabled\x18\x01 \x01(\x08\x12\x0f\n\x07tx_hash\x18\x02 \x01(\x0c\x12\x0e\n\x06\x64\x65lete\x18\x03 \x01(\x08\"G\n\rTokenMetadata\x12\x14\n\x0ctoken_txhash\x18\x01 \x01(\x0c\x12 \n\x18transfer_token_tx_hashes\x18\x02 \x03(\x0c\"\xc4\x01\n\x19\x45ncryptedEphemeralMessage\x12\x0e\n\x06msg_id\x18\x01 \x01(\x0c\x12\x0b\n\x03ttl\x18\x02 \x01(\x04\x12\x0b\n\x03ttr\x18\x03 \x01(\x04\x12\x37\n\x07\x63hannel\x18\x05 \x01(\x0b\x32&.qrl.EncryptedEphemeralMessage.Channel\x12\r\n\x05nonce\x18\x06 \x01(\x04\x12\x0f\n\x07payload\x18\x07 \x01(\x0c\x1a$\n\x07\x43hannel\x12\x19\n\x11\x65nc_aes256_symkey\x18\x04 \x01(\x0c\" \n\x0b\x41\x64\x64ressList\x12\x11\n\taddresses\x18\x01 \x03(\x0c\"`\n\x0f\x42lockHeightData\x12\x14\n\x0c\x62lock_number\x18\x01 \x01(\x04\x12\x18\n\x10\x62lock_headerhash\x18\x02 \x01(\x0c\x12\x1d\n\x15\x63umulative_difficulty\x18\x03 \x01(\x0c\"\x81\x01\n\rBlockMetaData\x12\x18\n\x10\x62lock_difficulty\x18\x01 \x01(\x0c\x12\x1d\n\x15\x63umulative_difficulty\x18\x02 \x01(\x0c\x12\x1a\n\x12\x63hild_headerhashes\x18\x03 \x03(\x0c\x12\x1b\n\x13last_N_headerhashes\x18\x04 \x03(\x0c\"A\n\x12\x42lockNumberMapping\x12\x12\n\nheaderhash\x18\x01 \x01(\x0c\x12\x17\n\x0fprev_headerhash\x18\x02 \x01(\x0c\"X\n\x08PeerStat\x12\x0f\n\x07peer_ip\x18\x01 \x01(\x0c\x12\x0c\n\x04port\x18\x02 \x01(\r\x12-\n\x10node_chain_state\x18\x03 \x01(\x0b\x32\x13.qrl.NodeChainState\"~\n\x0eNodeChainState\x12\x14\n\x0c\x62lock_number\x18\x01 \x01(\x04\x12\x13\n\x0bheader_hash\x18\x02 \x01(\x0c\x12\x1d\n\x15\x63umulative_difficulty\x18\x03 \x01(\x0c\x12\x0f\n\x07version\x18\x04 \x01(\t\x12\x11\n\ttimestamp\x18\x05 \x01(\x04\"<\n\x0eNodeHeaderHash\x12\x14\n\x0c\x62lock_number\x18\x01 \x01(\x04\x12\x14\n\x0cheaderhashes\x18\x02 \x03(\x0c\"-\n\x12P2PAcknowledgement\x12\x17\n\x0f\x62ytes_processed\x18\x01 \x01(\r\"|\n\x08PeerInfo\x12\x0f\n\x07peer_ip\x18\x01 \x01(\x0c\x12\x0c\n\x04port\x18\x02 \x01(\r\x12\x18\n\x10\x62\x61nned_timestamp\x18\x03 \x01(\r\x12\x13\n\x0b\x63redibility\x18\x04 \x01(\r\x12\"\n\x1alast_connections_timestamp\x18\x05 \x03(\r\".\n\x05Peers\x12%\n\x0epeer_info_list\x18\x01 \x03(\x0b\x32\r.qrl.PeerInfo\"\xbb\x0c\n\tDevConfig\x12\x16\n\x0eprev_state_key\x18\x01 \x01(\x0c\x12\x19\n\x11\x63urrent_state_key\x18\x02 \x01(\x0c\x12\x1e\n\x16\x61\x63tivation_header_hash\x18\x03 \x01(\x0c\x12\x1f\n\x17\x61\x63tivation_block_number\x18\x04 \x01(\x04\x12#\n\x05\x63hain\x18\x05 \x01(\x0b\x32\x14.qrl.DevConfig.Chain\x12#\n\x05\x62lock\x18\x06 \x01(\x0b\x32\x14.qrl.DevConfig.Block\x12/\n\x0btransaction\x18\x07 \x01(\x0b\x32\x1a.qrl.DevConfig.Transaction\x12\x1f\n\x03pow\x18\x08 \x01(\x0b\x32\x12.qrl.DevConfig.POW\x1a\x63\n\x05\x43hain\x12\x13\n\x0breorg_limit\x18\x01 \x01(\x04\x12\x17\n\x0fmax_coin_supply\x18\x02 \x01(\x04\x12,\n$complete_emission_time_span_in_years\x18\x03 \x01(\x04\x1a\xc6\x02\n\x05\x42lock\x12\x1b\n\x13mining_nonce_offset\x18\x01 \x01(\x04\x12\x1a\n\x12\x65xtra_nonce_offset\x18\x02 \x01(\x04\x12!\n\x19mining_blob_size_in_bytes\x18\x03 \x01(\x04\x12\x1f\n\x17\x62lock_timing_in_seconds\x18\x04 \x01(\x04\x12G\n\x15\x62lock_size_controller\x18\x05 \x01(\x0b\x32(.qrl.DevConfig.Block.BlockSizeController\x1aw\n\x13\x42lockSizeController\x12 \n\x18number_of_blocks_analyze\x18\x01 \x01(\x04\x12\x17\n\x0fsize_multiplier\x18\x02 \x01(\x04\x12%\n\x1d\x62lock_min_size_limit_in_bytes\x18\x03 \x01(\x04\x1a\xc5\x06\n\x0bTransaction\x12\x1a\n\x12multi_output_limit\x18\x01 \x01(\x04\x12\x33\n\x07message\x18\x02 \x01(\x0b\x32\".qrl.DevConfig.Transaction.Message\x12/\n\x05slave\x18\x03 \x01(\x0b\x32 .qrl.DevConfig.Transaction.Slave\x12/\n\x05token\x18\x04 \x01(\x0b\x32 .qrl.DevConfig.Transaction.Token\x12\x33\n\x07lattice\x18\x05 \x01(\x0b\x32\".qrl.DevConfig.Transaction.Lattice\x12K\n\x14\x66oundation_multi_sig\x18\x06 \x01(\x0b\x32-.qrl.DevConfig.Transaction.FoundationMultiSig\x12\x35\n\x08proposal\x18\x07 \x01(\x0b\x32#.qrl.DevConfig.Transaction.Proposal\x1a\x1d\n\x07Message\x12\x12\n\nmax_length\x18\x01 \x01(\x04\x1a$\n\x05Slave\x12\x1b\n\x13slave_pk_max_length\x18\x02 \x01(\x04\x1a;\n\x05Token\x12\x19\n\x11symbol_max_length\x18\x01 \x01(\x04\x12\x17\n\x0fname_max_length\x18\x02 \x01(\x04\x1aQ\n\x07Lattice\x12\x16\n\x0epk1_max_length\x18\x01 \x01(\x04\x12\x16\n\x0epk2_max_length\x18\x02 \x01(\x04\x12\x16\n\x0epk3_max_length\x18\x03 \x01(\x04\x1a\x32\n\x12\x46oundationMultiSig\x12\x1c\n\x14threshold_percentage\x18\x01 \x01(\x04\x1a\xc0\x01\n\x08Proposal\x12\x15\n\rthreshold_per\x18\x01 \x01(\x04\x12\x17\n\x0f\x64\x65\x66\x61ult_options\x18\x02 \x03(\t\x12\x1e\n\x16\x64\x65scription_max_length\x18\x03 \x01(\x04\x12\x1a\n\x12options_max_number\x18\x04 \x01(\x04\x12\x1e\n\x16option_max_text_length\x18\x05 \x01(\x04\x12(\n proposal_config_activation_delay\x18\x06 \x01(\x04\x1a(\n\x03POW\x12\x15\n\rN_measurement\x18\x01 \x01(\x04\x12\n\n\x02kp\x18\x02 \x01(\x04\x32\xbd\x1d\n\tPublicAPI\x12P\n\x0cGetNodeState\x12\x14.qrl.GetNodeStateReq\x1a\x15.qrl.GetNodeStateResp\"\x13\x82\xd3\xe4\x93\x02\r\x12\x0b/node-state\x12T\n\rGetKnownPeers\x12\x15.qrl.GetKnownPeersReq\x1a\x16.qrl.GetKnownPeersResp\"\x14\x82\xd3\xe4\x93\x02\x0e\x12\x0c/known-peers\x12P\n\x0cGetPeersStat\x12\x14.qrl.GetPeersStatReq\x1a\x15.qrl.GetPeersStatResp\"\x13\x82\xd3\xe4\x93\x02\r\x12\x0b/peers-stat\x12?\n\x08GetStats\x12\x10.qrl.GetStatsReq\x1a\x11.qrl.GetStatsResp\"\x0e\x82\xd3\xe4\x93\x02\x08\x12\x06/stats\x12\\\n\x0fGetAddressState\x12\x17.qrl.GetAddressStateReq\x1a\x18.qrl.GetAddressStateResp\"\x16\x82\xd3\xe4\x93\x02\x10\x12\x0e/address-state\x12x\n\x18GetOptimizedAddressState\x12\x17.qrl.GetAddressStateReq\x1a!.qrl.GetOptimizedAddressStateResp\" \x82\xd3\xe4\x93\x02\x1a\x12\x18/optimized-address-state\x12~\n\x17GetMultiSigAddressState\x12\x1f.qrl.GetMultiSigAddressStateReq\x1a .qrl.GetMultiSigAddressStateResp\" \x82\xd3\xe4\x93\x02\x1a\x12\x18/multi-sig-address-state\x12?\n\x07IsSlave\x12\x0f.qrl.IsSlaveReq\x1a\x10.qrl.IsSlaveResp\"\x11\x82\xd3\xe4\x93\x02\x0b\x12\t/is-slave\x12\x43\n\tGetObject\x12\x11.qrl.GetObjectReq\x1a\x12.qrl.GetObjectResp\"\x0f\x82\xd3\xe4\x93\x02\t\x12\x07/object\x12T\n\rGetLatestData\x12\x15.qrl.GetLatestDataReq\x1a\x16.qrl.GetLatestDataResp\"\x14\x82\xd3\xe4\x93\x02\x0e\x12\x0c/latest-data\x12_\n\x0fPushTransaction\x12\x17.qrl.PushTransactionReq\x1a\x18.qrl.PushTransactionResp\"\x19\x82\xd3\xe4\x93\x02\x13\"\x11/push-transaction\x12W\n\rTransferCoins\x12\x15.qrl.TransferCoinsReq\x1a\x16.qrl.TransferCoinsResp\"\x17\x82\xd3\xe4\x93\x02\x11\"\x0f/transfer-coins\x12S\n\x0cParseAddress\x12\x14.qrl.ParseAddressReq\x1a\x15.qrl.ParseAddressResp\"\x16\x82\xd3\xe4\x93\x02\x10\x12\x0e/parse-address\x12T\n\rGetChainStats\x12\x15.qrl.GetChainStatsReq\x1a\x16.qrl.GetChainStatsResp\"\x14\x82\xd3\xe4\x93\x02\x0e\x12\x0c/chain-stats\x12\x61\n\x10GetAddressFromPK\x12\x18.qrl.GetAddressFromPKReq\x1a\x19.qrl.GetAddressFromPKResp\"\x18\x82\xd3\xe4\x93\x02\x12\x12\x10/address-from-pk\x12h\n\x14GetMultiSigCreateTxn\x12\x19.qrl.MultiSigCreateTxnReq\x1a\x16.qrl.TransferCoinsResp\"\x1d\x82\xd3\xe4\x93\x02\x17\"\x15/multi-sig-create-txn\x12\x65\n\x13GetMultiSigSpendTxn\x12\x18.qrl.MultiSigSpendTxnReq\x1a\x16.qrl.TransferCoinsResp\"\x1c\x82\xd3\xe4\x93\x02\x16\"\x14/multi-sig-spend-txn\x12\x62\n\x12GetMultiSigVoteTxn\x12\x17.qrl.MultiSigVoteTxnReq\x1a\x16.qrl.TransferCoinsResp\"\x1b\x82\xd3\xe4\x93\x02\x15\"\x13/multi-sig-vote-txn\x12Q\n\rGetMessageTxn\x12\x12.qrl.MessageTxnReq\x1a\x16.qrl.TransferCoinsResp\"\x14\x82\xd3\xe4\x93\x02\x0e\"\x0c/message-txn\x12K\n\x0bGetTokenTxn\x12\x10.qrl.TokenTxnReq\x1a\x16.qrl.TransferCoinsResp\"\x12\x82\xd3\xe4\x93\x02\x0c\"\n/token-txn\x12\x64\n\x13GetTransferTokenTxn\x12\x18.qrl.TransferTokenTxnReq\x1a\x16.qrl.TransferCoinsResp\"\x1b\x82\xd3\xe4\x93\x02\x15\"\x13/transfer-token-txn\x12K\n\x0bGetSlaveTxn\x12\x10.qrl.SlaveTxnReq\x1a\x16.qrl.TransferCoinsResp\"\x12\x82\xd3\xe4\x93\x02\x0c\"\n/slave-txn\x12Q\n\rGetLatticeTxn\x12\x12.qrl.LatticeTxnReq\x1a\x16.qrl.TransferCoinsResp\"\x14\x82\xd3\xe4\x93\x02\x0e\"\x0c/lattice-txn\x12W\n\x0eGetTransaction\x12\x16.qrl.GetTransactionReq\x1a\x17.qrl.GetTransactionResp\"\x14\x82\xd3\xe4\x93\x02\x0e\x12\x0c/transaction\x12\x91\x01\n\x1cGetMiniTransactionsByAddress\x12$.qrl.GetMiniTransactionsByAddressReq\x1a%.qrl.GetMiniTransactionsByAddressResp\"$\x82\xd3\xe4\x93\x02\x1e\x12\x1c/mini-transaction-by-address\x12\x81\x01\n\x18GetTransactionsByAddress\x12 .qrl.GetTransactionsByAddressReq\x1a!.qrl.GetTransactionsByAddressResp\" \x82\xd3\xe4\x93\x02\x1a\x12\x18/transactions-by-address\x12o\n\x12GetTokensByAddress\x12 .qrl.GetTransactionsByAddressReq\x1a\x1b.qrl.GetTokensByAddressResp\"\x1a\x82\xd3\xe4\x93\x02\x14\x12\x12/tokens-by-address\x12o\n\x12GetSlavesByAddress\x12 .qrl.GetTransactionsByAddressReq\x1a\x1b.qrl.GetSlavesByAddressResp\"\x1a\x82\xd3\xe4\x93\x02\x14\x12\x12/slaves-by-address\x12|\n\x16GetLatticePKsByAddress\x12 .qrl.GetTransactionsByAddressReq\x1a\x1f.qrl.GetLatticePKsByAddressResp\"\x1f\x82\xd3\xe4\x93\x02\x19\x12\x17/lattice-pks-by-address\x12\x92\x01\n\x1dGetMultiSigAddressesByAddress\x12 .qrl.GetTransactionsByAddressReq\x1a&.qrl.GetMultiSigAddressesByAddressResp\"\'\x82\xd3\xe4\x93\x02!\x12\x1f/multi-sig-addresses-by-address\x12\x94\x01\n\x1cGetMultiSigSpendTxsByAddress\x12$.qrl.GetMultiSigSpendTxsByAddressReq\x1a%.qrl.GetMultiSigSpendTxsByAddressResp\"\'\x82\xd3\xe4\x93\x02!\x12\x1f/multi-sig-spend-txs-by-address\x12P\n\x0cGetVoteStats\x12\x14.qrl.GetVoteStatsReq\x1a\x15.qrl.GetVoteStatsResp\"\x13\x82\xd3\xe4\x93\x02\r\x12\x0b/vote-stats\x12\x85\x01\n\x19GetInboxMessagesByAddress\x12 .qrl.GetTransactionsByAddressReq\x1a\".qrl.GetInboxMessagesByAddressResp\"\"\x82\xd3\xe4\x93\x02\x1c\x12\x1a/inbox-messages-by-address\x12G\n\nGetBalance\x12\x12.qrl.GetBalanceReq\x1a\x13.qrl.GetBalanceResp\"\x10\x82\xd3\xe4\x93\x02\n\x12\x08/balance\x12\\\n\x0fGetTotalBalance\x12\x17.qrl.GetTotalBalanceReq\x1a\x18.qrl.GetTotalBalanceResp\"\x16\x82\xd3\xe4\x93\x02\x10\x12\x0e/total-balance\x12\x37\n\x06GetOTS\x12\x0e.qrl.GetOTSReq\x1a\x0f.qrl.GetOTSResp\"\x0c\x82\xd3\xe4\x93\x02\x06\x12\x04/ots\x12\x43\n\tGetHeight\x12\x11.qrl.GetHeightReq\x1a\x12.qrl.GetHeightResp\"\x0f\x82\xd3\xe4\x93\x02\t\x12\x07/height\x12?\n\x08GetBlock\x12\x10.qrl.GetBlockReq\x1a\x11.qrl.GetBlockResp\"\x0e\x82\xd3\xe4\x93\x02\x08\x12\x06/block\x12\x61\n\x10GetBlockByNumber\x12\x18.qrl.GetBlockByNumberReq\x1a\x19.qrl.GetBlockByNumberResp\"\x18\x82\xd3\xe4\x93\x02\x12\x12\x10/block-by-number2\n\n\x08\x41\x64minAPIb\x06proto3')
  ,
  dependencies=[google_dot_api_dot_annotations__pb2.DESCRIPTOR,])



_GETLATESTDATAREQ_FILTER = _descriptor.EnumDescriptor(
  name='Filter',
  full_name='qrl.GetLatestDataReq.Filter',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ALL', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BLOCKHEADERS', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TRANSACTIONS', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TRANSACTIONS_UNCONFIRMED', index=3, number=3,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1855,
  serialized_end=1938,
)
_sym_db.RegisterEnumDescriptor(_GETLATESTDATAREQ_FILTER)

_PUSHTRANSACTIONRESP_RESPONSECODE = _descriptor.EnumDescriptor(
  name='ResponseCode',
  full_name='qrl.PushTransactionResp.ResponseCode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ERROR', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='VALIDATION_FAILED', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SUBMITTED', index=3, number=3,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=2532,
  serialized_end=2608,
)
_sym_db.RegisterEnumDescriptor(_PUSHTRANSACTIONRESP_RESPONSECODE)

_GETMULTISIGSPENDTXSBYADDRESSREQ_FILTERTYPE = _descriptor.EnumDescriptor(
  name='FilterType',
  full_name='qrl.GetMultiSigSpendTxsByAddressReq.FilterType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NONE', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXECUTED_ONLY', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NON_EXECUTED', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXPIRED', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NON_EXPIRED', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NON_EXECUTED_EXPIRED', index=5, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NON_EXECUTED_NON_EXPIRED', index=6, number=6,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=4465,
  serialized_end=4610,
)
_sym_db.RegisterEnumDescriptor(_GETMULTISIGSPENDTXSBYADDRESSREQ_FILTERTYPE)

_NODEINFO_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='qrl.NodeInfo.State',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNSYNCED', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SYNCING', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SYNCED', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FORKED', index=4, number=4,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=6514,
  serialized_end=6585,
)
_sym_db.RegisterEnumDescriptor(_NODEINFO_STATE)


_EMPTY = _descriptor.Descriptor(
  name='Empty',
  full_name='qrl.Empty',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=48,
  serialized_end=55,
)


_GETNODESTATEREQ = _descriptor.Descriptor(
  name='GetNodeStateReq',
  full_name='qrl.GetNodeStateReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=57,
  serialized_end=74,
)


_GETNODESTATERESP = _descriptor.Descriptor(
  name='GetNodeStateResp',
  full_name='qrl.GetNodeStateResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='info', full_name='qrl.GetNodeStateResp.info', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=76,
  serialized_end=123,
)


_GETKNOWNPEERSREQ = _descriptor.Descriptor(
  name='GetKnownPeersReq',
  full_name='qrl.GetKnownPeersReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=125,
  serialized_end=143,
)


_GETKNOWNPEERSRESP = _descriptor.Descriptor(
  name='GetKnownPeersResp',
  full_name='qrl.GetKnownPeersResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='node_info', full_name='qrl.GetKnownPeersResp.node_info', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='known_peers', full_name='qrl.GetKnownPeersResp.known_peers', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=145,
  serialized_end=230,
)


_GETPEERSSTATREQ = _descriptor.Descriptor(
  name='GetPeersStatReq',
  full_name='qrl.GetPeersStatReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=232,
  serialized_end=249,
)


_GETPEERSSTATRESP = _descriptor.Descriptor(
  name='GetPeersStatResp',
  full_name='qrl.GetPeersStatResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='peers_stat', full_name='qrl.GetPeersStatResp.peers_stat', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=251,
  serialized_end=304,
)


_GETCHAINSTATSREQ = _descriptor.Descriptor(
  name='GetChainStatsReq',
  full_name='qrl.GetChainStatsReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=306,
  serialized_end=324,
)


_GETCHAINSTATSRESP = _descriptor.Descriptor(
  name='GetChainStatsResp',
  full_name='qrl.GetChainStatsResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='state_size', full_name='qrl.GetChainStatsResp.state_size', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='state_size_mb', full_name='qrl.GetChainStatsResp.state_size_mb', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='state_size_gb', full_name='qrl.GetChainStatsResp.state_size_gb', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=326,
  serialized_end=411,
)


_GETSTATSREQ = _descriptor.Descriptor(
  name='GetStatsReq',
  full_name='qrl.GetStatsReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='include_timeseries', full_name='qrl.GetStatsReq.include_timeseries', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=413,
  serialized_end=454,
)


_GETSTATSRESP = _descriptor.Descriptor(
  name='GetStatsResp',
  full_name='qrl.GetStatsResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='node_info', full_name='qrl.GetStatsResp.node_info', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='epoch', full_name='qrl.GetStatsResp.epoch', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='uptime_network', full_name='qrl.GetStatsResp.uptime_network', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_last_reward', full_name='qrl.GetStatsResp.block_last_reward', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_time_mean', full_name='qrl.GetStatsResp.block_time_mean', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_time_sd', full_name='qrl.GetStatsResp.block_time_sd', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='coins_total_supply', full_name='qrl.GetStatsResp.coins_total_supply', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='coins_emitted', full_name='qrl.GetStatsResp.coins_emitted', index=7,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_timeseries', full_name='qrl.GetStatsResp.block_timeseries', index=8,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=457,
  serialized_end=717,
)


_GETADDRESSFROMPKREQ = _descriptor.Descriptor(
  name='GetAddressFromPKReq',
  full_name='qrl.GetAddressFromPKReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pk', full_name='qrl.GetAddressFromPKReq.pk', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=719,
  serialized_end=752,
)


_GETADDRESSFROMPKRESP = _descriptor.Descriptor(
  name='GetAddressFromPKResp',
  full_name='qrl.GetAddressFromPKResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='qrl.GetAddressFromPKResp.address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=754,
  serialized_end=793,
)


_BLOCKDATAPOINT = _descriptor.Descriptor(
  name='BlockDataPoint',
  full_name='qrl.BlockDataPoint',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='number', full_name='qrl.BlockDataPoint.number', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='difficulty', full_name='qrl.BlockDataPoint.difficulty', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='qrl.BlockDataPoint.timestamp', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='time_last', full_name='qrl.BlockDataPoint.time_last', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='time_movavg', full_name='qrl.BlockDataPoint.time_movavg', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hash_power', full_name='qrl.BlockDataPoint.hash_power', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='header_hash', full_name='qrl.BlockDataPoint.header_hash', index=6,
      number=7, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='header_hash_prev', full_name='qrl.BlockDataPoint.header_hash_prev', index=7,
      number=8, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=796,
  serialized_end=974,
)


_GETADDRESSSTATEREQ = _descriptor.Descriptor(
  name='GetAddressStateReq',
  full_name='qrl.GetAddressStateReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='qrl.GetAddressStateReq.address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='exclude_ots_bitfield', full_name='qrl.GetAddressStateReq.exclude_ots_bitfield', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='exclude_transaction_hashes', full_name='qrl.GetAddressStateReq.exclude_transaction_hashes', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=976,
  serialized_end=1079,
)


_GETADDRESSSTATERESP = _descriptor.Descriptor(
  name='GetAddressStateResp',
  full_name='qrl.GetAddressStateResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='state', full_name='qrl.GetAddressStateResp.state', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1081,
  serialized_end=1136,
)


_GETOPTIMIZEDADDRESSSTATERESP = _descriptor.Descriptor(
  name='GetOptimizedAddressStateResp',
  full_name='qrl.GetOptimizedAddressStateResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='state', full_name='qrl.GetOptimizedAddressStateResp.state', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1138,
  serialized_end=1211,
)


_GETMULTISIGADDRESSSTATEREQ = _descriptor.Descriptor(
  name='GetMultiSigAddressStateReq',
  full_name='qrl.GetMultiSigAddressStateReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='qrl.GetMultiSigAddressStateReq.address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1213,
  serialized_end=1258,
)


_GETMULTISIGADDRESSSTATERESP = _descriptor.Descriptor(
  name='GetMultiSigAddressStateResp',
  full_name='qrl.GetMultiSigAddressStateResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='state', full_name='qrl.GetMultiSigAddressStateResp.state', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1260,
  serialized_end=1331,
)


_ISSLAVEREQ = _descriptor.Descriptor(
  name='IsSlaveReq',
  full_name='qrl.IsSlaveReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='master_address', full_name='qrl.IsSlaveReq.master_address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='slave_pk', full_name='qrl.IsSlaveReq.slave_pk', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1333,
  serialized_end=1387,
)


_ISSLAVERESP = _descriptor.Descriptor(
  name='IsSlaveResp',
  full_name='qrl.IsSlaveResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='result', full_name='qrl.IsSlaveResp.result', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1389,
  serialized_end=1418,
)


_PARSEADDRESSREQ = _descriptor.Descriptor(
  name='ParseAddressReq',
  full_name='qrl.ParseAddressReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='qrl.ParseAddressReq.address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1420,
  serialized_end=1454,
)


_PARSEADDRESSRESP = _descriptor.Descriptor(
  name='ParseAddressResp',
  full_name='qrl.ParseAddressResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='is_valid', full_name='qrl.ParseAddressResp.is_valid', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='desc', full_name='qrl.ParseAddressResp.desc', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1456,
  serialized_end=1530,
)


_GETOBJECTREQ = _descriptor.Descriptor(
  name='GetObjectReq',
  full_name='qrl.GetObjectReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='query', full_name='qrl.GetObjectReq.query', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1532,
  serialized_end=1561,
)


_GETOBJECTRESP = _descriptor.Descriptor(
  name='GetObjectResp',
  full_name='qrl.GetObjectResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='found', full_name='qrl.GetObjectResp.found', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='address_state', full_name='qrl.GetObjectResp.address_state', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transaction', full_name='qrl.GetObjectResp.transaction', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_extended', full_name='qrl.GetObjectResp.block_extended', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='result', full_name='qrl.GetObjectResp.result',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=1564,
  serialized_end=1752,
)


_GETLATESTDATAREQ = _descriptor.Descriptor(
  name='GetLatestDataReq',
  full_name='qrl.GetLatestDataReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='filter', full_name='qrl.GetLatestDataReq.filter', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='offset', full_name='qrl.GetLatestDataReq.offset', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='quantity', full_name='qrl.GetLatestDataReq.quantity', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _GETLATESTDATAREQ_FILTER,
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1755,
  serialized_end=1938,
)


_GETLATESTDATARESP = _descriptor.Descriptor(
  name='GetLatestDataResp',
  full_name='qrl.GetLatestDataResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='blockheaders', full_name='qrl.GetLatestDataResp.blockheaders', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transactions', full_name='qrl.GetLatestDataResp.transactions', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transactions_unconfirmed', full_name='qrl.GetLatestDataResp.transactions_unconfirmed', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1941,
  serialized_end=2116,
)


_TRANSFERCOINSREQ = _descriptor.Descriptor(
  name='TransferCoinsReq',
  full_name='qrl.TransferCoinsReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='master_addr', full_name='qrl.TransferCoinsReq.master_addr', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='addresses_to', full_name='qrl.TransferCoinsReq.addresses_to', index=1,
      number=2, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='amounts', full_name='qrl.TransferCoinsReq.amounts', index=2,
      number=3, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message_data', full_name='qrl.TransferCoinsReq.message_data', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fee', full_name='qrl.TransferCoinsReq.fee', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='xmss_pk', full_name='qrl.TransferCoinsReq.xmss_pk', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2119,
  serialized_end=2249,
)


_TRANSFERCOINSRESP = _descriptor.Descriptor(
  name='TransferCoinsResp',
  full_name='qrl.TransferCoinsResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='extended_transaction_unsigned', full_name='qrl.TransferCoinsResp.extended_transaction_unsigned', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2251,
  serialized_end=2335,
)


_PUSHTRANSACTIONREQ = _descriptor.Descriptor(
  name='PushTransactionReq',
  full_name='qrl.PushTransactionReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='transaction_signed', full_name='qrl.PushTransactionReq.transaction_signed', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2337,
  serialized_end=2403,
)


_PUSHTRANSACTIONRESP = _descriptor.Descriptor(
  name='PushTransactionResp',
  full_name='qrl.PushTransactionResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error_code', full_name='qrl.PushTransactionResp.error_code', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error_description', full_name='qrl.PushTransactionResp.error_description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tx_hash', full_name='qrl.PushTransactionResp.tx_hash', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _PUSHTRANSACTIONRESP_RESPONSECODE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2406,
  serialized_end=2608,
)


_MULTISIGCREATETXNREQ = _descriptor.Descriptor(
  name='MultiSigCreateTxnReq',
  full_name='qrl.MultiSigCreateTxnReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='master_addr', full_name='qrl.MultiSigCreateTxnReq.master_addr', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='signatories', full_name='qrl.MultiSigCreateTxnReq.signatories', index=1,
      number=2, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='weights', full_name='qrl.MultiSigCreateTxnReq.weights', index=2,
      number=3, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='threshold', full_name='qrl.MultiSigCreateTxnReq.threshold', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fee', full_name='qrl.MultiSigCreateTxnReq.fee', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='xmss_pk', full_name='qrl.MultiSigCreateTxnReq.xmss_pk', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2611,
  serialized_end=2741,
)


_MULTISIGSPENDTXNREQ = _descriptor.Descriptor(
  name='MultiSigSpendTxnReq',
  full_name='qrl.MultiSigSpendTxnReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='master_addr', full_name='qrl.MultiSigSpendTxnReq.master_addr', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='multi_sig_address', full_name='qrl.MultiSigSpendTxnReq.multi_sig_address', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='addrs_to', full_name='qrl.MultiSigSpendTxnReq.addrs_to', index=2,
      number=3, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='amounts', full_name='qrl.MultiSigSpendTxnReq.amounts', index=3,
      number=4, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='expiry_block_number', full_name='qrl.MultiSigSpendTxnReq.expiry_block_number', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fee', full_name='qrl.MultiSigSpendTxnReq.fee', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='xmss_pk', full_name='qrl.MultiSigSpendTxnReq.xmss_pk', index=6,
      number=7, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2744,
  serialized_end=2907,
)


_MULTISIGVOTETXNREQ = _descriptor.Descriptor(
  name='MultiSigVoteTxnReq',
  full_name='qrl.MultiSigVoteTxnReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='master_addr', full_name='qrl.MultiSigVoteTxnReq.master_addr', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shared_key', full_name='qrl.MultiSigVoteTxnReq.shared_key', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='unvote', full_name='qrl.MultiSigVoteTxnReq.unvote', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fee', full_name='qrl.MultiSigVoteTxnReq.fee', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='xmss_pk', full_name='qrl.MultiSigVoteTxnReq.xmss_pk', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2909,
  serialized_end=3016,
)


_MESSAGETXNREQ = _descriptor.Descriptor(
  name='MessageTxnReq',
  full_name='qrl.MessageTxnReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='master_addr', full_name='qrl.MessageTxnReq.master_addr', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='qrl.MessageTxnReq.message', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='addr_to', full_name='qrl.MessageTxnReq.addr_to', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fee', full_name='qrl.MessageTxnReq.fee', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='xmss_pk', full_name='qrl.MessageTxnReq.xmss_pk', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3018,
  serialized_end=3118,
)


_TOKENTXNREQ = _descriptor.Descriptor(
  name='TokenTxnReq',
  full_name='qrl.TokenTxnReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='master_addr', full_name='qrl.TokenTxnReq.master_addr', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='symbol', full_name='qrl.TokenTxnReq.symbol', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='qrl.TokenTxnReq.name', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='owner', full_name='qrl.TokenTxnReq.owner', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='decimals', full_name='qrl.TokenTxnReq.decimals', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='initial_balances', full_name='qrl.TokenTxnReq.initial_balances', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fee', full_name='qrl.TokenTxnReq.fee', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='xmss_pk', full_name='qrl.TokenTxnReq.xmss_pk', index=7,
      number=8, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3121,
  serialized_end=3294,
)


_TRANSFERTOKENTXNREQ = _descriptor.Descriptor(
  name='TransferTokenTxnReq',
  full_name='qrl.TransferTokenTxnReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='master_addr', full_name='qrl.TransferTokenTxnReq.master_addr', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='addresses_to', full_name='qrl.TransferTokenTxnReq.addresses_to', index=1,
      number=2, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='token_txhash', full_name='qrl.TransferTokenTxnReq.token_txhash', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='amounts', full_name='qrl.TransferTokenTxnReq.amounts', index=3,
      number=4, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fee', full_name='qrl.TransferTokenTxnReq.fee', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='xmss_pk', full_name='qrl.TransferTokenTxnReq.xmss_pk', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3297,
  serialized_end=3430,
)


_SLAVETXNREQ = _descriptor.Descriptor(
  name='SlaveTxnReq',
  full_name='qrl.SlaveTxnReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='master_addr', full_name='qrl.SlaveTxnReq.master_addr', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='slave_pks', full_name='qrl.SlaveTxnReq.slave_pks', index=1,
      number=2, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='access_types', full_name='qrl.SlaveTxnReq.access_types', index=2,
      number=3, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fee', full_name='qrl.SlaveTxnReq.fee', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='xmss_pk', full_name='qrl.SlaveTxnReq.xmss_pk', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3432,
  serialized_end=3537,
)


_LATTICETXNREQ = _descriptor.Descriptor(
  name='LatticeTxnReq',
  full_name='qrl.LatticeTxnReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='master_addr', full_name='qrl.LatticeTxnReq.master_addr', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pk1', full_name='qrl.LatticeTxnReq.pk1', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pk2', full_name='qrl.LatticeTxnReq.pk2', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pk3', full_name='qrl.LatticeTxnReq.pk3', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fee', full_name='qrl.LatticeTxnReq.fee', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='xmss_pk', full_name='qrl.LatticeTxnReq.xmss_pk', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3539,
  serialized_end=3644,
)


_MINITRANSACTION = _descriptor.Descriptor(
  name='MiniTransaction',
  full_name='qrl.MiniTransaction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='transaction_hash', full_name='qrl.MiniTransaction.transaction_hash', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out', full_name='qrl.MiniTransaction.out', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='amount', full_name='qrl.MiniTransaction.amount', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3646,
  serialized_end=3718,
)


_GETTRANSACTIONREQ = _descriptor.Descriptor(
  name='GetTransactionReq',
  full_name='qrl.GetTransactionReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='tx_hash', full_name='qrl.GetTransactionReq.tx_hash', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3720,
  serialized_end=3756,
)


_GETTRANSACTIONRESP = _descriptor.Descriptor(
  name='GetTransactionResp',
  full_name='qrl.GetTransactionResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='tx', full_name='qrl.GetTransactionResp.tx', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='confirmations', full_name='qrl.GetTransactionResp.confirmations', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_number', full_name='qrl.GetTransactionResp.block_number', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_header_hash', full_name='qrl.GetTransactionResp.block_header_hash', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='qrl.GetTransactionResp.timestamp', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='addr_from', full_name='qrl.GetTransactionResp.addr_from', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3759,
  serialized_end=3919,
)


_GETMINITRANSACTIONSBYADDRESSREQ = _descriptor.Descriptor(
  name='GetMiniTransactionsByAddressReq',
  full_name='qrl.GetMiniTransactionsByAddressReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='qrl.GetMiniTransactionsByAddressReq.address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='item_per_page', full_name='qrl.GetMiniTransactionsByAddressReq.item_per_page', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='page_number', full_name='qrl.GetMiniTransactionsByAddressReq.page_number', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3921,
  serialized_end=4015,
)


_GETMINITRANSACTIONSBYADDRESSRESP = _descriptor.Descriptor(
  name='GetMiniTransactionsByAddressResp',
  full_name='qrl.GetMiniTransactionsByAddressResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='mini_transactions', full_name='qrl.GetMiniTransactionsByAddressResp.mini_transactions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='balance', full_name='qrl.GetMiniTransactionsByAddressResp.balance', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4017,
  serialized_end=4117,
)


_GETTRANSACTIONSBYADDRESSREQ = _descriptor.Descriptor(
  name='GetTransactionsByAddressReq',
  full_name='qrl.GetTransactionsByAddressReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='qrl.GetTransactionsByAddressReq.address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='item_per_page', full_name='qrl.GetTransactionsByAddressReq.item_per_page', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='page_number', full_name='qrl.GetTransactionsByAddressReq.page_number', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4119,
  serialized_end=4209,
)


_GETTRANSACTIONSBYADDRESSRESP = _descriptor.Descriptor(
  name='GetTransactionsByAddressResp',
  full_name='qrl.GetTransactionsByAddressResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='transactions_detail', full_name='qrl.GetTransactionsByAddressResp.transactions_detail', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4211,
  serialized_end=4295,
)


_GETMULTISIGSPENDTXSBYADDRESSREQ = _descriptor.Descriptor(
  name='GetMultiSigSpendTxsByAddressReq',
  full_name='qrl.GetMultiSigSpendTxsByAddressReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='qrl.GetMultiSigSpendTxsByAddressReq.address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='item_per_page', full_name='qrl.GetMultiSigSpendTxsByAddressReq.item_per_page', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='page_number', full_name='qrl.GetMultiSigSpendTxsByAddressReq.page_number', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='filter_type', full_name='qrl.GetMultiSigSpendTxsByAddressReq.filter_type', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _GETMULTISIGSPENDTXSBYADDRESSREQ_FILTERTYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4298,
  serialized_end=4610,
)


_GETMULTISIGSPENDTXSBYADDRESSRESP = _descriptor.Descriptor(
  name='GetMultiSigSpendTxsByAddressResp',
  full_name='qrl.GetMultiSigSpendTxsByAddressResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='transactions_detail', full_name='qrl.GetMultiSigSpendTxsByAddressResp.transactions_detail', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4612,
  serialized_end=4700,
)


_GETVOTESTATSREQ = _descriptor.Descriptor(
  name='GetVoteStatsReq',
  full_name='qrl.GetVoteStatsReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='multi_sig_spend_tx_hash', full_name='qrl.GetVoteStatsReq.multi_sig_spend_tx_hash', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4702,
  serialized_end=4752,
)


_GETVOTESTATSRESP = _descriptor.Descriptor(
  name='GetVoteStatsResp',
  full_name='qrl.GetVoteStatsResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='vote_stats', full_name='qrl.GetVoteStatsResp.vote_stats', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4754,
  serialized_end=4808,
)


_GETINBOXMESSAGESBYADDRESSRESP = _descriptor.Descriptor(
  name='GetInboxMessagesByAddressResp',
  full_name='qrl.GetInboxMessagesByAddressResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='transactions_detail', full_name='qrl.GetInboxMessagesByAddressResp.transactions_detail', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4810,
  serialized_end=4895,
)


_INBOXMESSAGE = _descriptor.Descriptor(
  name='InboxMessage',
  full_name='qrl.InboxMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='addr_from', full_name='qrl.InboxMessage.addr_from', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='qrl.InboxMessage.timestamp', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='qrl.InboxMessage.message', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tx_hash', full_name='qrl.InboxMessage.tx_hash', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_number', full_name='qrl.InboxMessage.block_number', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4897,
  serialized_end=5005,
)


_TOKENDETAIL = _descriptor.Descriptor(
  name='TokenDetail',
  full_name='qrl.TokenDetail',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='token_txhash', full_name='qrl.TokenDetail.token_txhash', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='qrl.TokenDetail.name', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='symbol', full_name='qrl.TokenDetail.symbol', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='balance', full_name='qrl.TokenDetail.balance', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5007,
  serialized_end=5089,
)


_GETTOKENSBYADDRESSRESP = _descriptor.Descriptor(
  name='GetTokensByAddressResp',
  full_name='qrl.GetTokensByAddressResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='tokens_detail', full_name='qrl.GetTokensByAddressResp.tokens_detail', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5091,
  serialized_end=5156,
)


_SLAVEDETAIL = _descriptor.Descriptor(
  name='SlaveDetail',
  full_name='qrl.SlaveDetail',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='slave_address', full_name='qrl.SlaveDetail.slave_address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='access_type', full_name='qrl.SlaveDetail.access_type', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5158,
  serialized_end=5215,
)


_GETSLAVESBYADDRESSRESP = _descriptor.Descriptor(
  name='GetSlavesByAddressResp',
  full_name='qrl.GetSlavesByAddressResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='slaves_detail', full_name='qrl.GetSlavesByAddressResp.slaves_detail', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5217,
  serialized_end=5282,
)


_LATTICEPKSDETAIL = _descriptor.Descriptor(
  name='LatticePKsDetail',
  full_name='qrl.LatticePKsDetail',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pk1', full_name='qrl.LatticePKsDetail.pk1', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pk2', full_name='qrl.LatticePKsDetail.pk2', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pk3', full_name='qrl.LatticePKsDetail.pk3', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tx_hash', full_name='qrl.LatticePKsDetail.tx_hash', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5284,
  serialized_end=5358,
)


_GETLATTICEPKSBYADDRESSRESP = _descriptor.Descriptor(
  name='GetLatticePKsByAddressResp',
  full_name='qrl.GetLatticePKsByAddressResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='lattice_pks_detail', full_name='qrl.GetLatticePKsByAddressResp.lattice_pks_detail', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5360,
  serialized_end=5439,
)


_MULTISIGDETAIL = _descriptor.Descriptor(
  name='MultiSigDetail',
  full_name='qrl.MultiSigDetail',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='qrl.MultiSigDetail.address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='balance', full_name='qrl.MultiSigDetail.balance', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5441,
  serialized_end=5491,
)


_GETMULTISIGADDRESSESBYADDRESSRESP = _descriptor.Descriptor(
  name='GetMultiSigAddressesByAddressResp',
  full_name='qrl.GetMultiSigAddressesByAddressResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='multi_sig_detail', full_name='qrl.GetMultiSigAddressesByAddressResp.multi_sig_detail', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5493,
  serialized_end=5575,
)


_GETBALANCEREQ = _descriptor.Descriptor(
  name='GetBalanceReq',
  full_name='qrl.GetBalanceReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='qrl.GetBalanceReq.address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5577,
  serialized_end=5609,
)


_GETBALANCERESP = _descriptor.Descriptor(
  name='GetBalanceResp',
  full_name='qrl.GetBalanceResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='balance', full_name='qrl.GetBalanceResp.balance', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5611,
  serialized_end=5644,
)


_GETTOTALBALANCEREQ = _descriptor.Descriptor(
  name='GetTotalBalanceReq',
  full_name='qrl.GetTotalBalanceReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='addresses', full_name='qrl.GetTotalBalanceReq.addresses', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5646,
  serialized_end=5685,
)


_GETTOTALBALANCERESP = _descriptor.Descriptor(
  name='GetTotalBalanceResp',
  full_name='qrl.GetTotalBalanceResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='balance', full_name='qrl.GetTotalBalanceResp.balance', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5687,
  serialized_end=5725,
)


_GETOTSREQ = _descriptor.Descriptor(
  name='GetOTSReq',
  full_name='qrl.GetOTSReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='qrl.GetOTSReq.address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='page_from', full_name='qrl.GetOTSReq.page_from', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='page_count', full_name='qrl.GetOTSReq.page_count', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='unused_ots_index_from', full_name='qrl.GetOTSReq.unused_ots_index_from', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5727,
  serialized_end=5825,
)


_OTSBITFIELDBYPAGE = _descriptor.Descriptor(
  name='OTSBitfieldByPage',
  full_name='qrl.OTSBitfieldByPage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ots_bitfield', full_name='qrl.OTSBitfieldByPage.ots_bitfield', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='page_number', full_name='qrl.OTSBitfieldByPage.page_number', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5827,
  serialized_end=5889,
)


_GETOTSRESP = _descriptor.Descriptor(
  name='GetOTSResp',
  full_name='qrl.GetOTSResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ots_bitfield_by_page', full_name='qrl.GetOTSResp.ots_bitfield_by_page', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='next_unused_ots_index', full_name='qrl.GetOTSResp.next_unused_ots_index', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='unused_ots_index_found', full_name='qrl.GetOTSResp.unused_ots_index_found', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5892,
  serialized_end=6021,
)


_GETHEIGHTREQ = _descriptor.Descriptor(
  name='GetHeightReq',
  full_name='qrl.GetHeightReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6023,
  serialized_end=6037,
)


_GETHEIGHTRESP = _descriptor.Descriptor(
  name='GetHeightResp',
  full_name='qrl.GetHeightResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='height', full_name='qrl.GetHeightResp.height', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6039,
  serialized_end=6070,
)


_GETBLOCKREQ = _descriptor.Descriptor(
  name='GetBlockReq',
  full_name='qrl.GetBlockReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header_hash', full_name='qrl.GetBlockReq.header_hash', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6072,
  serialized_end=6106,
)


_GETBLOCKRESP = _descriptor.Descriptor(
  name='GetBlockResp',
  full_name='qrl.GetBlockResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='block', full_name='qrl.GetBlockResp.block', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6108,
  serialized_end=6149,
)


_GETBLOCKBYNUMBERREQ = _descriptor.Descriptor(
  name='GetBlockByNumberReq',
  full_name='qrl.GetBlockByNumberReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='block_number', full_name='qrl.GetBlockByNumberReq.block_number', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6151,
  serialized_end=6194,
)


_GETBLOCKBYNUMBERRESP = _descriptor.Descriptor(
  name='GetBlockByNumberResp',
  full_name='qrl.GetBlockByNumberResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='block', full_name='qrl.GetBlockByNumberResp.block', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6196,
  serialized_end=6245,
)


_GETLOCALADDRESSESREQ = _descriptor.Descriptor(
  name='GetLocalAddressesReq',
  full_name='qrl.GetLocalAddressesReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6247,
  serialized_end=6269,
)


_GETLOCALADDRESSESRESP = _descriptor.Descriptor(
  name='GetLocalAddressesResp',
  full_name='qrl.GetLocalAddressesResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='addresses', full_name='qrl.GetLocalAddressesResp.addresses', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6271,
  serialized_end=6313,
)


_NODEINFO = _descriptor.Descriptor(
  name='NodeInfo',
  full_name='qrl.NodeInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='qrl.NodeInfo.version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='state', full_name='qrl.NodeInfo.state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='num_connections', full_name='qrl.NodeInfo.num_connections', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='num_known_peers', full_name='qrl.NodeInfo.num_known_peers', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='uptime', full_name='qrl.NodeInfo.uptime', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_height', full_name='qrl.NodeInfo.block_height', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_last_hash', full_name='qrl.NodeInfo.block_last_hash', index=6,
      number=7, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='network_id', full_name='qrl.NodeInfo.network_id', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _NODEINFO_STATE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6316,
  serialized_end=6585,
)


_ADDRESSDESCRIPTOR = _descriptor.Descriptor(
  name='AddressDescriptor',
  full_name='qrl.AddressDescriptor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='hash_function', full_name='qrl.AddressDescriptor.hash_function', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='signature_scheme', full_name='qrl.AddressDescriptor.signature_scheme', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tree_height', full_name='qrl.AddressDescriptor.tree_height', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='signatures', full_name='qrl.AddressDescriptor.signatures', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='address_format', full_name='qrl.AddressDescriptor.address_format', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6588,
  serialized_end=6721,
)


_STOREDPEERS = _descriptor.Descriptor(
  name='StoredPeers',
  full_name='qrl.StoredPeers',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='peers', full_name='qrl.StoredPeers.peers', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6723,
  serialized_end=6762,
)


_PEER = _descriptor.Descriptor(
  name='Peer',
  full_name='qrl.Peer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ip', full_name='qrl.Peer.ip', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6764,
  serialized_end=6782,
)


_ADDRESSSTATE_TOKENSENTRY = _descriptor.Descriptor(
  name='TokensEntry',
  full_name='qrl.AddressState.TokensEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='qrl.AddressState.TokensEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='qrl.AddressState.TokensEntry.value', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=_descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001')),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7082,
  serialized_end=7127,
)

_ADDRESSSTATE_SLAVEPKSACCESSTYPEENTRY = _descriptor.Descriptor(
  name='SlavePksAccessTypeEntry',
  full_name='qrl.AddressState.SlavePksAccessTypeEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='qrl.AddressState.SlavePksAccessTypeEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='qrl.AddressState.SlavePksAccessTypeEntry.value', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=_descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001')),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7129,
  serialized_end=7186,
)

_ADDRESSSTATE = _descriptor.Descriptor(
  name='AddressState',
  full_name='qrl.AddressState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='qrl.AddressState.address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='balance', full_name='qrl.AddressState.balance', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='nonce', full_name='qrl.AddressState.nonce', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ots_bitfield', full_name='qrl.AddressState.ots_bitfield', index=3,
      number=4, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transaction_hashes', full_name='qrl.AddressState.transaction_hashes', index=4,
      number=5, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tokens', full_name='qrl.AddressState.tokens', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='latticePK_list', full_name='qrl.AddressState.latticePK_list', index=6,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='slave_pks_access_type', full_name='qrl.AddressState.slave_pks_access_type', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ots_counter', full_name='qrl.AddressState.ots_counter', index=8,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_ADDRESSSTATE_TOKENSENTRY, _ADDRESSSTATE_SLAVEPKSACCESSTYPEENTRY, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6785,
  serialized_end=7186,
)


_OPTIMIZEDADDRESSSTATE = _descriptor.Descriptor(
  name='OptimizedAddressState',
  full_name='qrl.OptimizedAddressState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='qrl.OptimizedAddressState.address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='balance', full_name='qrl.OptimizedAddressState.balance', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='nonce', full_name='qrl.OptimizedAddressState.nonce', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ots_bitfield_used_page', full_name='qrl.OptimizedAddressState.ots_bitfield_used_page', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='used_ots_key_count', full_name='qrl.OptimizedAddressState.used_ots_key_count', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transaction_hash_count', full_name='qrl.OptimizedAddressState.transaction_hash_count', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tokens_count', full_name='qrl.OptimizedAddressState.tokens_count', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='slaves_count', full_name='qrl.OptimizedAddressState.slaves_count', index=7,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lattice_pk_count', full_name='qrl.OptimizedAddressState.lattice_pk_count', index=8,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='multi_sig_address_count', full_name='qrl.OptimizedAddressState.multi_sig_address_count', index=9,
      number=10, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='multi_sig_spend_count', full_name='qrl.OptimizedAddressState.multi_sig_spend_count', index=10,
      number=11, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='inbox_message_count', full_name='qrl.OptimizedAddressState.inbox_message_count', index=11,
      number=12, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='foundation_multi_sig_spend_txn_hash', full_name='qrl.OptimizedAddressState.foundation_multi_sig_spend_txn_hash', index=12,
      number=13, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='foundation_multi_sig_vote_txn_hash', full_name='qrl.OptimizedAddressState.foundation_multi_sig_vote_txn_hash', index=13,
      number=14, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='unvotes', full_name='qrl.OptimizedAddressState.unvotes', index=14,
      number=15, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proposal_vote_stats', full_name='qrl.OptimizedAddressState.proposal_vote_stats', index=15,
      number=16, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7189,
  serialized_end=7669,
)


_MULTISIGADDRESSSTATE = _descriptor.Descriptor(
  name='MultiSigAddressState',
  full_name='qrl.MultiSigAddressState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='qrl.MultiSigAddressState.address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='creation_tx_hash', full_name='qrl.MultiSigAddressState.creation_tx_hash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='nonce', full_name='qrl.MultiSigAddressState.nonce', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='balance', full_name='qrl.MultiSigAddressState.balance', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='signatories', full_name='qrl.MultiSigAddressState.signatories', index=4,
      number=5, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='weights', full_name='qrl.MultiSigAddressState.weights', index=5,
      number=6, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='threshold', full_name='qrl.MultiSigAddressState.threshold', index=6,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transaction_hash_count', full_name='qrl.MultiSigAddressState.transaction_hash_count', index=7,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='multi_sig_spend_count', full_name='qrl.MultiSigAddressState.multi_sig_spend_count', index=8,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='multi_sig_address_count', full_name='qrl.MultiSigAddressState.multi_sig_address_count', index=9,
      number=10, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='foundation_multi_sig_spend_txn_hash', full_name='qrl.MultiSigAddressState.foundation_multi_sig_spend_txn_hash', index=10,
      number=11, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='foundation_multi_sig_vote_txn_hash', full_name='qrl.MultiSigAddressState.foundation_multi_sig_vote_txn_hash', index=11,
      number=12, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='unvotes', full_name='qrl.MultiSigAddressState.unvotes', index=12,
      number=13, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proposal_vote_stats', full_name='qrl.MultiSigAddressState.proposal_vote_stats', index=13,
      number=14, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7672,
  serialized_end=8075,
)


_MULTISIGADDRESSESLIST = _descriptor.Descriptor(
  name='MultiSigAddressesList',
  full_name='qrl.MultiSigAddressesList',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='hashes', full_name='qrl.MultiSigAddressesList.hashes', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8077,
  serialized_end=8116,
)


_DATALIST = _descriptor.Descriptor(
  name='DataList',
  full_name='qrl.DataList',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='values', full_name='qrl.DataList.values', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8118,
  serialized_end=8144,
)


_BITFIELD = _descriptor.Descriptor(
  name='Bitfield',
  full_name='qrl.Bitfield',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='bitfields', full_name='qrl.Bitfield.bitfields', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8146,
  serialized_end=8175,
)


_TRANSACTIONHASHLIST = _descriptor.Descriptor(
  name='TransactionHashList',
  full_name='qrl.TransactionHashList',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='hashes', full_name='qrl.TransactionHashList.hashes', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8177,
  serialized_end=8214,
)


_LATTICEPK = _descriptor.Descriptor(
  name='LatticePK',
  full_name='qrl.LatticePK',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='kyber_pk', full_name='qrl.LatticePK.kyber_pk', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dilithium_pk', full_name='qrl.LatticePK.dilithium_pk', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8216,
  serialized_end=8267,
)


_ADDRESSAMOUNT = _descriptor.Descriptor(
  name='AddressAmount',
  full_name='qrl.AddressAmount',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='qrl.AddressAmount.address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='amount', full_name='qrl.AddressAmount.amount', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8269,
  serialized_end=8317,
)


_BLOCKHEADER = _descriptor.Descriptor(
  name='BlockHeader',
  full_name='qrl.BlockHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='hash_header', full_name='qrl.BlockHeader.hash_header', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_number', full_name='qrl.BlockHeader.block_number', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp_seconds', full_name='qrl.BlockHeader.timestamp_seconds', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hash_header_prev', full_name='qrl.BlockHeader.hash_header_prev', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='reward_block', full_name='qrl.BlockHeader.reward_block', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='reward_fee', full_name='qrl.BlockHeader.reward_fee', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='merkle_root', full_name='qrl.BlockHeader.merkle_root', index=6,
      number=7, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mining_nonce', full_name='qrl.BlockHeader.mining_nonce', index=7,
      number=8, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='extra_nonce', full_name='qrl.BlockHeader.extra_nonce', index=8,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8320,
  serialized_end=8535,
)


_BLOCKHEADEREXTENDED = _descriptor.Descriptor(
  name='BlockHeaderExtended',
  full_name='qrl.BlockHeaderExtended',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='qrl.BlockHeaderExtended.header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transaction_count', full_name='qrl.BlockHeaderExtended.transaction_count', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8537,
  serialized_end=8642,
)


_TRANSACTIONCOUNT_COUNTENTRY = _descriptor.Descriptor(
  name='CountEntry',
  full_name='qrl.TransactionCount.CountEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='qrl.TransactionCount.CountEntry.key', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='qrl.TransactionCount.CountEntry.value', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=_descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001')),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8713,
  serialized_end=8757,
)

_TRANSACTIONCOUNT = _descriptor.Descriptor(
  name='TransactionCount',
  full_name='qrl.TransactionCount',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='count', full_name='qrl.TransactionCount.count', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTIONCOUNT_COUNTENTRY, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8644,
  serialized_end=8757,
)


_TRANSACTIONEXTENDED = _descriptor.Descriptor(
  name='TransactionExtended',
  full_name='qrl.TransactionExtended',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='qrl.TransactionExtended.header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tx', full_name='qrl.TransactionExtended.tx', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='addr_from', full_name='qrl.TransactionExtended.addr_from', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='size', full_name='qrl.TransactionExtended.size', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp_seconds', full_name='qrl.TransactionExtended.timestamp_seconds', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8760,
  serialized_end=8905,
)


_BLOCKEXTENDED = _descriptor.Descriptor(
  name='BlockExtended',
  full_name='qrl.BlockExtended',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='qrl.BlockExtended.header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='extended_transactions', full_name='qrl.BlockExtended.extended_transactions', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='genesis_balance', full_name='qrl.BlockExtended.genesis_balance', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='size', full_name='qrl.BlockExtended.size', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8908,
  serialized_end=9074,
)


_BLOCK = _descriptor.Descriptor(
  name='Block',
  full_name='qrl.Block',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='qrl.Block.header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transactions', full_name='qrl.Block.transactions', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='genesis_balance', full_name='qrl.Block.genesis_balance', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9076,
  serialized_end=9203,
)


_GENESISBALANCE = _descriptor.Descriptor(
  name='GenesisBalance',
  full_name='qrl.GenesisBalance',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='qrl.GenesisBalance.address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='balance', full_name='qrl.GenesisBalance.balance', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9205,
  serialized_end=9255,
)


_BLOCKMETADATALIST = _descriptor.Descriptor(
  name='BlockMetaDataList',
  full_name='qrl.BlockMetaDataList',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='block_number_hashes', full_name='qrl.BlockMetaDataList.block_number_hashes', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9257,
  serialized_end=9325,
)


_TRANSACTION_TRANSFER = _descriptor.Descriptor(
  name='Transfer',
  full_name='qrl.Transaction.Transfer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='addrs_to', full_name='qrl.Transaction.Transfer.addrs_to', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='amounts', full_name='qrl.Transaction.Transfer.amounts', index=1,
      number=2, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message_data', full_name='qrl.Transaction.Transfer.message_data', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10085,
  serialized_end=10152,
)

_TRANSACTION_COINBASE = _descriptor.Descriptor(
  name='CoinBase',
  full_name='qrl.Transaction.CoinBase',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='addr_to', full_name='qrl.Transaction.CoinBase.addr_to', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='amount', full_name='qrl.Transaction.CoinBase.amount', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10154,
  serialized_end=10197,
)

_TRANSACTION_LATTICEPUBLICKEY = _descriptor.Descriptor(
  name='LatticePublicKey',
  full_name='qrl.Transaction.LatticePublicKey',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pk1', full_name='qrl.Transaction.LatticePublicKey.pk1', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pk2', full_name='qrl.Transaction.LatticePublicKey.pk2', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pk3', full_name='qrl.Transaction.LatticePublicKey.pk3', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10199,
  serialized_end=10256,
)

_TRANSACTION_MESSAGE = _descriptor.Descriptor(
  name='Message',
  full_name='qrl.Transaction.Message',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='message_hash', full_name='qrl.Transaction.Message.message_hash', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='addr_to', full_name='qrl.Transaction.Message.addr_to', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10258,
  serialized_end=10306,
)

_TRANSACTION_TOKEN = _descriptor.Descriptor(
  name='Token',
  full_name='qrl.Transaction.Token',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='symbol', full_name='qrl.Transaction.Token.symbol', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='qrl.Transaction.Token.name', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='owner', full_name='qrl.Transaction.Token.owner', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='decimals', full_name='qrl.Transaction.Token.decimals', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='initial_balances', full_name='qrl.Transaction.Token.initial_balances', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10308,
  serialized_end=10424,
)

_TRANSACTION_TRANSFERTOKEN = _descriptor.Descriptor(
  name='TransferToken',
  full_name='qrl.Transaction.TransferToken',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='token_txhash', full_name='qrl.Transaction.TransferToken.token_txhash', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='addrs_to', full_name='qrl.Transaction.TransferToken.addrs_to', index=1,
      number=2, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='amounts', full_name='qrl.Transaction.TransferToken.amounts', index=2,
      number=3, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10426,
  serialized_end=10498,
)

_TRANSACTION_SLAVE = _descriptor.Descriptor(
  name='Slave',
  full_name='qrl.Transaction.Slave',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='slave_pks', full_name='qrl.Transaction.Slave.slave_pks', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='access_types', full_name='qrl.Transaction.Slave.access_types', index=1,
      number=2, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10500,
  serialized_end=10548,
)

_TRANSACTION_MULTISIGCREATE = _descriptor.Descriptor(
  name='MultiSigCreate',
  full_name='qrl.Transaction.MultiSigCreate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='signatories', full_name='qrl.Transaction.MultiSigCreate.signatories', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='weights', full_name='qrl.Transaction.MultiSigCreate.weights', index=1,
      number=2, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='threshold', full_name='qrl.Transaction.MultiSigCreate.threshold', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10550,
  serialized_end=10623,
)

_TRANSACTION_MULTISIGSPEND = _descriptor.Descriptor(
  name='MultiSigSpend',
  full_name='qrl.Transaction.MultiSigSpend',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='multi_sig_address', full_name='qrl.Transaction.MultiSigSpend.multi_sig_address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='addrs_to', full_name='qrl.Transaction.MultiSigSpend.addrs_to', index=1,
      number=2, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='amounts', full_name='qrl.Transaction.MultiSigSpend.amounts', index=2,
      number=3, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='expiry_block_number', full_name='qrl.Transaction.MultiSigSpend.expiry_block_number', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10625,
  serialized_end=10731,
)

_TRANSACTION_MULTISIGVOTE = _descriptor.Descriptor(
  name='MultiSigVote',
  full_name='qrl.Transaction.MultiSigVote',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='shared_key', full_name='qrl.Transaction.MultiSigVote.shared_key', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='unvote', full_name='qrl.Transaction.MultiSigVote.unvote', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='prev_tx_hash', full_name='qrl.Transaction.MultiSigVote.prev_tx_hash', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10733,
  serialized_end=10805,
)

_TRANSACTION_PROPOSALCREATE_QIP = _descriptor.Descriptor(
  name='QIP',
  full_name='qrl.Transaction.ProposalCreate.QIP',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='qip_link', full_name='qrl.Transaction.ProposalCreate.QIP.qip_link', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11042,
  serialized_end=11065,
)

_TRANSACTION_PROPOSALCREATE_CONFIG = _descriptor.Descriptor(
  name='Config',
  full_name='qrl.Transaction.ProposalCreate.Config',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='changes_bitfield', full_name='qrl.Transaction.ProposalCreate.Config.changes_bitfield', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='reorg_limit', full_name='qrl.Transaction.ProposalCreate.Config.reorg_limit', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_coin_supply', full_name='qrl.Transaction.ProposalCreate.Config.max_coin_supply', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='complete_emission_time_span_in_years', full_name='qrl.Transaction.ProposalCreate.Config.complete_emission_time_span_in_years', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mining_nonce_offset', full_name='qrl.Transaction.ProposalCreate.Config.mining_nonce_offset', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='extra_nonce_offset', full_name='qrl.Transaction.ProposalCreate.Config.extra_nonce_offset', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mining_blob_size_in_bytes', full_name='qrl.Transaction.ProposalCreate.Config.mining_blob_size_in_bytes', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_timing_in_seconds', full_name='qrl.Transaction.ProposalCreate.Config.block_timing_in_seconds', index=7,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='number_of_blocks_analyze', full_name='qrl.Transaction.ProposalCreate.Config.number_of_blocks_analyze', index=8,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_size_multiplier', full_name='qrl.Transaction.ProposalCreate.Config.block_size_multiplier', index=9,
      number=10, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_min_size_limit_in_bytes', full_name='qrl.Transaction.ProposalCreate.Config.block_min_size_limit_in_bytes', index=10,
      number=11, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transaction_multi_output_limit', full_name='qrl.Transaction.ProposalCreate.Config.transaction_multi_output_limit', index=11,
      number=12, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message_max_length', full_name='qrl.Transaction.ProposalCreate.Config.message_max_length', index=12,
      number=13, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='token_symbol_max_length', full_name='qrl.Transaction.ProposalCreate.Config.token_symbol_max_length', index=13,
      number=14, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='token_name_max_length', full_name='qrl.Transaction.ProposalCreate.Config.token_name_max_length', index=14,
      number=15, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lattice_pk1_max_length', full_name='qrl.Transaction.ProposalCreate.Config.lattice_pk1_max_length', index=15,
      number=16, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lattice_pk2_max_length', full_name='qrl.Transaction.ProposalCreate.Config.lattice_pk2_max_length', index=16,
      number=17, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lattice_pk3_max_length', full_name='qrl.Transaction.ProposalCreate.Config.lattice_pk3_max_length', index=17,
      number=18, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='foundation_multi_sig_address_threshold_percentage', full_name='qrl.Transaction.ProposalCreate.Config.foundation_multi_sig_address_threshold_percentage', index=18,
      number=19, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proposal_threshold_per', full_name='qrl.Transaction.ProposalCreate.Config.proposal_threshold_per', index=19,
      number=20, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proposal_default_options', full_name='qrl.Transaction.ProposalCreate.Config.proposal_default_options', index=20,
      number=21, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description_max_length', full_name='qrl.Transaction.ProposalCreate.Config.description_max_length', index=21,
      number=22, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='options_max_number', full_name='qrl.Transaction.ProposalCreate.Config.options_max_number', index=22,
      number=23, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='option_max_text_length', full_name='qrl.Transaction.ProposalCreate.Config.option_max_text_length', index=23,
      number=24, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proposal_config_activation_delay', full_name='qrl.Transaction.ProposalCreate.Config.proposal_config_activation_delay', index=24,
      number=25, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='N_measurement', full_name='qrl.Transaction.ProposalCreate.Config.N_measurement', index=25,
      number=26, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kp', full_name='qrl.Transaction.ProposalCreate.Config.kp', index=26,
      number=27, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11068,
  serialized_end=11945,
)

_TRANSACTION_PROPOSALCREATE_OTHER = _descriptor.Descriptor(
  name='Other',
  full_name='qrl.Transaction.ProposalCreate.Other',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='options', full_name='qrl.Transaction.ProposalCreate.Other.options', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11947,
  serialized_end=11971,
)

_TRANSACTION_PROPOSALCREATE = _descriptor.Descriptor(
  name='ProposalCreate',
  full_name='qrl.Transaction.ProposalCreate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='expiry_block_number', full_name='qrl.Transaction.ProposalCreate.expiry_block_number', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='qrl.Transaction.ProposalCreate.description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='qip', full_name='qrl.Transaction.ProposalCreate.qip', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='config', full_name='qrl.Transaction.ProposalCreate.config', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='other', full_name='qrl.Transaction.ProposalCreate.other', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_PROPOSALCREATE_QIP, _TRANSACTION_PROPOSALCREATE_CONFIG, _TRANSACTION_PROPOSALCREATE_OTHER, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='proposalType', full_name='qrl.Transaction.ProposalCreate.proposalType',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=10808,
  serialized_end=11987,
)

_TRANSACTION_PROPOSALVOTE = _descriptor.Descriptor(
  name='ProposalVote',
  full_name='qrl.Transaction.ProposalVote',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='shared_key', full_name='qrl.Transaction.ProposalVote.shared_key', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='option', full_name='qrl.Transaction.ProposalVote.option', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11989,
  serialized_end=12039,
)

_TRANSACTION = _descriptor.Descriptor(
  name='Transaction',
  full_name='qrl.Transaction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='master_addr', full_name='qrl.Transaction.master_addr', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fee', full_name='qrl.Transaction.fee', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='public_key', full_name='qrl.Transaction.public_key', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='signature', full_name='qrl.Transaction.signature', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='nonce', full_name='qrl.Transaction.nonce', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transaction_hash', full_name='qrl.Transaction.transaction_hash', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transfer', full_name='qrl.Transaction.transfer', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='coinbase', full_name='qrl.Transaction.coinbase', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='latticePK', full_name='qrl.Transaction.latticePK', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='qrl.Transaction.message', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='token', full_name='qrl.Transaction.token', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transfer_token', full_name='qrl.Transaction.transfer_token', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='slave', full_name='qrl.Transaction.slave', index=12,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='multi_sig_create', full_name='qrl.Transaction.multi_sig_create', index=13,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='multi_sig_spend', full_name='qrl.Transaction.multi_sig_spend', index=14,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='multi_sig_vote', full_name='qrl.Transaction.multi_sig_vote', index=15,
      number=16, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proposal_create', full_name='qrl.Transaction.proposal_create', index=16,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proposal_vote', full_name='qrl.Transaction.proposal_vote', index=17,
      number=18, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_TRANSACTION_TRANSFER, _TRANSACTION_COINBASE, _TRANSACTION_LATTICEPUBLICKEY, _TRANSACTION_MESSAGE, _TRANSACTION_TOKEN, _TRANSACTION_TRANSFERTOKEN, _TRANSACTION_SLAVE, _TRANSACTION_MULTISIGCREATE, _TRANSACTION_MULTISIGSPEND, _TRANSACTION_MULTISIGVOTE, _TRANSACTION_PROPOSALCREATE, _TRANSACTION_PROPOSALVOTE, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='transactionType', full_name='qrl.Transaction.transactionType',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=9328,
  serialized_end=12058,
)


_VOTESTATS = _descriptor.Descriptor(
  name='VoteStats',
  full_name='qrl.VoteStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='multi_sig_address', full_name='qrl.VoteStats.multi_sig_address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shared_key', full_name='qrl.VoteStats.shared_key', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='signatories', full_name='qrl.VoteStats.signatories', index=2,
      number=3, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tx_hashes', full_name='qrl.VoteStats.tx_hashes', index=3,
      number=4, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='unvotes', full_name='qrl.VoteStats.unvotes', index=4,
      number=5, type=8, cpp_type=7, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='expiry_block_number', full_name='qrl.VoteStats.expiry_block_number', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='total_weight', full_name='qrl.VoteStats.total_weight', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='executed', full_name='qrl.VoteStats.executed', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12061,
  serialized_end=12245,
)


_PROPOSALVOTESTATS = _descriptor.Descriptor(
  name='ProposalVoteStats',
  full_name='qrl.ProposalVoteStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='addr_from', full_name='qrl.ProposalVoteStats.addr_from', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shared_key', full_name='qrl.ProposalVoteStats.shared_key', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proposal_type', full_name='qrl.ProposalVoteStats.proposal_type', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='weight_by_option', full_name='qrl.ProposalVoteStats.weight_by_option', index=3,
      number=4, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='expiry_block_number', full_name='qrl.ProposalVoteStats.expiry_block_number', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='executed', full_name='qrl.ProposalVoteStats.executed', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='number_of_tx_hashes', full_name='qrl.ProposalVoteStats.number_of_tx_hashes', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12248,
  serialized_end=12431,
)


_PROPOSALRECORD = _descriptor.Descriptor(
  name='ProposalRecord',
  full_name='qrl.ProposalRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='number_of_tx_hashes', full_name='qrl.ProposalRecord.number_of_tx_hashes', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12433,
  serialized_end=12478,
)


_TOKENLIST = _descriptor.Descriptor(
  name='TokenList',
  full_name='qrl.TokenList',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='token_txhash', full_name='qrl.TokenList.token_txhash', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12480,
  serialized_end=12513,
)


_TOKENBALANCE = _descriptor.Descriptor(
  name='TokenBalance',
  full_name='qrl.TokenBalance',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='balance', full_name='qrl.TokenBalance.balance', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='decimals', full_name='qrl.TokenBalance.decimals', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tx_hash', full_name='qrl.TokenBalance.tx_hash', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='delete', full_name='qrl.TokenBalance.delete', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12515,
  serialized_end=12597,
)


_SLAVEMETADATA = _descriptor.Descriptor(
  name='SlaveMetadata',
  full_name='qrl.SlaveMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='access_type', full_name='qrl.SlaveMetadata.access_type', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tx_hash', full_name='qrl.SlaveMetadata.tx_hash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='delete', full_name='qrl.SlaveMetadata.delete', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12599,
  serialized_end=12668,
)


_LATTICEPKMETADATA = _descriptor.Descriptor(
  name='LatticePKMetadata',
  full_name='qrl.LatticePKMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='enabled', full_name='qrl.LatticePKMetadata.enabled', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tx_hash', full_name='qrl.LatticePKMetadata.tx_hash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='delete', full_name='qrl.LatticePKMetadata.delete', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12670,
  serialized_end=12739,
)


_TOKENMETADATA = _descriptor.Descriptor(
  name='TokenMetadata',
  full_name='qrl.TokenMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='token_txhash', full_name='qrl.TokenMetadata.token_txhash', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transfer_token_tx_hashes', full_name='qrl.TokenMetadata.transfer_token_tx_hashes', index=1,
      number=2, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12741,
  serialized_end=12812,
)


_ENCRYPTEDEPHEMERALMESSAGE_CHANNEL = _descriptor.Descriptor(
  name='Channel',
  full_name='qrl.EncryptedEphemeralMessage.Channel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='enc_aes256_symkey', full_name='qrl.EncryptedEphemeralMessage.Channel.enc_aes256_symkey', index=0,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12975,
  serialized_end=13011,
)

_ENCRYPTEDEPHEMERALMESSAGE = _descriptor.Descriptor(
  name='EncryptedEphemeralMessage',
  full_name='qrl.EncryptedEphemeralMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='msg_id', full_name='qrl.EncryptedEphemeralMessage.msg_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ttl', full_name='qrl.EncryptedEphemeralMessage.ttl', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ttr', full_name='qrl.EncryptedEphemeralMessage.ttr', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='channel', full_name='qrl.EncryptedEphemeralMessage.channel', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='nonce', full_name='qrl.EncryptedEphemeralMessage.nonce', index=4,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='payload', full_name='qrl.EncryptedEphemeralMessage.payload', index=5,
      number=7, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_ENCRYPTEDEPHEMERALMESSAGE_CHANNEL, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12815,
  serialized_end=13011,
)


_ADDRESSLIST = _descriptor.Descriptor(
  name='AddressList',
  full_name='qrl.AddressList',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='addresses', full_name='qrl.AddressList.addresses', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13013,
  serialized_end=13045,
)


_BLOCKHEIGHTDATA = _descriptor.Descriptor(
  name='BlockHeightData',
  full_name='qrl.BlockHeightData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='block_number', full_name='qrl.BlockHeightData.block_number', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_headerhash', full_name='qrl.BlockHeightData.block_headerhash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cumulative_difficulty', full_name='qrl.BlockHeightData.cumulative_difficulty', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13047,
  serialized_end=13143,
)


_BLOCKMETADATA = _descriptor.Descriptor(
  name='BlockMetaData',
  full_name='qrl.BlockMetaData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='block_difficulty', full_name='qrl.BlockMetaData.block_difficulty', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cumulative_difficulty', full_name='qrl.BlockMetaData.cumulative_difficulty', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='child_headerhashes', full_name='qrl.BlockMetaData.child_headerhashes', index=2,
      number=3, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='last_N_headerhashes', full_name='qrl.BlockMetaData.last_N_headerhashes', index=3,
      number=4, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13146,
  serialized_end=13275,
)


_BLOCKNUMBERMAPPING = _descriptor.Descriptor(
  name='BlockNumberMapping',
  full_name='qrl.BlockNumberMapping',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='headerhash', full_name='qrl.BlockNumberMapping.headerhash', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='prev_headerhash', full_name='qrl.BlockNumberMapping.prev_headerhash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13277,
  serialized_end=13342,
)


_PEERSTAT = _descriptor.Descriptor(
  name='PeerStat',
  full_name='qrl.PeerStat',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='peer_ip', full_name='qrl.PeerStat.peer_ip', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='port', full_name='qrl.PeerStat.port', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='node_chain_state', full_name='qrl.PeerStat.node_chain_state', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13344,
  serialized_end=13432,
)


_NODECHAINSTATE = _descriptor.Descriptor(
  name='NodeChainState',
  full_name='qrl.NodeChainState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='block_number', full_name='qrl.NodeChainState.block_number', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='header_hash', full_name='qrl.NodeChainState.header_hash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cumulative_difficulty', full_name='qrl.NodeChainState.cumulative_difficulty', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='qrl.NodeChainState.version', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='qrl.NodeChainState.timestamp', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13434,
  serialized_end=13560,
)


_NODEHEADERHASH = _descriptor.Descriptor(
  name='NodeHeaderHash',
  full_name='qrl.NodeHeaderHash',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='block_number', full_name='qrl.NodeHeaderHash.block_number', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='headerhashes', full_name='qrl.NodeHeaderHash.headerhashes', index=1,
      number=2, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13562,
  serialized_end=13622,
)


_P2PACKNOWLEDGEMENT = _descriptor.Descriptor(
  name='P2PAcknowledgement',
  full_name='qrl.P2PAcknowledgement',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='bytes_processed', full_name='qrl.P2PAcknowledgement.bytes_processed', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13624,
  serialized_end=13669,
)


_PEERINFO = _descriptor.Descriptor(
  name='PeerInfo',
  full_name='qrl.PeerInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='peer_ip', full_name='qrl.PeerInfo.peer_ip', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='port', full_name='qrl.PeerInfo.port', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='banned_timestamp', full_name='qrl.PeerInfo.banned_timestamp', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='credibility', full_name='qrl.PeerInfo.credibility', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='last_connections_timestamp', full_name='qrl.PeerInfo.last_connections_timestamp', index=4,
      number=5, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13671,
  serialized_end=13795,
)


_PEERS = _descriptor.Descriptor(
  name='Peers',
  full_name='qrl.Peers',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='peer_info_list', full_name='qrl.Peers.peer_info_list', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13797,
  serialized_end=13843,
)


_DEVCONFIG_CHAIN = _descriptor.Descriptor(
  name='Chain',
  full_name='qrl.DevConfig.Chain',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='reorg_limit', full_name='qrl.DevConfig.Chain.reorg_limit', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_coin_supply', full_name='qrl.DevConfig.Chain.max_coin_supply', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='complete_emission_time_span_in_years', full_name='qrl.DevConfig.Chain.complete_emission_time_span_in_years', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14131,
  serialized_end=14230,
)

_DEVCONFIG_BLOCK_BLOCKSIZECONTROLLER = _descriptor.Descriptor(
  name='BlockSizeController',
  full_name='qrl.DevConfig.Block.BlockSizeController',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='number_of_blocks_analyze', full_name='qrl.DevConfig.Block.BlockSizeController.number_of_blocks_analyze', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='size_multiplier', full_name='qrl.DevConfig.Block.BlockSizeController.size_multiplier', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_min_size_limit_in_bytes', full_name='qrl.DevConfig.Block.BlockSizeController.block_min_size_limit_in_bytes', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14440,
  serialized_end=14559,
)

_DEVCONFIG_BLOCK = _descriptor.Descriptor(
  name='Block',
  full_name='qrl.DevConfig.Block',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='mining_nonce_offset', full_name='qrl.DevConfig.Block.mining_nonce_offset', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='extra_nonce_offset', full_name='qrl.DevConfig.Block.extra_nonce_offset', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mining_blob_size_in_bytes', full_name='qrl.DevConfig.Block.mining_blob_size_in_bytes', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_timing_in_seconds', full_name='qrl.DevConfig.Block.block_timing_in_seconds', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_size_controller', full_name='qrl.DevConfig.Block.block_size_controller', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_DEVCONFIG_BLOCK_BLOCKSIZECONTROLLER, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14233,
  serialized_end=14559,
)

_DEVCONFIG_TRANSACTION_MESSAGE = _descriptor.Descriptor(
  name='Message',
  full_name='qrl.DevConfig.Transaction.Message',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='max_length', full_name='qrl.DevConfig.Transaction.Message.max_length', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14941,
  serialized_end=14970,
)

_DEVCONFIG_TRANSACTION_SLAVE = _descriptor.Descriptor(
  name='Slave',
  full_name='qrl.DevConfig.Transaction.Slave',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='slave_pk_max_length', full_name='qrl.DevConfig.Transaction.Slave.slave_pk_max_length', index=0,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14972,
  serialized_end=15008,
)

_DEVCONFIG_TRANSACTION_TOKEN = _descriptor.Descriptor(
  name='Token',
  full_name='qrl.DevConfig.Transaction.Token',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='symbol_max_length', full_name='qrl.DevConfig.Transaction.Token.symbol_max_length', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name_max_length', full_name='qrl.DevConfig.Transaction.Token.name_max_length', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15010,
  serialized_end=15069,
)

_DEVCONFIG_TRANSACTION_LATTICE = _descriptor.Descriptor(
  name='Lattice',
  full_name='qrl.DevConfig.Transaction.Lattice',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pk1_max_length', full_name='qrl.DevConfig.Transaction.Lattice.pk1_max_length', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pk2_max_length', full_name='qrl.DevConfig.Transaction.Lattice.pk2_max_length', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pk3_max_length', full_name='qrl.DevConfig.Transaction.Lattice.pk3_max_length', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15071,
  serialized_end=15152,
)

_DEVCONFIG_TRANSACTION_FOUNDATIONMULTISIG = _descriptor.Descriptor(
  name='FoundationMultiSig',
  full_name='qrl.DevConfig.Transaction.FoundationMultiSig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='threshold_percentage', full_name='qrl.DevConfig.Transaction.FoundationMultiSig.threshold_percentage', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15154,
  serialized_end=15204,
)

_DEVCONFIG_TRANSACTION_PROPOSAL = _descriptor.Descriptor(
  name='Proposal',
  full_name='qrl.DevConfig.Transaction.Proposal',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='threshold_per', full_name='qrl.DevConfig.Transaction.Proposal.threshold_per', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='default_options', full_name='qrl.DevConfig.Transaction.Proposal.default_options', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description_max_length', full_name='qrl.DevConfig.Transaction.Proposal.description_max_length', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='options_max_number', full_name='qrl.DevConfig.Transaction.Proposal.options_max_number', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='option_max_text_length', full_name='qrl.DevConfig.Transaction.Proposal.option_max_text_length', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proposal_config_activation_delay', full_name='qrl.DevConfig.Transaction.Proposal.proposal_config_activation_delay', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15207,
  serialized_end=15399,
)

_DEVCONFIG_TRANSACTION = _descriptor.Descriptor(
  name='Transaction',
  full_name='qrl.DevConfig.Transaction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='multi_output_limit', full_name='qrl.DevConfig.Transaction.multi_output_limit', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='qrl.DevConfig.Transaction.message', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='slave', full_name='qrl.DevConfig.Transaction.slave', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='token', full_name='qrl.DevConfig.Transaction.token', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lattice', full_name='qrl.DevConfig.Transaction.lattice', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='foundation_multi_sig', full_name='qrl.DevConfig.Transaction.foundation_multi_sig', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proposal', full_name='qrl.DevConfig.Transaction.proposal', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_DEVCONFIG_TRANSACTION_MESSAGE, _DEVCONFIG_TRANSACTION_SLAVE, _DEVCONFIG_TRANSACTION_TOKEN, _DEVCONFIG_TRANSACTION_LATTICE, _DEVCONFIG_TRANSACTION_FOUNDATIONMULTISIG, _DEVCONFIG_TRANSACTION_PROPOSAL, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14562,
  serialized_end=15399,
)

_DEVCONFIG_POW = _descriptor.Descriptor(
  name='POW',
  full_name='qrl.DevConfig.POW',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='N_measurement', full_name='qrl.DevConfig.POW.N_measurement', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kp', full_name='qrl.DevConfig.POW.kp', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15401,
  serialized_end=15441,
)

_DEVCONFIG = _descriptor.Descriptor(
  name='DevConfig',
  full_name='qrl.DevConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='prev_state_key', full_name='qrl.DevConfig.prev_state_key', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='current_state_key', full_name='qrl.DevConfig.current_state_key', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='activation_header_hash', full_name='qrl.DevConfig.activation_header_hash', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='activation_block_number', full_name='qrl.DevConfig.activation_block_number', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='chain', full_name='qrl.DevConfig.chain', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block', full_name='qrl.DevConfig.block', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transaction', full_name='qrl.DevConfig.transaction', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pow', full_name='qrl.DevConfig.pow', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_DEVCONFIG_CHAIN, _DEVCONFIG_BLOCK, _DEVCONFIG_TRANSACTION, _DEVCONFIG_POW, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13846,
  serialized_end=15441,
)

_GETNODESTATERESP.fields_by_name['info'].message_type = _NODEINFO
_GETKNOWNPEERSRESP.fields_by_name['node_info'].message_type = _NODEINFO
_GETKNOWNPEERSRESP.fields_by_name['known_peers'].message_type = _PEER
_GETPEERSSTATRESP.fields_by_name['peers_stat'].message_type = _PEERSTAT
_GETSTATSRESP.fields_by_name['node_info'].message_type = _NODEINFO
_GETSTATSRESP.fields_by_name['block_timeseries'].message_type = _BLOCKDATAPOINT
_GETADDRESSSTATERESP.fields_by_name['state'].message_type = _ADDRESSSTATE
_GETOPTIMIZEDADDRESSSTATERESP.fields_by_name['state'].message_type = _OPTIMIZEDADDRESSSTATE
_GETMULTISIGADDRESSSTATERESP.fields_by_name['state'].message_type = _MULTISIGADDRESSSTATE
_PARSEADDRESSRESP.fields_by_name['desc'].message_type = _ADDRESSDESCRIPTOR
_GETOBJECTRESP.fields_by_name['address_state'].message_type = _OPTIMIZEDADDRESSSTATE
_GETOBJECTRESP.fields_by_name['transaction'].message_type = _TRANSACTIONEXTENDED
_GETOBJECTRESP.fields_by_name['block_extended'].message_type = _BLOCKEXTENDED
_GETOBJECTRESP.oneofs_by_name['result'].fields.append(
  _GETOBJECTRESP.fields_by_name['address_state'])
_GETOBJECTRESP.fields_by_name['address_state'].containing_oneof = _GETOBJECTRESP.oneofs_by_name['result']
_GETOBJECTRESP.oneofs_by_name['result'].fields.append(
  _GETOBJECTRESP.fields_by_name['transaction'])
_GETOBJECTRESP.fields_by_name['transaction'].containing_oneof = _GETOBJECTRESP.oneofs_by_name['result']
_GETOBJECTRESP.oneofs_by_name['result'].fields.append(
  _GETOBJECTRESP.fields_by_name['block_extended'])
_GETOBJECTRESP.fields_by_name['block_extended'].containing_oneof = _GETOBJECTRESP.oneofs_by_name['result']
_GETLATESTDATAREQ.fields_by_name['filter'].enum_type = _GETLATESTDATAREQ_FILTER
_GETLATESTDATAREQ_FILTER.containing_type = _GETLATESTDATAREQ
_GETLATESTDATARESP.fields_by_name['blockheaders'].message_type = _BLOCKHEADEREXTENDED
_GETLATESTDATARESP.fields_by_name['transactions'].message_type = _TRANSACTIONEXTENDED
_GETLATESTDATARESP.fields_by_name['transactions_unconfirmed'].message_type = _TRANSACTIONEXTENDED
_TRANSFERCOINSRESP.fields_by_name['extended_transaction_unsigned'].message_type = _TRANSACTIONEXTENDED
_PUSHTRANSACTIONREQ.fields_by_name['transaction_signed'].message_type = _TRANSACTION
_PUSHTRANSACTIONRESP.fields_by_name['error_code'].enum_type = _PUSHTRANSACTIONRESP_RESPONSECODE
_PUSHTRANSACTIONRESP_RESPONSECODE.containing_type = _PUSHTRANSACTIONRESP
_TOKENTXNREQ.fields_by_name['initial_balances'].message_type = _ADDRESSAMOUNT
_GETTRANSACTIONRESP.fields_by_name['tx'].message_type = _TRANSACTION
_GETMINITRANSACTIONSBYADDRESSRESP.fields_by_name['mini_transactions'].message_type = _MINITRANSACTION
_GETTRANSACTIONSBYADDRESSRESP.fields_by_name['transactions_detail'].message_type = _GETTRANSACTIONRESP
_GETMULTISIGSPENDTXSBYADDRESSREQ.fields_by_name['filter_type'].enum_type = _GETMULTISIGSPENDTXSBYADDRESSREQ_FILTERTYPE
_GETMULTISIGSPENDTXSBYADDRESSREQ_FILTERTYPE.containing_type = _GETMULTISIGSPENDTXSBYADDRESSREQ
_GETMULTISIGSPENDTXSBYADDRESSRESP.fields_by_name['transactions_detail'].message_type = _GETTRANSACTIONRESP
_GETVOTESTATSRESP.fields_by_name['vote_stats'].message_type = _VOTESTATS
_GETINBOXMESSAGESBYADDRESSRESP.fields_by_name['transactions_detail'].message_type = _GETTRANSACTIONRESP
_GETTOKENSBYADDRESSRESP.fields_by_name['tokens_detail'].message_type = _TOKENDETAIL
_GETSLAVESBYADDRESSRESP.fields_by_name['slaves_detail'].message_type = _SLAVEDETAIL
_GETLATTICEPKSBYADDRESSRESP.fields_by_name['lattice_pks_detail'].message_type = _LATTICEPKSDETAIL
_GETMULTISIGADDRESSESBYADDRESSRESP.fields_by_name['multi_sig_detail'].message_type = _MULTISIGDETAIL
_GETOTSRESP.fields_by_name['ots_bitfield_by_page'].message_type = _OTSBITFIELDBYPAGE
_GETBLOCKRESP.fields_by_name['block'].message_type = _BLOCK
_GETBLOCKBYNUMBERRESP.fields_by_name['block'].message_type = _BLOCK
_NODEINFO.fields_by_name['state'].enum_type = _NODEINFO_STATE
_NODEINFO_STATE.containing_type = _NODEINFO
_STOREDPEERS.fields_by_name['peers'].message_type = _PEER
_ADDRESSSTATE_TOKENSENTRY.containing_type = _ADDRESSSTATE
_ADDRESSSTATE_SLAVEPKSACCESSTYPEENTRY.containing_type = _ADDRESSSTATE
_ADDRESSSTATE.fields_by_name['tokens'].message_type = _ADDRESSSTATE_TOKENSENTRY
_ADDRESSSTATE.fields_by_name['latticePK_list'].message_type = _LATTICEPK
_ADDRESSSTATE.fields_by_name['slave_pks_access_type'].message_type = _ADDRESSSTATE_SLAVEPKSACCESSTYPEENTRY
_OPTIMIZEDADDRESSSTATE.fields_by_name['proposal_vote_stats'].message_type = _TRANSACTION
_MULTISIGADDRESSSTATE.fields_by_name['proposal_vote_stats'].message_type = _TRANSACTION
_BLOCKHEADEREXTENDED.fields_by_name['header'].message_type = _BLOCKHEADER
_BLOCKHEADEREXTENDED.fields_by_name['transaction_count'].message_type = _TRANSACTIONCOUNT
_TRANSACTIONCOUNT_COUNTENTRY.containing_type = _TRANSACTIONCOUNT
_TRANSACTIONCOUNT.fields_by_name['count'].message_type = _TRANSACTIONCOUNT_COUNTENTRY
_TRANSACTIONEXTENDED.fields_by_name['header'].message_type = _BLOCKHEADER
_TRANSACTIONEXTENDED.fields_by_name['tx'].message_type = _TRANSACTION
_BLOCKEXTENDED.fields_by_name['header'].message_type = _BLOCKHEADER
_BLOCKEXTENDED.fields_by_name['extended_transactions'].message_type = _TRANSACTIONEXTENDED
_BLOCKEXTENDED.fields_by_name['genesis_balance'].message_type = _GENESISBALANCE
_BLOCK.fields_by_name['header'].message_type = _BLOCKHEADER
_BLOCK.fields_by_name['transactions'].message_type = _TRANSACTION
_BLOCK.fields_by_name['genesis_balance'].message_type = _GENESISBALANCE
_BLOCKMETADATALIST.fields_by_name['block_number_hashes'].message_type = _BLOCKMETADATA
_TRANSACTION_TRANSFER.containing_type = _TRANSACTION
_TRANSACTION_COINBASE.containing_type = _TRANSACTION
_TRANSACTION_LATTICEPUBLICKEY.containing_type = _TRANSACTION
_TRANSACTION_MESSAGE.containing_type = _TRANSACTION
_TRANSACTION_TOKEN.fields_by_name['initial_balances'].message_type = _ADDRESSAMOUNT
_TRANSACTION_TOKEN.containing_type = _TRANSACTION
_TRANSACTION_TRANSFERTOKEN.containing_type = _TRANSACTION
_TRANSACTION_SLAVE.containing_type = _TRANSACTION
_TRANSACTION_MULTISIGCREATE.containing_type = _TRANSACTION
_TRANSACTION_MULTISIGSPEND.containing_type = _TRANSACTION
_TRANSACTION_MULTISIGVOTE.containing_type = _TRANSACTION
_TRANSACTION_PROPOSALCREATE_QIP.containing_type = _TRANSACTION_PROPOSALCREATE
_TRANSACTION_PROPOSALCREATE_CONFIG.containing_type = _TRANSACTION_PROPOSALCREATE
_TRANSACTION_PROPOSALCREATE_OTHER.containing_type = _TRANSACTION_PROPOSALCREATE
_TRANSACTION_PROPOSALCREATE.fields_by_name['qip'].message_type = _TRANSACTION_PROPOSALCREATE_QIP
_TRANSACTION_PROPOSALCREATE.fields_by_name['config'].message_type = _TRANSACTION_PROPOSALCREATE_CONFIG
_TRANSACTION_PROPOSALCREATE.fields_by_name['other'].message_type = _TRANSACTION_PROPOSALCREATE_OTHER
_TRANSACTION_PROPOSALCREATE.containing_type = _TRANSACTION
_TRANSACTION_PROPOSALCREATE.oneofs_by_name['proposalType'].fields.append(
  _TRANSACTION_PROPOSALCREATE.fields_by_name['qip'])
_TRANSACTION_PROPOSALCREATE.fields_by_name['qip'].containing_oneof = _TRANSACTION_PROPOSALCREATE.oneofs_by_name['proposalType']
_TRANSACTION_PROPOSALCREATE.oneofs_by_name['proposalType'].fields.append(
  _TRANSACTION_PROPOSALCREATE.fields_by_name['config'])
_TRANSACTION_PROPOSALCREATE.fields_by_name['config'].containing_oneof = _TRANSACTION_PROPOSALCREATE.oneofs_by_name['proposalType']
_TRANSACTION_PROPOSALCREATE.oneofs_by_name['proposalType'].fields.append(
  _TRANSACTION_PROPOSALCREATE.fields_by_name['other'])
_TRANSACTION_PROPOSALCREATE.fields_by_name['other'].containing_oneof = _TRANSACTION_PROPOSALCREATE.oneofs_by_name['proposalType']
_TRANSACTION_PROPOSALVOTE.containing_type = _TRANSACTION
_TRANSACTION.fields_by_name['transfer'].message_type = _TRANSACTION_TRANSFER
_TRANSACTION.fields_by_name['coinbase'].message_type = _TRANSACTION_COINBASE
_TRANSACTION.fields_by_name['latticePK'].message_type = _TRANSACTION_LATTICEPUBLICKEY
_TRANSACTION.fields_by_name['message'].message_type = _TRANSACTION_MESSAGE
_TRANSACTION.fields_by_name['token'].message_type = _TRANSACTION_TOKEN
_TRANSACTION.fields_by_name['transfer_token'].message_type = _TRANSACTION_TRANSFERTOKEN
_TRANSACTION.fields_by_name['slave'].message_type = _TRANSACTION_SLAVE
_TRANSACTION.fields_by_name['multi_sig_create'].message_type = _TRANSACTION_MULTISIGCREATE
_TRANSACTION.fields_by_name['multi_sig_spend'].message_type = _TRANSACTION_MULTISIGSPEND
_TRANSACTION.fields_by_name['multi_sig_vote'].message_type = _TRANSACTION_MULTISIGVOTE
_TRANSACTION.fields_by_name['proposal_create'].message_type = _TRANSACTION_PROPOSALCREATE
_TRANSACTION.fields_by_name['proposal_vote'].message_type = _TRANSACTION_PROPOSALVOTE
_TRANSACTION.oneofs_by_name['transactionType'].fields.append(
  _TRANSACTION.fields_by_name['transfer'])
_TRANSACTION.fields_by_name['transfer'].containing_oneof = _TRANSACTION.oneofs_by_name['transactionType']
_TRANSACTION.oneofs_by_name['transactionType'].fields.append(
  _TRANSACTION.fields_by_name['coinbase'])
_TRANSACTION.fields_by_name['coinbase'].containing_oneof = _TRANSACTION.oneofs_by_name['transactionType']
_TRANSACTION.oneofs_by_name['transactionType'].fields.append(
  _TRANSACTION.fields_by_name['latticePK'])
_TRANSACTION.fields_by_name['latticePK'].containing_oneof = _TRANSACTION.oneofs_by_name['transactionType']
_TRANSACTION.oneofs_by_name['transactionType'].fields.append(
  _TRANSACTION.fields_by_name['message'])
_TRANSACTION.fields_by_name['message'].containing_oneof = _TRANSACTION.oneofs_by_name['transactionType']
_TRANSACTION.oneofs_by_name['transactionType'].fields.append(
  _TRANSACTION.fields_by_name['token'])
_TRANSACTION.fields_by_name['token'].containing_oneof = _TRANSACTION.oneofs_by_name['transactionType']
_TRANSACTION.oneofs_by_name['transactionType'].fields.append(
  _TRANSACTION.fields_by_name['transfer_token'])
_TRANSACTION.fields_by_name['transfer_token'].containing_oneof = _TRANSACTION.oneofs_by_name['transactionType']
_TRANSACTION.oneofs_by_name['transactionType'].fields.append(
  _TRANSACTION.fields_by_name['slave'])
_TRANSACTION.fields_by_name['slave'].containing_oneof = _TRANSACTION.oneofs_by_name['transactionType']
_TRANSACTION.oneofs_by_name['transactionType'].fields.append(
  _TRANSACTION.fields_by_name['multi_sig_create'])
_TRANSACTION.fields_by_name['multi_sig_create'].containing_oneof = _TRANSACTION.oneofs_by_name['transactionType']
_TRANSACTION.oneofs_by_name['transactionType'].fields.append(
  _TRANSACTION.fields_by_name['multi_sig_spend'])
_TRANSACTION.fields_by_name['multi_sig_spend'].containing_oneof = _TRANSACTION.oneofs_by_name['transactionType']
_TRANSACTION.oneofs_by_name['transactionType'].fields.append(
  _TRANSACTION.fields_by_name['multi_sig_vote'])
_TRANSACTION.fields_by_name['multi_sig_vote'].containing_oneof = _TRANSACTION.oneofs_by_name['transactionType']
_TRANSACTION.oneofs_by_name['transactionType'].fields.append(
  _TRANSACTION.fields_by_name['proposal_create'])
_TRANSACTION.fields_by_name['proposal_create'].containing_oneof = _TRANSACTION.oneofs_by_name['transactionType']
_TRANSACTION.oneofs_by_name['transactionType'].fields.append(
  _TRANSACTION.fields_by_name['proposal_vote'])
_TRANSACTION.fields_by_name['proposal_vote'].containing_oneof = _TRANSACTION.oneofs_by_name['transactionType']
_ENCRYPTEDEPHEMERALMESSAGE_CHANNEL.containing_type = _ENCRYPTEDEPHEMERALMESSAGE
_ENCRYPTEDEPHEMERALMESSAGE.fields_by_name['channel'].message_type = _ENCRYPTEDEPHEMERALMESSAGE_CHANNEL
_PEERSTAT.fields_by_name['node_chain_state'].message_type = _NODECHAINSTATE
_PEERS.fields_by_name['peer_info_list'].message_type = _PEERINFO
_DEVCONFIG_CHAIN.containing_type = _DEVCONFIG
_DEVCONFIG_BLOCK_BLOCKSIZECONTROLLER.containing_type = _DEVCONFIG_BLOCK
_DEVCONFIG_BLOCK.fields_by_name['block_size_controller'].message_type = _DEVCONFIG_BLOCK_BLOCKSIZECONTROLLER
_DEVCONFIG_BLOCK.containing_type = _DEVCONFIG
_DEVCONFIG_TRANSACTION_MESSAGE.containing_type = _DEVCONFIG_TRANSACTION
_DEVCONFIG_TRANSACTION_SLAVE.containing_type = _DEVCONFIG_TRANSACTION
_DEVCONFIG_TRANSACTION_TOKEN.containing_type = _DEVCONFIG_TRANSACTION
_DEVCONFIG_TRANSACTION_LATTICE.containing_type = _DEVCONFIG_TRANSACTION
_DEVCONFIG_TRANSACTION_FOUNDATIONMULTISIG.containing_type = _DEVCONFIG_TRANSACTION
_DEVCONFIG_TRANSACTION_PROPOSAL.containing_type = _DEVCONFIG_TRANSACTION
_DEVCONFIG_TRANSACTION.fields_by_name['message'].message_type = _DEVCONFIG_TRANSACTION_MESSAGE
_DEVCONFIG_TRANSACTION.fields_by_name['slave'].message_type = _DEVCONFIG_TRANSACTION_SLAVE
_DEVCONFIG_TRANSACTION.fields_by_name['token'].message_type = _DEVCONFIG_TRANSACTION_TOKEN
_DEVCONFIG_TRANSACTION.fields_by_name['lattice'].message_type = _DEVCONFIG_TRANSACTION_LATTICE
_DEVCONFIG_TRANSACTION.fields_by_name['foundation_multi_sig'].message_type = _DEVCONFIG_TRANSACTION_FOUNDATIONMULTISIG
_DEVCONFIG_TRANSACTION.fields_by_name['proposal'].message_type = _DEVCONFIG_TRANSACTION_PROPOSAL
_DEVCONFIG_TRANSACTION.containing_type = _DEVCONFIG
_DEVCONFIG_POW.containing_type = _DEVCONFIG
_DEVCONFIG.fields_by_name['chain'].message_type = _DEVCONFIG_CHAIN
_DEVCONFIG.fields_by_name['block'].message_type = _DEVCONFIG_BLOCK
_DEVCONFIG.fields_by_name['transaction'].message_type = _DEVCONFIG_TRANSACTION
_DEVCONFIG.fields_by_name['pow'].message_type = _DEVCONFIG_POW
DESCRIPTOR.message_types_by_name['Empty'] = _EMPTY
DESCRIPTOR.message_types_by_name['GetNodeStateReq'] = _GETNODESTATEREQ
DESCRIPTOR.message_types_by_name['GetNodeStateResp'] = _GETNODESTATERESP
DESCRIPTOR.message_types_by_name['GetKnownPeersReq'] = _GETKNOWNPEERSREQ
DESCRIPTOR.message_types_by_name['GetKnownPeersResp'] = _GETKNOWNPEERSRESP
DESCRIPTOR.message_types_by_name['GetPeersStatReq'] = _GETPEERSSTATREQ
DESCRIPTOR.message_types_by_name['GetPeersStatResp'] = _GETPEERSSTATRESP
DESCRIPTOR.message_types_by_name['GetChainStatsReq'] = _GETCHAINSTATSREQ
DESCRIPTOR.message_types_by_name['GetChainStatsResp'] = _GETCHAINSTATSRESP
DESCRIPTOR.message_types_by_name['GetStatsReq'] = _GETSTATSREQ
DESCRIPTOR.message_types_by_name['GetStatsResp'] = _GETSTATSRESP
DESCRIPTOR.message_types_by_name['GetAddressFromPKReq'] = _GETADDRESSFROMPKREQ
DESCRIPTOR.message_types_by_name['GetAddressFromPKResp'] = _GETADDRESSFROMPKRESP
DESCRIPTOR.message_types_by_name['BlockDataPoint'] = _BLOCKDATAPOINT
DESCRIPTOR.message_types_by_name['GetAddressStateReq'] = _GETADDRESSSTATEREQ
DESCRIPTOR.message_types_by_name['GetAddressStateResp'] = _GETADDRESSSTATERESP
DESCRIPTOR.message_types_by_name['GetOptimizedAddressStateResp'] = _GETOPTIMIZEDADDRESSSTATERESP
DESCRIPTOR.message_types_by_name['GetMultiSigAddressStateReq'] = _GETMULTISIGADDRESSSTATEREQ
DESCRIPTOR.message_types_by_name['GetMultiSigAddressStateResp'] = _GETMULTISIGADDRESSSTATERESP
DESCRIPTOR.message_types_by_name['IsSlaveReq'] = _ISSLAVEREQ
DESCRIPTOR.message_types_by_name['IsSlaveResp'] = _ISSLAVERESP
DESCRIPTOR.message_types_by_name['ParseAddressReq'] = _PARSEADDRESSREQ
DESCRIPTOR.message_types_by_name['ParseAddressResp'] = _PARSEADDRESSRESP
DESCRIPTOR.message_types_by_name['GetObjectReq'] = _GETOBJECTREQ
DESCRIPTOR.message_types_by_name['GetObjectResp'] = _GETOBJECTRESP
DESCRIPTOR.message_types_by_name['GetLatestDataReq'] = _GETLATESTDATAREQ
DESCRIPTOR.message_types_by_name['GetLatestDataResp'] = _GETLATESTDATARESP
DESCRIPTOR.message_types_by_name['TransferCoinsReq'] = _TRANSFERCOINSREQ
DESCRIPTOR.message_types_by_name['TransferCoinsResp'] = _TRANSFERCOINSRESP
DESCRIPTOR.message_types_by_name['PushTransactionReq'] = _PUSHTRANSACTIONREQ
DESCRIPTOR.message_types_by_name['PushTransactionResp'] = _PUSHTRANSACTIONRESP
DESCRIPTOR.message_types_by_name['MultiSigCreateTxnReq'] = _MULTISIGCREATETXNREQ
DESCRIPTOR.message_types_by_name['MultiSigSpendTxnReq'] = _MULTISIGSPENDTXNREQ
DESCRIPTOR.message_types_by_name['MultiSigVoteTxnReq'] = _MULTISIGVOTETXNREQ
DESCRIPTOR.message_types_by_name['MessageTxnReq'] = _MESSAGETXNREQ
DESCRIPTOR.message_types_by_name['TokenTxnReq'] = _TOKENTXNREQ
DESCRIPTOR.message_types_by_name['TransferTokenTxnReq'] = _TRANSFERTOKENTXNREQ
DESCRIPTOR.message_types_by_name['SlaveTxnReq'] = _SLAVETXNREQ
DESCRIPTOR.message_types_by_name['LatticeTxnReq'] = _LATTICETXNREQ
DESCRIPTOR.message_types_by_name['MiniTransaction'] = _MINITRANSACTION
DESCRIPTOR.message_types_by_name['GetTransactionReq'] = _GETTRANSACTIONREQ
DESCRIPTOR.message_types_by_name['GetTransactionResp'] = _GETTRANSACTIONRESP
DESCRIPTOR.message_types_by_name['GetMiniTransactionsByAddressReq'] = _GETMINITRANSACTIONSBYADDRESSREQ
DESCRIPTOR.message_types_by_name['GetMiniTransactionsByAddressResp'] = _GETMINITRANSACTIONSBYADDRESSRESP
DESCRIPTOR.message_types_by_name['GetTransactionsByAddressReq'] = _GETTRANSACTIONSBYADDRESSREQ
DESCRIPTOR.message_types_by_name['GetTransactionsByAddressResp'] = _GETTRANSACTIONSBYADDRESSRESP
DESCRIPTOR.message_types_by_name['GetMultiSigSpendTxsByAddressReq'] = _GETMULTISIGSPENDTXSBYADDRESSREQ
DESCRIPTOR.message_types_by_name['GetMultiSigSpendTxsByAddressResp'] = _GETMULTISIGSPENDTXSBYADDRESSRESP
DESCRIPTOR.message_types_by_name['GetVoteStatsReq'] = _GETVOTESTATSREQ
DESCRIPTOR.message_types_by_name['GetVoteStatsResp'] = _GETVOTESTATSRESP
DESCRIPTOR.message_types_by_name['GetInboxMessagesByAddressResp'] = _GETINBOXMESSAGESBYADDRESSRESP
DESCRIPTOR.message_types_by_name['InboxMessage'] = _INBOXMESSAGE
DESCRIPTOR.message_types_by_name['TokenDetail'] = _TOKENDETAIL
DESCRIPTOR.message_types_by_name['GetTokensByAddressResp'] = _GETTOKENSBYADDRESSRESP
DESCRIPTOR.message_types_by_name['SlaveDetail'] = _SLAVEDETAIL
DESCRIPTOR.message_types_by_name['GetSlavesByAddressResp'] = _GETSLAVESBYADDRESSRESP
DESCRIPTOR.message_types_by_name['LatticePKsDetail'] = _LATTICEPKSDETAIL
DESCRIPTOR.message_types_by_name['GetLatticePKsByAddressResp'] = _GETLATTICEPKSBYADDRESSRESP
DESCRIPTOR.message_types_by_name['MultiSigDetail'] = _MULTISIGDETAIL
DESCRIPTOR.message_types_by_name['GetMultiSigAddressesByAddressResp'] = _GETMULTISIGADDRESSESBYADDRESSRESP
DESCRIPTOR.message_types_by_name['GetBalanceReq'] = _GETBALANCEREQ
DESCRIPTOR.message_types_by_name['GetBalanceResp'] = _GETBALANCERESP
DESCRIPTOR.message_types_by_name['GetTotalBalanceReq'] = _GETTOTALBALANCEREQ
DESCRIPTOR.message_types_by_name['GetTotalBalanceResp'] = _GETTOTALBALANCERESP
DESCRIPTOR.message_types_by_name['GetOTSReq'] = _GETOTSREQ
DESCRIPTOR.message_types_by_name['OTSBitfieldByPage'] = _OTSBITFIELDBYPAGE
DESCRIPTOR.message_types_by_name['GetOTSResp'] = _GETOTSRESP
DESCRIPTOR.message_types_by_name['GetHeightReq'] = _GETHEIGHTREQ
DESCRIPTOR.message_types_by_name['GetHeightResp'] = _GETHEIGHTRESP
DESCRIPTOR.message_types_by_name['GetBlockReq'] = _GETBLOCKREQ
DESCRIPTOR.message_types_by_name['GetBlockResp'] = _GETBLOCKRESP
DESCRIPTOR.message_types_by_name['GetBlockByNumberReq'] = _GETBLOCKBYNUMBERREQ
DESCRIPTOR.message_types_by_name['GetBlockByNumberResp'] = _GETBLOCKBYNUMBERRESP
DESCRIPTOR.message_types_by_name['GetLocalAddressesReq'] = _GETLOCALADDRESSESREQ
DESCRIPTOR.message_types_by_name['GetLocalAddressesResp'] = _GETLOCALADDRESSESRESP
DESCRIPTOR.message_types_by_name['NodeInfo'] = _NODEINFO
DESCRIPTOR.message_types_by_name['AddressDescriptor'] = _ADDRESSDESCRIPTOR
DESCRIPTOR.message_types_by_name['StoredPeers'] = _STOREDPEERS
DESCRIPTOR.message_types_by_name['Peer'] = _PEER
DESCRIPTOR.message_types_by_name['AddressState'] = _ADDRESSSTATE
DESCRIPTOR.message_types_by_name['OptimizedAddressState'] = _OPTIMIZEDADDRESSSTATE
DESCRIPTOR.message_types_by_name['MultiSigAddressState'] = _MULTISIGADDRESSSTATE
DESCRIPTOR.message_types_by_name['MultiSigAddressesList'] = _MULTISIGADDRESSESLIST
DESCRIPTOR.message_types_by_name['DataList'] = _DATALIST
DESCRIPTOR.message_types_by_name['Bitfield'] = _BITFIELD
DESCRIPTOR.message_types_by_name['TransactionHashList'] = _TRANSACTIONHASHLIST
DESCRIPTOR.message_types_by_name['LatticePK'] = _LATTICEPK
DESCRIPTOR.message_types_by_name['AddressAmount'] = _ADDRESSAMOUNT
DESCRIPTOR.message_types_by_name['BlockHeader'] = _BLOCKHEADER
DESCRIPTOR.message_types_by_name['BlockHeaderExtended'] = _BLOCKHEADEREXTENDED
DESCRIPTOR.message_types_by_name['TransactionCount'] = _TRANSACTIONCOUNT
DESCRIPTOR.message_types_by_name['TransactionExtended'] = _TRANSACTIONEXTENDED
DESCRIPTOR.message_types_by_name['BlockExtended'] = _BLOCKEXTENDED
DESCRIPTOR.message_types_by_name['Block'] = _BLOCK
DESCRIPTOR.message_types_by_name['GenesisBalance'] = _GENESISBALANCE
DESCRIPTOR.message_types_by_name['BlockMetaDataList'] = _BLOCKMETADATALIST
DESCRIPTOR.message_types_by_name['Transaction'] = _TRANSACTION
DESCRIPTOR.message_types_by_name['VoteStats'] = _VOTESTATS
DESCRIPTOR.message_types_by_name['ProposalVoteStats'] = _PROPOSALVOTESTATS
DESCRIPTOR.message_types_by_name['ProposalRecord'] = _PROPOSALRECORD
DESCRIPTOR.message_types_by_name['TokenList'] = _TOKENLIST
DESCRIPTOR.message_types_by_name['TokenBalance'] = _TOKENBALANCE
DESCRIPTOR.message_types_by_name['SlaveMetadata'] = _SLAVEMETADATA
DESCRIPTOR.message_types_by_name['LatticePKMetadata'] = _LATTICEPKMETADATA
DESCRIPTOR.message_types_by_name['TokenMetadata'] = _TOKENMETADATA
DESCRIPTOR.message_types_by_name['EncryptedEphemeralMessage'] = _ENCRYPTEDEPHEMERALMESSAGE
DESCRIPTOR.message_types_by_name['AddressList'] = _ADDRESSLIST
DESCRIPTOR.message_types_by_name['BlockHeightData'] = _BLOCKHEIGHTDATA
DESCRIPTOR.message_types_by_name['BlockMetaData'] = _BLOCKMETADATA
DESCRIPTOR.message_types_by_name['BlockNumberMapping'] = _BLOCKNUMBERMAPPING
DESCRIPTOR.message_types_by_name['PeerStat'] = _PEERSTAT
DESCRIPTOR.message_types_by_name['NodeChainState'] = _NODECHAINSTATE
DESCRIPTOR.message_types_by_name['NodeHeaderHash'] = _NODEHEADERHASH
DESCRIPTOR.message_types_by_name['P2PAcknowledgement'] = _P2PACKNOWLEDGEMENT
DESCRIPTOR.message_types_by_name['PeerInfo'] = _PEERINFO
DESCRIPTOR.message_types_by_name['Peers'] = _PEERS
DESCRIPTOR.message_types_by_name['DevConfig'] = _DEVCONFIG
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Empty = _reflection.GeneratedProtocolMessageType('Empty', (_message.Message,), dict(
  DESCRIPTOR = _EMPTY,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.Empty)
  ))
_sym_db.RegisterMessage(Empty)

GetNodeStateReq = _reflection.GeneratedProtocolMessageType('GetNodeStateReq', (_message.Message,), dict(
  DESCRIPTOR = _GETNODESTATEREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetNodeStateReq)
  ))
_sym_db.RegisterMessage(GetNodeStateReq)

GetNodeStateResp = _reflection.GeneratedProtocolMessageType('GetNodeStateResp', (_message.Message,), dict(
  DESCRIPTOR = _GETNODESTATERESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetNodeStateResp)
  ))
_sym_db.RegisterMessage(GetNodeStateResp)

GetKnownPeersReq = _reflection.GeneratedProtocolMessageType('GetKnownPeersReq', (_message.Message,), dict(
  DESCRIPTOR = _GETKNOWNPEERSREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetKnownPeersReq)
  ))
_sym_db.RegisterMessage(GetKnownPeersReq)

GetKnownPeersResp = _reflection.GeneratedProtocolMessageType('GetKnownPeersResp', (_message.Message,), dict(
  DESCRIPTOR = _GETKNOWNPEERSRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetKnownPeersResp)
  ))
_sym_db.RegisterMessage(GetKnownPeersResp)

GetPeersStatReq = _reflection.GeneratedProtocolMessageType('GetPeersStatReq', (_message.Message,), dict(
  DESCRIPTOR = _GETPEERSSTATREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetPeersStatReq)
  ))
_sym_db.RegisterMessage(GetPeersStatReq)

GetPeersStatResp = _reflection.GeneratedProtocolMessageType('GetPeersStatResp', (_message.Message,), dict(
  DESCRIPTOR = _GETPEERSSTATRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetPeersStatResp)
  ))
_sym_db.RegisterMessage(GetPeersStatResp)

GetChainStatsReq = _reflection.GeneratedProtocolMessageType('GetChainStatsReq', (_message.Message,), dict(
  DESCRIPTOR = _GETCHAINSTATSREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetChainStatsReq)
  ))
_sym_db.RegisterMessage(GetChainStatsReq)

GetChainStatsResp = _reflection.GeneratedProtocolMessageType('GetChainStatsResp', (_message.Message,), dict(
  DESCRIPTOR = _GETCHAINSTATSRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetChainStatsResp)
  ))
_sym_db.RegisterMessage(GetChainStatsResp)

GetStatsReq = _reflection.GeneratedProtocolMessageType('GetStatsReq', (_message.Message,), dict(
  DESCRIPTOR = _GETSTATSREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetStatsReq)
  ))
_sym_db.RegisterMessage(GetStatsReq)

GetStatsResp = _reflection.GeneratedProtocolMessageType('GetStatsResp', (_message.Message,), dict(
  DESCRIPTOR = _GETSTATSRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetStatsResp)
  ))
_sym_db.RegisterMessage(GetStatsResp)

GetAddressFromPKReq = _reflection.GeneratedProtocolMessageType('GetAddressFromPKReq', (_message.Message,), dict(
  DESCRIPTOR = _GETADDRESSFROMPKREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetAddressFromPKReq)
  ))
_sym_db.RegisterMessage(GetAddressFromPKReq)

GetAddressFromPKResp = _reflection.GeneratedProtocolMessageType('GetAddressFromPKResp', (_message.Message,), dict(
  DESCRIPTOR = _GETADDRESSFROMPKRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetAddressFromPKResp)
  ))
_sym_db.RegisterMessage(GetAddressFromPKResp)

BlockDataPoint = _reflection.GeneratedProtocolMessageType('BlockDataPoint', (_message.Message,), dict(
  DESCRIPTOR = _BLOCKDATAPOINT,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.BlockDataPoint)
  ))
_sym_db.RegisterMessage(BlockDataPoint)

GetAddressStateReq = _reflection.GeneratedProtocolMessageType('GetAddressStateReq', (_message.Message,), dict(
  DESCRIPTOR = _GETADDRESSSTATEREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetAddressStateReq)
  ))
_sym_db.RegisterMessage(GetAddressStateReq)

GetAddressStateResp = _reflection.GeneratedProtocolMessageType('GetAddressStateResp', (_message.Message,), dict(
  DESCRIPTOR = _GETADDRESSSTATERESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetAddressStateResp)
  ))
_sym_db.RegisterMessage(GetAddressStateResp)

GetOptimizedAddressStateResp = _reflection.GeneratedProtocolMessageType('GetOptimizedAddressStateResp', (_message.Message,), dict(
  DESCRIPTOR = _GETOPTIMIZEDADDRESSSTATERESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetOptimizedAddressStateResp)
  ))
_sym_db.RegisterMessage(GetOptimizedAddressStateResp)

GetMultiSigAddressStateReq = _reflection.GeneratedProtocolMessageType('GetMultiSigAddressStateReq', (_message.Message,), dict(
  DESCRIPTOR = _GETMULTISIGADDRESSSTATEREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetMultiSigAddressStateReq)
  ))
_sym_db.RegisterMessage(GetMultiSigAddressStateReq)

GetMultiSigAddressStateResp = _reflection.GeneratedProtocolMessageType('GetMultiSigAddressStateResp', (_message.Message,), dict(
  DESCRIPTOR = _GETMULTISIGADDRESSSTATERESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetMultiSigAddressStateResp)
  ))
_sym_db.RegisterMessage(GetMultiSigAddressStateResp)

IsSlaveReq = _reflection.GeneratedProtocolMessageType('IsSlaveReq', (_message.Message,), dict(
  DESCRIPTOR = _ISSLAVEREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.IsSlaveReq)
  ))
_sym_db.RegisterMessage(IsSlaveReq)

IsSlaveResp = _reflection.GeneratedProtocolMessageType('IsSlaveResp', (_message.Message,), dict(
  DESCRIPTOR = _ISSLAVERESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.IsSlaveResp)
  ))
_sym_db.RegisterMessage(IsSlaveResp)

ParseAddressReq = _reflection.GeneratedProtocolMessageType('ParseAddressReq', (_message.Message,), dict(
  DESCRIPTOR = _PARSEADDRESSREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.ParseAddressReq)
  ))
_sym_db.RegisterMessage(ParseAddressReq)

ParseAddressResp = _reflection.GeneratedProtocolMessageType('ParseAddressResp', (_message.Message,), dict(
  DESCRIPTOR = _PARSEADDRESSRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.ParseAddressResp)
  ))
_sym_db.RegisterMessage(ParseAddressResp)

GetObjectReq = _reflection.GeneratedProtocolMessageType('GetObjectReq', (_message.Message,), dict(
  DESCRIPTOR = _GETOBJECTREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetObjectReq)
  ))
_sym_db.RegisterMessage(GetObjectReq)

GetObjectResp = _reflection.GeneratedProtocolMessageType('GetObjectResp', (_message.Message,), dict(
  DESCRIPTOR = _GETOBJECTRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetObjectResp)
  ))
_sym_db.RegisterMessage(GetObjectResp)

GetLatestDataReq = _reflection.GeneratedProtocolMessageType('GetLatestDataReq', (_message.Message,), dict(
  DESCRIPTOR = _GETLATESTDATAREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetLatestDataReq)
  ))
_sym_db.RegisterMessage(GetLatestDataReq)

GetLatestDataResp = _reflection.GeneratedProtocolMessageType('GetLatestDataResp', (_message.Message,), dict(
  DESCRIPTOR = _GETLATESTDATARESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetLatestDataResp)
  ))
_sym_db.RegisterMessage(GetLatestDataResp)

TransferCoinsReq = _reflection.GeneratedProtocolMessageType('TransferCoinsReq', (_message.Message,), dict(
  DESCRIPTOR = _TRANSFERCOINSREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.TransferCoinsReq)
  ))
_sym_db.RegisterMessage(TransferCoinsReq)

TransferCoinsResp = _reflection.GeneratedProtocolMessageType('TransferCoinsResp', (_message.Message,), dict(
  DESCRIPTOR = _TRANSFERCOINSRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.TransferCoinsResp)
  ))
_sym_db.RegisterMessage(TransferCoinsResp)

PushTransactionReq = _reflection.GeneratedProtocolMessageType('PushTransactionReq', (_message.Message,), dict(
  DESCRIPTOR = _PUSHTRANSACTIONREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.PushTransactionReq)
  ))
_sym_db.RegisterMessage(PushTransactionReq)

PushTransactionResp = _reflection.GeneratedProtocolMessageType('PushTransactionResp', (_message.Message,), dict(
  DESCRIPTOR = _PUSHTRANSACTIONRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.PushTransactionResp)
  ))
_sym_db.RegisterMessage(PushTransactionResp)

MultiSigCreateTxnReq = _reflection.GeneratedProtocolMessageType('MultiSigCreateTxnReq', (_message.Message,), dict(
  DESCRIPTOR = _MULTISIGCREATETXNREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.MultiSigCreateTxnReq)
  ))
_sym_db.RegisterMessage(MultiSigCreateTxnReq)

MultiSigSpendTxnReq = _reflection.GeneratedProtocolMessageType('MultiSigSpendTxnReq', (_message.Message,), dict(
  DESCRIPTOR = _MULTISIGSPENDTXNREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.MultiSigSpendTxnReq)
  ))
_sym_db.RegisterMessage(MultiSigSpendTxnReq)

MultiSigVoteTxnReq = _reflection.GeneratedProtocolMessageType('MultiSigVoteTxnReq', (_message.Message,), dict(
  DESCRIPTOR = _MULTISIGVOTETXNREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.MultiSigVoteTxnReq)
  ))
_sym_db.RegisterMessage(MultiSigVoteTxnReq)

MessageTxnReq = _reflection.GeneratedProtocolMessageType('MessageTxnReq', (_message.Message,), dict(
  DESCRIPTOR = _MESSAGETXNREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.MessageTxnReq)
  ))
_sym_db.RegisterMessage(MessageTxnReq)

TokenTxnReq = _reflection.GeneratedProtocolMessageType('TokenTxnReq', (_message.Message,), dict(
  DESCRIPTOR = _TOKENTXNREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.TokenTxnReq)
  ))
_sym_db.RegisterMessage(TokenTxnReq)

TransferTokenTxnReq = _reflection.GeneratedProtocolMessageType('TransferTokenTxnReq', (_message.Message,), dict(
  DESCRIPTOR = _TRANSFERTOKENTXNREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.TransferTokenTxnReq)
  ))
_sym_db.RegisterMessage(TransferTokenTxnReq)

SlaveTxnReq = _reflection.GeneratedProtocolMessageType('SlaveTxnReq', (_message.Message,), dict(
  DESCRIPTOR = _SLAVETXNREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.SlaveTxnReq)
  ))
_sym_db.RegisterMessage(SlaveTxnReq)

LatticeTxnReq = _reflection.GeneratedProtocolMessageType('LatticeTxnReq', (_message.Message,), dict(
  DESCRIPTOR = _LATTICETXNREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.LatticeTxnReq)
  ))
_sym_db.RegisterMessage(LatticeTxnReq)

MiniTransaction = _reflection.GeneratedProtocolMessageType('MiniTransaction', (_message.Message,), dict(
  DESCRIPTOR = _MINITRANSACTION,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.MiniTransaction)
  ))
_sym_db.RegisterMessage(MiniTransaction)

GetTransactionReq = _reflection.GeneratedProtocolMessageType('GetTransactionReq', (_message.Message,), dict(
  DESCRIPTOR = _GETTRANSACTIONREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetTransactionReq)
  ))
_sym_db.RegisterMessage(GetTransactionReq)

GetTransactionResp = _reflection.GeneratedProtocolMessageType('GetTransactionResp', (_message.Message,), dict(
  DESCRIPTOR = _GETTRANSACTIONRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetTransactionResp)
  ))
_sym_db.RegisterMessage(GetTransactionResp)

GetMiniTransactionsByAddressReq = _reflection.GeneratedProtocolMessageType('GetMiniTransactionsByAddressReq', (_message.Message,), dict(
  DESCRIPTOR = _GETMINITRANSACTIONSBYADDRESSREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetMiniTransactionsByAddressReq)
  ))
_sym_db.RegisterMessage(GetMiniTransactionsByAddressReq)

GetMiniTransactionsByAddressResp = _reflection.GeneratedProtocolMessageType('GetMiniTransactionsByAddressResp', (_message.Message,), dict(
  DESCRIPTOR = _GETMINITRANSACTIONSBYADDRESSRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetMiniTransactionsByAddressResp)
  ))
_sym_db.RegisterMessage(GetMiniTransactionsByAddressResp)

GetTransactionsByAddressReq = _reflection.GeneratedProtocolMessageType('GetTransactionsByAddressReq', (_message.Message,), dict(
  DESCRIPTOR = _GETTRANSACTIONSBYADDRESSREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetTransactionsByAddressReq)
  ))
_sym_db.RegisterMessage(GetTransactionsByAddressReq)

GetTransactionsByAddressResp = _reflection.GeneratedProtocolMessageType('GetTransactionsByAddressResp', (_message.Message,), dict(
  DESCRIPTOR = _GETTRANSACTIONSBYADDRESSRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetTransactionsByAddressResp)
  ))
_sym_db.RegisterMessage(GetTransactionsByAddressResp)

GetMultiSigSpendTxsByAddressReq = _reflection.GeneratedProtocolMessageType('GetMultiSigSpendTxsByAddressReq', (_message.Message,), dict(
  DESCRIPTOR = _GETMULTISIGSPENDTXSBYADDRESSREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetMultiSigSpendTxsByAddressReq)
  ))
_sym_db.RegisterMessage(GetMultiSigSpendTxsByAddressReq)

GetMultiSigSpendTxsByAddressResp = _reflection.GeneratedProtocolMessageType('GetMultiSigSpendTxsByAddressResp', (_message.Message,), dict(
  DESCRIPTOR = _GETMULTISIGSPENDTXSBYADDRESSRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetMultiSigSpendTxsByAddressResp)
  ))
_sym_db.RegisterMessage(GetMultiSigSpendTxsByAddressResp)

GetVoteStatsReq = _reflection.GeneratedProtocolMessageType('GetVoteStatsReq', (_message.Message,), dict(
  DESCRIPTOR = _GETVOTESTATSREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetVoteStatsReq)
  ))
_sym_db.RegisterMessage(GetVoteStatsReq)

GetVoteStatsResp = _reflection.GeneratedProtocolMessageType('GetVoteStatsResp', (_message.Message,), dict(
  DESCRIPTOR = _GETVOTESTATSRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetVoteStatsResp)
  ))
_sym_db.RegisterMessage(GetVoteStatsResp)

GetInboxMessagesByAddressResp = _reflection.GeneratedProtocolMessageType('GetInboxMessagesByAddressResp', (_message.Message,), dict(
  DESCRIPTOR = _GETINBOXMESSAGESBYADDRESSRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetInboxMessagesByAddressResp)
  ))
_sym_db.RegisterMessage(GetInboxMessagesByAddressResp)

InboxMessage = _reflection.GeneratedProtocolMessageType('InboxMessage', (_message.Message,), dict(
  DESCRIPTOR = _INBOXMESSAGE,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.InboxMessage)
  ))
_sym_db.RegisterMessage(InboxMessage)

TokenDetail = _reflection.GeneratedProtocolMessageType('TokenDetail', (_message.Message,), dict(
  DESCRIPTOR = _TOKENDETAIL,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.TokenDetail)
  ))
_sym_db.RegisterMessage(TokenDetail)

GetTokensByAddressResp = _reflection.GeneratedProtocolMessageType('GetTokensByAddressResp', (_message.Message,), dict(
  DESCRIPTOR = _GETTOKENSBYADDRESSRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetTokensByAddressResp)
  ))
_sym_db.RegisterMessage(GetTokensByAddressResp)

SlaveDetail = _reflection.GeneratedProtocolMessageType('SlaveDetail', (_message.Message,), dict(
  DESCRIPTOR = _SLAVEDETAIL,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.SlaveDetail)
  ))
_sym_db.RegisterMessage(SlaveDetail)

GetSlavesByAddressResp = _reflection.GeneratedProtocolMessageType('GetSlavesByAddressResp', (_message.Message,), dict(
  DESCRIPTOR = _GETSLAVESBYADDRESSRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetSlavesByAddressResp)
  ))
_sym_db.RegisterMessage(GetSlavesByAddressResp)

LatticePKsDetail = _reflection.GeneratedProtocolMessageType('LatticePKsDetail', (_message.Message,), dict(
  DESCRIPTOR = _LATTICEPKSDETAIL,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.LatticePKsDetail)
  ))
_sym_db.RegisterMessage(LatticePKsDetail)

GetLatticePKsByAddressResp = _reflection.GeneratedProtocolMessageType('GetLatticePKsByAddressResp', (_message.Message,), dict(
  DESCRIPTOR = _GETLATTICEPKSBYADDRESSRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetLatticePKsByAddressResp)
  ))
_sym_db.RegisterMessage(GetLatticePKsByAddressResp)

MultiSigDetail = _reflection.GeneratedProtocolMessageType('MultiSigDetail', (_message.Message,), dict(
  DESCRIPTOR = _MULTISIGDETAIL,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.MultiSigDetail)
  ))
_sym_db.RegisterMessage(MultiSigDetail)

GetMultiSigAddressesByAddressResp = _reflection.GeneratedProtocolMessageType('GetMultiSigAddressesByAddressResp', (_message.Message,), dict(
  DESCRIPTOR = _GETMULTISIGADDRESSESBYADDRESSRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetMultiSigAddressesByAddressResp)
  ))
_sym_db.RegisterMessage(GetMultiSigAddressesByAddressResp)

GetBalanceReq = _reflection.GeneratedProtocolMessageType('GetBalanceReq', (_message.Message,), dict(
  DESCRIPTOR = _GETBALANCEREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetBalanceReq)
  ))
_sym_db.RegisterMessage(GetBalanceReq)

GetBalanceResp = _reflection.GeneratedProtocolMessageType('GetBalanceResp', (_message.Message,), dict(
  DESCRIPTOR = _GETBALANCERESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetBalanceResp)
  ))
_sym_db.RegisterMessage(GetBalanceResp)

GetTotalBalanceReq = _reflection.GeneratedProtocolMessageType('GetTotalBalanceReq', (_message.Message,), dict(
  DESCRIPTOR = _GETTOTALBALANCEREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetTotalBalanceReq)
  ))
_sym_db.RegisterMessage(GetTotalBalanceReq)

GetTotalBalanceResp = _reflection.GeneratedProtocolMessageType('GetTotalBalanceResp', (_message.Message,), dict(
  DESCRIPTOR = _GETTOTALBALANCERESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetTotalBalanceResp)
  ))
_sym_db.RegisterMessage(GetTotalBalanceResp)

GetOTSReq = _reflection.GeneratedProtocolMessageType('GetOTSReq', (_message.Message,), dict(
  DESCRIPTOR = _GETOTSREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetOTSReq)
  ))
_sym_db.RegisterMessage(GetOTSReq)

OTSBitfieldByPage = _reflection.GeneratedProtocolMessageType('OTSBitfieldByPage', (_message.Message,), dict(
  DESCRIPTOR = _OTSBITFIELDBYPAGE,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.OTSBitfieldByPage)
  ))
_sym_db.RegisterMessage(OTSBitfieldByPage)

GetOTSResp = _reflection.GeneratedProtocolMessageType('GetOTSResp', (_message.Message,), dict(
  DESCRIPTOR = _GETOTSRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetOTSResp)
  ))
_sym_db.RegisterMessage(GetOTSResp)

GetHeightReq = _reflection.GeneratedProtocolMessageType('GetHeightReq', (_message.Message,), dict(
  DESCRIPTOR = _GETHEIGHTREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetHeightReq)
  ))
_sym_db.RegisterMessage(GetHeightReq)

GetHeightResp = _reflection.GeneratedProtocolMessageType('GetHeightResp', (_message.Message,), dict(
  DESCRIPTOR = _GETHEIGHTRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetHeightResp)
  ))
_sym_db.RegisterMessage(GetHeightResp)

GetBlockReq = _reflection.GeneratedProtocolMessageType('GetBlockReq', (_message.Message,), dict(
  DESCRIPTOR = _GETBLOCKREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetBlockReq)
  ))
_sym_db.RegisterMessage(GetBlockReq)

GetBlockResp = _reflection.GeneratedProtocolMessageType('GetBlockResp', (_message.Message,), dict(
  DESCRIPTOR = _GETBLOCKRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetBlockResp)
  ))
_sym_db.RegisterMessage(GetBlockResp)

GetBlockByNumberReq = _reflection.GeneratedProtocolMessageType('GetBlockByNumberReq', (_message.Message,), dict(
  DESCRIPTOR = _GETBLOCKBYNUMBERREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetBlockByNumberReq)
  ))
_sym_db.RegisterMessage(GetBlockByNumberReq)

GetBlockByNumberResp = _reflection.GeneratedProtocolMessageType('GetBlockByNumberResp', (_message.Message,), dict(
  DESCRIPTOR = _GETBLOCKBYNUMBERRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetBlockByNumberResp)
  ))
_sym_db.RegisterMessage(GetBlockByNumberResp)

GetLocalAddressesReq = _reflection.GeneratedProtocolMessageType('GetLocalAddressesReq', (_message.Message,), dict(
  DESCRIPTOR = _GETLOCALADDRESSESREQ,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetLocalAddressesReq)
  ))
_sym_db.RegisterMessage(GetLocalAddressesReq)

GetLocalAddressesResp = _reflection.GeneratedProtocolMessageType('GetLocalAddressesResp', (_message.Message,), dict(
  DESCRIPTOR = _GETLOCALADDRESSESRESP,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetLocalAddressesResp)
  ))
_sym_db.RegisterMessage(GetLocalAddressesResp)

NodeInfo = _reflection.GeneratedProtocolMessageType('NodeInfo', (_message.Message,), dict(
  DESCRIPTOR = _NODEINFO,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.NodeInfo)
  ))
_sym_db.RegisterMessage(NodeInfo)

AddressDescriptor = _reflection.GeneratedProtocolMessageType('AddressDescriptor', (_message.Message,), dict(
  DESCRIPTOR = _ADDRESSDESCRIPTOR,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.AddressDescriptor)
  ))
_sym_db.RegisterMessage(AddressDescriptor)

StoredPeers = _reflection.GeneratedProtocolMessageType('StoredPeers', (_message.Message,), dict(
  DESCRIPTOR = _STOREDPEERS,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.StoredPeers)
  ))
_sym_db.RegisterMessage(StoredPeers)

Peer = _reflection.GeneratedProtocolMessageType('Peer', (_message.Message,), dict(
  DESCRIPTOR = _PEER,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.Peer)
  ))
_sym_db.RegisterMessage(Peer)

AddressState = _reflection.GeneratedProtocolMessageType('AddressState', (_message.Message,), dict(

  TokensEntry = _reflection.GeneratedProtocolMessageType('TokensEntry', (_message.Message,), dict(
    DESCRIPTOR = _ADDRESSSTATE_TOKENSENTRY,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.AddressState.TokensEntry)
    ))
  ,

  SlavePksAccessTypeEntry = _reflection.GeneratedProtocolMessageType('SlavePksAccessTypeEntry', (_message.Message,), dict(
    DESCRIPTOR = _ADDRESSSTATE_SLAVEPKSACCESSTYPEENTRY,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.AddressState.SlavePksAccessTypeEntry)
    ))
  ,
  DESCRIPTOR = _ADDRESSSTATE,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.AddressState)
  ))
_sym_db.RegisterMessage(AddressState)
_sym_db.RegisterMessage(AddressState.TokensEntry)
_sym_db.RegisterMessage(AddressState.SlavePksAccessTypeEntry)

OptimizedAddressState = _reflection.GeneratedProtocolMessageType('OptimizedAddressState', (_message.Message,), dict(
  DESCRIPTOR = _OPTIMIZEDADDRESSSTATE,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.OptimizedAddressState)
  ))
_sym_db.RegisterMessage(OptimizedAddressState)

MultiSigAddressState = _reflection.GeneratedProtocolMessageType('MultiSigAddressState', (_message.Message,), dict(
  DESCRIPTOR = _MULTISIGADDRESSSTATE,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.MultiSigAddressState)
  ))
_sym_db.RegisterMessage(MultiSigAddressState)

MultiSigAddressesList = _reflection.GeneratedProtocolMessageType('MultiSigAddressesList', (_message.Message,), dict(
  DESCRIPTOR = _MULTISIGADDRESSESLIST,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.MultiSigAddressesList)
  ))
_sym_db.RegisterMessage(MultiSigAddressesList)

DataList = _reflection.GeneratedProtocolMessageType('DataList', (_message.Message,), dict(
  DESCRIPTOR = _DATALIST,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.DataList)
  ))
_sym_db.RegisterMessage(DataList)

Bitfield = _reflection.GeneratedProtocolMessageType('Bitfield', (_message.Message,), dict(
  DESCRIPTOR = _BITFIELD,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.Bitfield)
  ))
_sym_db.RegisterMessage(Bitfield)

TransactionHashList = _reflection.GeneratedProtocolMessageType('TransactionHashList', (_message.Message,), dict(
  DESCRIPTOR = _TRANSACTIONHASHLIST,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.TransactionHashList)
  ))
_sym_db.RegisterMessage(TransactionHashList)

LatticePK = _reflection.GeneratedProtocolMessageType('LatticePK', (_message.Message,), dict(
  DESCRIPTOR = _LATTICEPK,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.LatticePK)
  ))
_sym_db.RegisterMessage(LatticePK)

AddressAmount = _reflection.GeneratedProtocolMessageType('AddressAmount', (_message.Message,), dict(
  DESCRIPTOR = _ADDRESSAMOUNT,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.AddressAmount)
  ))
_sym_db.RegisterMessage(AddressAmount)

BlockHeader = _reflection.GeneratedProtocolMessageType('BlockHeader', (_message.Message,), dict(
  DESCRIPTOR = _BLOCKHEADER,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.BlockHeader)
  ))
_sym_db.RegisterMessage(BlockHeader)

BlockHeaderExtended = _reflection.GeneratedProtocolMessageType('BlockHeaderExtended', (_message.Message,), dict(
  DESCRIPTOR = _BLOCKHEADEREXTENDED,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.BlockHeaderExtended)
  ))
_sym_db.RegisterMessage(BlockHeaderExtended)

TransactionCount = _reflection.GeneratedProtocolMessageType('TransactionCount', (_message.Message,), dict(

  CountEntry = _reflection.GeneratedProtocolMessageType('CountEntry', (_message.Message,), dict(
    DESCRIPTOR = _TRANSACTIONCOUNT_COUNTENTRY,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.TransactionCount.CountEntry)
    ))
  ,
  DESCRIPTOR = _TRANSACTIONCOUNT,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.TransactionCount)
  ))
_sym_db.RegisterMessage(TransactionCount)
_sym_db.RegisterMessage(TransactionCount.CountEntry)

TransactionExtended = _reflection.GeneratedProtocolMessageType('TransactionExtended', (_message.Message,), dict(
  DESCRIPTOR = _TRANSACTIONEXTENDED,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.TransactionExtended)
  ))
_sym_db.RegisterMessage(TransactionExtended)

BlockExtended = _reflection.GeneratedProtocolMessageType('BlockExtended', (_message.Message,), dict(
  DESCRIPTOR = _BLOCKEXTENDED,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.BlockExtended)
  ))
_sym_db.RegisterMessage(BlockExtended)

Block = _reflection.GeneratedProtocolMessageType('Block', (_message.Message,), dict(
  DESCRIPTOR = _BLOCK,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.Block)
  ))
_sym_db.RegisterMessage(Block)

GenesisBalance = _reflection.GeneratedProtocolMessageType('GenesisBalance', (_message.Message,), dict(
  DESCRIPTOR = _GENESISBALANCE,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GenesisBalance)
  ))
_sym_db.RegisterMessage(GenesisBalance)

BlockMetaDataList = _reflection.GeneratedProtocolMessageType('BlockMetaDataList', (_message.Message,), dict(
  DESCRIPTOR = _BLOCKMETADATALIST,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.BlockMetaDataList)
  ))
_sym_db.RegisterMessage(BlockMetaDataList)

Transaction = _reflection.GeneratedProtocolMessageType('Transaction', (_message.Message,), dict(

  Transfer = _reflection.GeneratedProtocolMessageType('Transfer', (_message.Message,), dict(
    DESCRIPTOR = _TRANSACTION_TRANSFER,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.Transaction.Transfer)
    ))
  ,

  CoinBase = _reflection.GeneratedProtocolMessageType('CoinBase', (_message.Message,), dict(
    DESCRIPTOR = _TRANSACTION_COINBASE,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.Transaction.CoinBase)
    ))
  ,

  LatticePublicKey = _reflection.GeneratedProtocolMessageType('LatticePublicKey', (_message.Message,), dict(
    DESCRIPTOR = _TRANSACTION_LATTICEPUBLICKEY,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.Transaction.LatticePublicKey)
    ))
  ,

  Message = _reflection.GeneratedProtocolMessageType('Message', (_message.Message,), dict(
    DESCRIPTOR = _TRANSACTION_MESSAGE,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.Transaction.Message)
    ))
  ,

  Token = _reflection.GeneratedProtocolMessageType('Token', (_message.Message,), dict(
    DESCRIPTOR = _TRANSACTION_TOKEN,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.Transaction.Token)
    ))
  ,

  TransferToken = _reflection.GeneratedProtocolMessageType('TransferToken', (_message.Message,), dict(
    DESCRIPTOR = _TRANSACTION_TRANSFERTOKEN,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.Transaction.TransferToken)
    ))
  ,

  Slave = _reflection.GeneratedProtocolMessageType('Slave', (_message.Message,), dict(
    DESCRIPTOR = _TRANSACTION_SLAVE,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.Transaction.Slave)
    ))
  ,

  MultiSigCreate = _reflection.GeneratedProtocolMessageType('MultiSigCreate', (_message.Message,), dict(
    DESCRIPTOR = _TRANSACTION_MULTISIGCREATE,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.Transaction.MultiSigCreate)
    ))
  ,

  MultiSigSpend = _reflection.GeneratedProtocolMessageType('MultiSigSpend', (_message.Message,), dict(
    DESCRIPTOR = _TRANSACTION_MULTISIGSPEND,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.Transaction.MultiSigSpend)
    ))
  ,

  MultiSigVote = _reflection.GeneratedProtocolMessageType('MultiSigVote', (_message.Message,), dict(
    DESCRIPTOR = _TRANSACTION_MULTISIGVOTE,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.Transaction.MultiSigVote)
    ))
  ,

  ProposalCreate = _reflection.GeneratedProtocolMessageType('ProposalCreate', (_message.Message,), dict(

    QIP = _reflection.GeneratedProtocolMessageType('QIP', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_PROPOSALCREATE_QIP,
      __module__ = 'qrl_pb2'
      # @@protoc_insertion_point(class_scope:qrl.Transaction.ProposalCreate.QIP)
      ))
    ,

    Config = _reflection.GeneratedProtocolMessageType('Config', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_PROPOSALCREATE_CONFIG,
      __module__ = 'qrl_pb2'
      # @@protoc_insertion_point(class_scope:qrl.Transaction.ProposalCreate.Config)
      ))
    ,

    Other = _reflection.GeneratedProtocolMessageType('Other', (_message.Message,), dict(
      DESCRIPTOR = _TRANSACTION_PROPOSALCREATE_OTHER,
      __module__ = 'qrl_pb2'
      # @@protoc_insertion_point(class_scope:qrl.Transaction.ProposalCreate.Other)
      ))
    ,
    DESCRIPTOR = _TRANSACTION_PROPOSALCREATE,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.Transaction.ProposalCreate)
    ))
  ,

  ProposalVote = _reflection.GeneratedProtocolMessageType('ProposalVote', (_message.Message,), dict(
    DESCRIPTOR = _TRANSACTION_PROPOSALVOTE,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.Transaction.ProposalVote)
    ))
  ,
  DESCRIPTOR = _TRANSACTION,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.Transaction)
  ))
_sym_db.RegisterMessage(Transaction)
_sym_db.RegisterMessage(Transaction.Transfer)
_sym_db.RegisterMessage(Transaction.CoinBase)
_sym_db.RegisterMessage(Transaction.LatticePublicKey)
_sym_db.RegisterMessage(Transaction.Message)
_sym_db.RegisterMessage(Transaction.Token)
_sym_db.RegisterMessage(Transaction.TransferToken)
_sym_db.RegisterMessage(Transaction.Slave)
_sym_db.RegisterMessage(Transaction.MultiSigCreate)
_sym_db.RegisterMessage(Transaction.MultiSigSpend)
_sym_db.RegisterMessage(Transaction.MultiSigVote)
_sym_db.RegisterMessage(Transaction.ProposalCreate)
_sym_db.RegisterMessage(Transaction.ProposalCreate.QIP)
_sym_db.RegisterMessage(Transaction.ProposalCreate.Config)
_sym_db.RegisterMessage(Transaction.ProposalCreate.Other)
_sym_db.RegisterMessage(Transaction.ProposalVote)

VoteStats = _reflection.GeneratedProtocolMessageType('VoteStats', (_message.Message,), dict(
  DESCRIPTOR = _VOTESTATS,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.VoteStats)
  ))
_sym_db.RegisterMessage(VoteStats)

ProposalVoteStats = _reflection.GeneratedProtocolMessageType('ProposalVoteStats', (_message.Message,), dict(
  DESCRIPTOR = _PROPOSALVOTESTATS,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.ProposalVoteStats)
  ))
_sym_db.RegisterMessage(ProposalVoteStats)

ProposalRecord = _reflection.GeneratedProtocolMessageType('ProposalRecord', (_message.Message,), dict(
  DESCRIPTOR = _PROPOSALRECORD,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.ProposalRecord)
  ))
_sym_db.RegisterMessage(ProposalRecord)

TokenList = _reflection.GeneratedProtocolMessageType('TokenList', (_message.Message,), dict(
  DESCRIPTOR = _TOKENLIST,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.TokenList)
  ))
_sym_db.RegisterMessage(TokenList)

TokenBalance = _reflection.GeneratedProtocolMessageType('TokenBalance', (_message.Message,), dict(
  DESCRIPTOR = _TOKENBALANCE,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.TokenBalance)
  ))
_sym_db.RegisterMessage(TokenBalance)

SlaveMetadata = _reflection.GeneratedProtocolMessageType('SlaveMetadata', (_message.Message,), dict(
  DESCRIPTOR = _SLAVEMETADATA,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.SlaveMetadata)
  ))
_sym_db.RegisterMessage(SlaveMetadata)

LatticePKMetadata = _reflection.GeneratedProtocolMessageType('LatticePKMetadata', (_message.Message,), dict(
  DESCRIPTOR = _LATTICEPKMETADATA,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.LatticePKMetadata)
  ))
_sym_db.RegisterMessage(LatticePKMetadata)

TokenMetadata = _reflection.GeneratedProtocolMessageType('TokenMetadata', (_message.Message,), dict(
  DESCRIPTOR = _TOKENMETADATA,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.TokenMetadata)
  ))
_sym_db.RegisterMessage(TokenMetadata)

EncryptedEphemeralMessage = _reflection.GeneratedProtocolMessageType('EncryptedEphemeralMessage', (_message.Message,), dict(

  Channel = _reflection.GeneratedProtocolMessageType('Channel', (_message.Message,), dict(
    DESCRIPTOR = _ENCRYPTEDEPHEMERALMESSAGE_CHANNEL,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.EncryptedEphemeralMessage.Channel)
    ))
  ,
  DESCRIPTOR = _ENCRYPTEDEPHEMERALMESSAGE,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.EncryptedEphemeralMessage)
  ))
_sym_db.RegisterMessage(EncryptedEphemeralMessage)
_sym_db.RegisterMessage(EncryptedEphemeralMessage.Channel)

AddressList = _reflection.GeneratedProtocolMessageType('AddressList', (_message.Message,), dict(
  DESCRIPTOR = _ADDRESSLIST,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.AddressList)
  ))
_sym_db.RegisterMessage(AddressList)

BlockHeightData = _reflection.GeneratedProtocolMessageType('BlockHeightData', (_message.Message,), dict(
  DESCRIPTOR = _BLOCKHEIGHTDATA,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.BlockHeightData)
  ))
_sym_db.RegisterMessage(BlockHeightData)

BlockMetaData = _reflection.GeneratedProtocolMessageType('BlockMetaData', (_message.Message,), dict(
  DESCRIPTOR = _BLOCKMETADATA,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.BlockMetaData)
  ))
_sym_db.RegisterMessage(BlockMetaData)

BlockNumberMapping = _reflection.GeneratedProtocolMessageType('BlockNumberMapping', (_message.Message,), dict(
  DESCRIPTOR = _BLOCKNUMBERMAPPING,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.BlockNumberMapping)
  ))
_sym_db.RegisterMessage(BlockNumberMapping)

PeerStat = _reflection.GeneratedProtocolMessageType('PeerStat', (_message.Message,), dict(
  DESCRIPTOR = _PEERSTAT,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.PeerStat)
  ))
_sym_db.RegisterMessage(PeerStat)

NodeChainState = _reflection.GeneratedProtocolMessageType('NodeChainState', (_message.Message,), dict(
  DESCRIPTOR = _NODECHAINSTATE,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.NodeChainState)
  ))
_sym_db.RegisterMessage(NodeChainState)

NodeHeaderHash = _reflection.GeneratedProtocolMessageType('NodeHeaderHash', (_message.Message,), dict(
  DESCRIPTOR = _NODEHEADERHASH,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.NodeHeaderHash)
  ))
_sym_db.RegisterMessage(NodeHeaderHash)

P2PAcknowledgement = _reflection.GeneratedProtocolMessageType('P2PAcknowledgement', (_message.Message,), dict(
  DESCRIPTOR = _P2PACKNOWLEDGEMENT,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.P2PAcknowledgement)
  ))
_sym_db.RegisterMessage(P2PAcknowledgement)

PeerInfo = _reflection.GeneratedProtocolMessageType('PeerInfo', (_message.Message,), dict(
  DESCRIPTOR = _PEERINFO,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.PeerInfo)
  ))
_sym_db.RegisterMessage(PeerInfo)

Peers = _reflection.GeneratedProtocolMessageType('Peers', (_message.Message,), dict(
  DESCRIPTOR = _PEERS,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.Peers)
  ))
_sym_db.RegisterMessage(Peers)

DevConfig = _reflection.GeneratedProtocolMessageType('DevConfig', (_message.Message,), dict(

  Chain = _reflection.GeneratedProtocolMessageType('Chain', (_message.Message,), dict(
    DESCRIPTOR = _DEVCONFIG_CHAIN,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.DevConfig.Chain)
    ))
  ,

  Block = _reflection.GeneratedProtocolMessageType('Block', (_message.Message,), dict(

    BlockSizeController = _reflection.GeneratedProtocolMessageType('BlockSizeController', (_message.Message,), dict(
      DESCRIPTOR = _DEVCONFIG_BLOCK_BLOCKSIZECONTROLLER,
      __module__ = 'qrl_pb2'
      # @@protoc_insertion_point(class_scope:qrl.DevConfig.Block.BlockSizeController)
      ))
    ,
    DESCRIPTOR = _DEVCONFIG_BLOCK,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.DevConfig.Block)
    ))
  ,

  Transaction = _reflection.GeneratedProtocolMessageType('Transaction', (_message.Message,), dict(

    Message = _reflection.GeneratedProtocolMessageType('Message', (_message.Message,), dict(
      DESCRIPTOR = _DEVCONFIG_TRANSACTION_MESSAGE,
      __module__ = 'qrl_pb2'
      # @@protoc_insertion_point(class_scope:qrl.DevConfig.Transaction.Message)
      ))
    ,

    Slave = _reflection.GeneratedProtocolMessageType('Slave', (_message.Message,), dict(
      DESCRIPTOR = _DEVCONFIG_TRANSACTION_SLAVE,
      __module__ = 'qrl_pb2'
      # @@protoc_insertion_point(class_scope:qrl.DevConfig.Transaction.Slave)
      ))
    ,

    Token = _reflection.GeneratedProtocolMessageType('Token', (_message.Message,), dict(
      DESCRIPTOR = _DEVCONFIG_TRANSACTION_TOKEN,
      __module__ = 'qrl_pb2'
      # @@protoc_insertion_point(class_scope:qrl.DevConfig.Transaction.Token)
      ))
    ,

    Lattice = _reflection.GeneratedProtocolMessageType('Lattice', (_message.Message,), dict(
      DESCRIPTOR = _DEVCONFIG_TRANSACTION_LATTICE,
      __module__ = 'qrl_pb2'
      # @@protoc_insertion_point(class_scope:qrl.DevConfig.Transaction.Lattice)
      ))
    ,

    FoundationMultiSig = _reflection.GeneratedProtocolMessageType('FoundationMultiSig', (_message.Message,), dict(
      DESCRIPTOR = _DEVCONFIG_TRANSACTION_FOUNDATIONMULTISIG,
      __module__ = 'qrl_pb2'
      # @@protoc_insertion_point(class_scope:qrl.DevConfig.Transaction.FoundationMultiSig)
      ))
    ,

    Proposal = _reflection.GeneratedProtocolMessageType('Proposal', (_message.Message,), dict(
      DESCRIPTOR = _DEVCONFIG_TRANSACTION_PROPOSAL,
      __module__ = 'qrl_pb2'
      # @@protoc_insertion_point(class_scope:qrl.DevConfig.Transaction.Proposal)
      ))
    ,
    DESCRIPTOR = _DEVCONFIG_TRANSACTION,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.DevConfig.Transaction)
    ))
  ,

  POW = _reflection.GeneratedProtocolMessageType('POW', (_message.Message,), dict(
    DESCRIPTOR = _DEVCONFIG_POW,
    __module__ = 'qrl_pb2'
    # @@protoc_insertion_point(class_scope:qrl.DevConfig.POW)
    ))
  ,
  DESCRIPTOR = _DEVCONFIG,
  __module__ = 'qrl_pb2'
  # @@protoc_insertion_point(class_scope:qrl.DevConfig)
  ))
_sym_db.RegisterMessage(DevConfig)
_sym_db.RegisterMessage(DevConfig.Chain)
_sym_db.RegisterMessage(DevConfig.Block)
_sym_db.RegisterMessage(DevConfig.Block.BlockSizeController)
_sym_db.RegisterMessage(DevConfig.Transaction)
_sym_db.RegisterMessage(DevConfig.Transaction.Message)
_sym_db.RegisterMessage(DevConfig.Transaction.Slave)
_sym_db.RegisterMessage(DevConfig.Transaction.Token)
_sym_db.RegisterMessage(DevConfig.Transaction.Lattice)
_sym_db.RegisterMessage(DevConfig.Transaction.FoundationMultiSig)
_sym_db.RegisterMessage(DevConfig.Transaction.Proposal)
_sym_db.RegisterMessage(DevConfig.POW)


_ADDRESSSTATE_TOKENSENTRY.has_options = True
_ADDRESSSTATE_TOKENSENTRY._options = _descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001'))
_ADDRESSSTATE_SLAVEPKSACCESSTYPEENTRY.has_options = True
_ADDRESSSTATE_SLAVEPKSACCESSTYPEENTRY._options = _descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001'))
_TRANSACTIONCOUNT_COUNTENTRY.has_options = True
_TRANSACTIONCOUNT_COUNTENTRY._options = _descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001'))

_PUBLICAPI = _descriptor.ServiceDescriptor(
  name='PublicAPI',
  full_name='qrl.PublicAPI',
  file=DESCRIPTOR,
  index=0,
  options=None,
  serialized_start=15444,
  serialized_end=19217,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetNodeState',
    full_name='qrl.PublicAPI.GetNodeState',
    index=0,
    containing_service=None,
    input_type=_GETNODESTATEREQ,
    output_type=_GETNODESTATERESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\r\022\013/node-state')),
  ),
  _descriptor.MethodDescriptor(
    name='GetKnownPeers',
    full_name='qrl.PublicAPI.GetKnownPeers',
    index=1,
    containing_service=None,
    input_type=_GETKNOWNPEERSREQ,
    output_type=_GETKNOWNPEERSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\016\022\014/known-peers')),
  ),
  _descriptor.MethodDescriptor(
    name='GetPeersStat',
    full_name='qrl.PublicAPI.GetPeersStat',
    index=2,
    containing_service=None,
    input_type=_GETPEERSSTATREQ,
    output_type=_GETPEERSSTATRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\r\022\013/peers-stat')),
  ),
  _descriptor.MethodDescriptor(
    name='GetStats',
    full_name='qrl.PublicAPI.GetStats',
    index=3,
    containing_service=None,
    input_type=_GETSTATSREQ,
    output_type=_GETSTATSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\010\022\006/stats')),
  ),
  _descriptor.MethodDescriptor(
    name='GetAddressState',
    full_name='qrl.PublicAPI.GetAddressState',
    index=4,
    containing_service=None,
    input_type=_GETADDRESSSTATEREQ,
    output_type=_GETADDRESSSTATERESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\020\022\016/address-state')),
  ),
  _descriptor.MethodDescriptor(
    name='GetOptimizedAddressState',
    full_name='qrl.PublicAPI.GetOptimizedAddressState',
    index=5,
    containing_service=None,
    input_type=_GETADDRESSSTATEREQ,
    output_type=_GETOPTIMIZEDADDRESSSTATERESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\032\022\030/optimized-address-state')),
  ),
  _descriptor.MethodDescriptor(
    name='GetMultiSigAddressState',
    full_name='qrl.PublicAPI.GetMultiSigAddressState',
    index=6,
    containing_service=None,
    input_type=_GETMULTISIGADDRESSSTATEREQ,
    output_type=_GETMULTISIGADDRESSSTATERESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\032\022\030/multi-sig-address-state')),
  ),
  _descriptor.MethodDescriptor(
    name='IsSlave',
    full_name='qrl.PublicAPI.IsSlave',
    index=7,
    containing_service=None,
    input_type=_ISSLAVEREQ,
    output_type=_ISSLAVERESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\013\022\t/is-slave')),
  ),
  _descriptor.MethodDescriptor(
    name='GetObject',
    full_name='qrl.PublicAPI.GetObject',
    index=8,
    containing_service=None,
    input_type=_GETOBJECTREQ,
    output_type=_GETOBJECTRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\t\022\007/object')),
  ),
  _descriptor.MethodDescriptor(
    name='GetLatestData',
    full_name='qrl.PublicAPI.GetLatestData',
    index=9,
    containing_service=None,
    input_type=_GETLATESTDATAREQ,
    output_type=_GETLATESTDATARESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\016\022\014/latest-data')),
  ),
  _descriptor.MethodDescriptor(
    name='PushTransaction',
    full_name='qrl.PublicAPI.PushTransaction',
    index=10,
    containing_service=None,
    input_type=_PUSHTRANSACTIONREQ,
    output_type=_PUSHTRANSACTIONRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\023\"\021/push-transaction')),
  ),
  _descriptor.MethodDescriptor(
    name='TransferCoins',
    full_name='qrl.PublicAPI.TransferCoins',
    index=11,
    containing_service=None,
    input_type=_TRANSFERCOINSREQ,
    output_type=_TRANSFERCOINSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\021\"\017/transfer-coins')),
  ),
  _descriptor.MethodDescriptor(
    name='ParseAddress',
    full_name='qrl.PublicAPI.ParseAddress',
    index=12,
    containing_service=None,
    input_type=_PARSEADDRESSREQ,
    output_type=_PARSEADDRESSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\020\022\016/parse-address')),
  ),
  _descriptor.MethodDescriptor(
    name='GetChainStats',
    full_name='qrl.PublicAPI.GetChainStats',
    index=13,
    containing_service=None,
    input_type=_GETCHAINSTATSREQ,
    output_type=_GETCHAINSTATSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\016\022\014/chain-stats')),
  ),
  _descriptor.MethodDescriptor(
    name='GetAddressFromPK',
    full_name='qrl.PublicAPI.GetAddressFromPK',
    index=14,
    containing_service=None,
    input_type=_GETADDRESSFROMPKREQ,
    output_type=_GETADDRESSFROMPKRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\022\022\020/address-from-pk')),
  ),
  _descriptor.MethodDescriptor(
    name='GetMultiSigCreateTxn',
    full_name='qrl.PublicAPI.GetMultiSigCreateTxn',
    index=15,
    containing_service=None,
    input_type=_MULTISIGCREATETXNREQ,
    output_type=_TRANSFERCOINSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\027\"\025/multi-sig-create-txn')),
  ),
  _descriptor.MethodDescriptor(
    name='GetMultiSigSpendTxn',
    full_name='qrl.PublicAPI.GetMultiSigSpendTxn',
    index=16,
    containing_service=None,
    input_type=_MULTISIGSPENDTXNREQ,
    output_type=_TRANSFERCOINSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\026\"\024/multi-sig-spend-txn')),
  ),
  _descriptor.MethodDescriptor(
    name='GetMultiSigVoteTxn',
    full_name='qrl.PublicAPI.GetMultiSigVoteTxn',
    index=17,
    containing_service=None,
    input_type=_MULTISIGVOTETXNREQ,
    output_type=_TRANSFERCOINSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\025\"\023/multi-sig-vote-txn')),
  ),
  _descriptor.MethodDescriptor(
    name='GetMessageTxn',
    full_name='qrl.PublicAPI.GetMessageTxn',
    index=18,
    containing_service=None,
    input_type=_MESSAGETXNREQ,
    output_type=_TRANSFERCOINSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\016\"\014/message-txn')),
  ),
  _descriptor.MethodDescriptor(
    name='GetTokenTxn',
    full_name='qrl.PublicAPI.GetTokenTxn',
    index=19,
    containing_service=None,
    input_type=_TOKENTXNREQ,
    output_type=_TRANSFERCOINSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\014\"\n/token-txn')),
  ),
  _descriptor.MethodDescriptor(
    name='GetTransferTokenTxn',
    full_name='qrl.PublicAPI.GetTransferTokenTxn',
    index=20,
    containing_service=None,
    input_type=_TRANSFERTOKENTXNREQ,
    output_type=_TRANSFERCOINSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\025\"\023/transfer-token-txn')),
  ),
  _descriptor.MethodDescriptor(
    name='GetSlaveTxn',
    full_name='qrl.PublicAPI.GetSlaveTxn',
    index=21,
    containing_service=None,
    input_type=_SLAVETXNREQ,
    output_type=_TRANSFERCOINSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\014\"\n/slave-txn')),
  ),
  _descriptor.MethodDescriptor(
    name='GetLatticeTxn',
    full_name='qrl.PublicAPI.GetLatticeTxn',
    index=22,
    containing_service=None,
    input_type=_LATTICETXNREQ,
    output_type=_TRANSFERCOINSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\016\"\014/lattice-txn')),
  ),
  _descriptor.MethodDescriptor(
    name='GetTransaction',
    full_name='qrl.PublicAPI.GetTransaction',
    index=23,
    containing_service=None,
    input_type=_GETTRANSACTIONREQ,
    output_type=_GETTRANSACTIONRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\016\022\014/transaction')),
  ),
  _descriptor.MethodDescriptor(
    name='GetMiniTransactionsByAddress',
    full_name='qrl.PublicAPI.GetMiniTransactionsByAddress',
    index=24,
    containing_service=None,
    input_type=_GETMINITRANSACTIONSBYADDRESSREQ,
    output_type=_GETMINITRANSACTIONSBYADDRESSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\036\022\034/mini-transaction-by-address')),
  ),
  _descriptor.MethodDescriptor(
    name='GetTransactionsByAddress',
    full_name='qrl.PublicAPI.GetTransactionsByAddress',
    index=25,
    containing_service=None,
    input_type=_GETTRANSACTIONSBYADDRESSREQ,
    output_type=_GETTRANSACTIONSBYADDRESSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\032\022\030/transactions-by-address')),
  ),
  _descriptor.MethodDescriptor(
    name='GetTokensByAddress',
    full_name='qrl.PublicAPI.GetTokensByAddress',
    index=26,
    containing_service=None,
    input_type=_GETTRANSACTIONSBYADDRESSREQ,
    output_type=_GETTOKENSBYADDRESSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\024\022\022/tokens-by-address')),
  ),
  _descriptor.MethodDescriptor(
    name='GetSlavesByAddress',
    full_name='qrl.PublicAPI.GetSlavesByAddress',
    index=27,
    containing_service=None,
    input_type=_GETTRANSACTIONSBYADDRESSREQ,
    output_type=_GETSLAVESBYADDRESSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\024\022\022/slaves-by-address')),
  ),
  _descriptor.MethodDescriptor(
    name='GetLatticePKsByAddress',
    full_name='qrl.PublicAPI.GetLatticePKsByAddress',
    index=28,
    containing_service=None,
    input_type=_GETTRANSACTIONSBYADDRESSREQ,
    output_type=_GETLATTICEPKSBYADDRESSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\031\022\027/lattice-pks-by-address')),
  ),
  _descriptor.MethodDescriptor(
    name='GetMultiSigAddressesByAddress',
    full_name='qrl.PublicAPI.GetMultiSigAddressesByAddress',
    index=29,
    containing_service=None,
    input_type=_GETTRANSACTIONSBYADDRESSREQ,
    output_type=_GETMULTISIGADDRESSESBYADDRESSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002!\022\037/multi-sig-addresses-by-address')),
  ),
  _descriptor.MethodDescriptor(
    name='GetMultiSigSpendTxsByAddress',
    full_name='qrl.PublicAPI.GetMultiSigSpendTxsByAddress',
    index=30,
    containing_service=None,
    input_type=_GETMULTISIGSPENDTXSBYADDRESSREQ,
    output_type=_GETMULTISIGSPENDTXSBYADDRESSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002!\022\037/multi-sig-spend-txs-by-address')),
  ),
  _descriptor.MethodDescriptor(
    name='GetVoteStats',
    full_name='qrl.PublicAPI.GetVoteStats',
    index=31,
    containing_service=None,
    input_type=_GETVOTESTATSREQ,
    output_type=_GETVOTESTATSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\r\022\013/vote-stats')),
  ),
  _descriptor.MethodDescriptor(
    name='GetInboxMessagesByAddress',
    full_name='qrl.PublicAPI.GetInboxMessagesByAddress',
    index=32,
    containing_service=None,
    input_type=_GETTRANSACTIONSBYADDRESSREQ,
    output_type=_GETINBOXMESSAGESBYADDRESSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\034\022\032/inbox-messages-by-address')),
  ),
  _descriptor.MethodDescriptor(
    name='GetBalance',
    full_name='qrl.PublicAPI.GetBalance',
    index=33,
    containing_service=None,
    input_type=_GETBALANCEREQ,
    output_type=_GETBALANCERESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\n\022\010/balance')),
  ),
  _descriptor.MethodDescriptor(
    name='GetTotalBalance',
    full_name='qrl.PublicAPI.GetTotalBalance',
    index=34,
    containing_service=None,
    input_type=_GETTOTALBALANCEREQ,
    output_type=_GETTOTALBALANCERESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\020\022\016/total-balance')),
  ),
  _descriptor.MethodDescriptor(
    name='GetOTS',
    full_name='qrl.PublicAPI.GetOTS',
    index=35,
    containing_service=None,
    input_type=_GETOTSREQ,
    output_type=_GETOTSRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\006\022\004/ots')),
  ),
  _descriptor.MethodDescriptor(
    name='GetHeight',
    full_name='qrl.PublicAPI.GetHeight',
    index=36,
    containing_service=None,
    input_type=_GETHEIGHTREQ,
    output_type=_GETHEIGHTRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\t\022\007/height')),
  ),
  _descriptor.MethodDescriptor(
    name='GetBlock',
    full_name='qrl.PublicAPI.GetBlock',
    index=37,
    containing_service=None,
    input_type=_GETBLOCKREQ,
    output_type=_GETBLOCKRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\010\022\006/block')),
  ),
  _descriptor.MethodDescriptor(
    name='GetBlockByNumber',
    full_name='qrl.PublicAPI.GetBlockByNumber',
    index=38,
    containing_service=None,
    input_type=_GETBLOCKBYNUMBERREQ,
    output_type=_GETBLOCKBYNUMBERRESP,
    options=_descriptor._ParseOptions(descriptor_pb2.MethodOptions(), _b('\202\323\344\223\002\022\022\020/block-by-number')),
  ),
])
_sym_db.RegisterServiceDescriptor(_PUBLICAPI)

DESCRIPTOR.services_by_name['PublicAPI'] = _PUBLICAPI


_ADMINAPI = _descriptor.ServiceDescriptor(
  name='AdminAPI',
  full_name='qrl.AdminAPI',
  file=DESCRIPTOR,
  index=1,
  options=None,
  serialized_start=19219,
  serialized_end=19229,
  methods=[
])
_sym_db.RegisterServiceDescriptor(_ADMINAPI)

DESCRIPTOR.services_by_name['AdminAPI'] = _ADMINAPI

# @@protoc_insertion_point(module_scope)
